/*
	$Id: Base.java 7260 2016-12-12 07:50:49Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web;

import net.sf.basedb.core.AnnotatableProxy;
import net.sf.basedb.core.AnyToAny;
import net.sf.basedb.core.Coloring;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DateUtil;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileSet;
import net.sf.basedb.core.FileStoreEnabled;
import net.sf.basedb.core.ItemList;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.Location;
import net.sf.basedb.core.Platform;
import net.sf.basedb.core.PlatformVariant;
import net.sf.basedb.core.Protocol;
import net.sf.basedb.core.ReporterList;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Application;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.Nameable;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.PropertyFilter;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.StringUtil;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Operator;
import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.AnnotationSet;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.Annotation;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.Unit;
import net.sf.basedb.core.ItemContext.SortDirection;
import net.sf.basedb.core.query.EntityQuery;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.ItemModifiedException;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.NotLoggedInException;
import net.sf.basedb.util.ColorGenerator;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.formatter.Formatter;
import net.sf.basedb.clients.web.formatter.FormatterFactory;
import net.sf.basedb.clients.web.util.HTML;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Arrays;
import java.util.Enumeration;

import javax.servlet.jsp.PageContext;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.ServletRequest;
import javax.servlet.http.Cookie;

import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.json.simple.parser.JSONParser;
import org.json.simple.parser.ParseException;


/**
	This class is used to collect some useful methods for the web client
	into a single place. All methods in this class are static methods.

	@author Nicklas
	@version 2.0
	@base.modified $Date$
*/
public final class Base
{

	/**
		The external ID of the web client application.
		@since 3.9
	*/
	public static final String WEBCLIENT_ID = "net.sf.basedb.clients.web";
	
	/**
		Get a new or existing <code>SessionControl</code> object
		given a <code>PageContext</code>. This method will get
		the <code>ID</code> parameter from the page URL and
		try to retrieve the corresponding object from the
		BASE {@link Application}. If no {@link SessionControl}
		object exists, a new one will be created if the <code>create</code> 
		parameter is TRUE.

		@param pageContext The JSP <code>pageContext</code> variable
		@param create If a new <code>SessionControl</code> should be created or not
		@return A <code>SessionControl</code> object
		@throws BaseException If there is an error
	*/
	public static final SessionControl getSessionControl(PageContext pageContext, boolean create)
		throws BaseException
	{
		return getSessionControl(pageContext.getRequest(), create);
	}

	/**
		Get a session control for the BASE web client "net.sf.basedb.clients.web".
		
		@see #getSessionControl(ServletRequest, String, boolean)
		@since 2.7
	*/
	public static final SessionControl getSessionControl(ServletRequest request, boolean create)
	{
		return getSessionControl(request, Base.WEBCLIENT_ID, create);
	}

	/**
		Get a new or existing <code>SessionControl</code> object
		given a <code>ServletRequest</code> instance. This method will get
		the <code>ID</code> parameter from the page URL and
		try to retrieve the corresponding object from the
		BASE {@link Application}. If no {@link SessionControl}
		object exists, a new one will be created if the <code>create</code> 
		parameter is TRUE.
	
		@param request The Servlet <code>request</code> variable
		@param externalClientId ID of the client application
		@param create If a new <code>SessionControl</code> should be created or not
		@return A <code>SessionControl</code> object
		@throws BaseException If there is an error
		@since 3.9
	*/
	public static final SessionControl getSessionControl(ServletRequest request, String externalClientId, boolean create)
	{
		SessionControl sc = null;
		String id = request.getParameter("ID");
		if (id == null && request.getAttribute("ID") instanceof String) 
		{
			id = (String)request.getAttribute("ID");
		}
		String remoteId = request.getRemoteAddr();
		try
		{
			sc = Application.getSessionControl(id, externalClientId, remoteId, false);
			if (sc.isImpersonated())
			{
				// To keep the original session control from timing out we update the last access
				SessionControl impersonator = (SessionControl)sc.getSessionSetting("impersonate.originalSessionControl");
				if (impersonator != null) impersonator.updateLastAccess();
			}
		}
		catch (ItemNotFoundException ex)
		{}
		return sc == null && create ? Application.newSessionControl(externalClientId, remoteId, id) : sc;
	}
	
	/**
		Get an existing {@link SessionControl} object
		given a <code>PageContext</code> and optionally check that a user is logged in. 
		This method will get the <code>ID</code> parameter from the page URL and
		try to retrieve the corresponding object from the
		{@link Application} and check that a user is logged in if
		the <code>checkLoggedIn</code> parameter is TRUE.

		@param pageContext The JSP <code>pageContext</code> variable
		@param checkLoggedIn If the method should check if a user is logged in or not
		@return A <code>SessionControl</code> object
		@throws NotLoggedInException If a <code>SessionControl</code> object isn't found 
			or no user is logged in
		@throws BaseException If there is another error
	*/
	public static final SessionControl getExistingSessionControl(PageContext pageContext, boolean checkLoggedIn)
		throws NotLoggedInException, BaseException
	{
		return getExistingSessionControl(pageContext.getRequest(), checkLoggedIn);
	}

	/**
		Get a session control for the BASE web client "net.sf.basedb.clients.web".
		
		@see #getExistingSessionControl(ServletRequest, String, boolean)
		@since 2.7
	*/
	public static final SessionControl getExistingSessionControl(ServletRequest request, boolean checkLoggedIn)
		throws NotLoggedInException, BaseException
	{
		return getExistingSessionControl(request, WEBCLIENT_ID, checkLoggedIn);
	}
	
	/**
		Get an existing {@link SessionControl} object
		given a <code>ServletRequest</code> object and optionally check that a user is logged in. 
		This method will get the <code>ID</code> parameter from the page URL and
		try to retrieve the corresponding object from the
		{@link Application} and check that a user is logged in if
		the <code>checkLoggedIn</code> parameter is TRUE.
	
		@param request The Servlet <code>request</code> variable
		@param externalClientId ID of the client application
		@param checkLoggedIn If the method should check if a user is logged in or not
		@return A <code>SessionControl</code> object
		@throws NotLoggedInException If a <code>SessionControl</code> object isn't found 
			or no user is logged in
		@throws BaseException If there is another error
		@since 3.9
	*/
	public static final SessionControl getExistingSessionControl(ServletRequest request, String externalClientId, boolean checkLoggedIn)
		throws NotLoggedInException, BaseException
	{
		SessionControl sc = getSessionControl(request, externalClientId, false);
		if (sc == null || (checkLoggedIn && !sc.isLoggedIn()))
		{
			throw new NotLoggedInException();
		}
		return sc;
	}
	
	
	/**
		Get an existing {@link SessionControl} object given 
		a <code>PageContext</code> and check for generic (role-based)
		permission to an item type. 
 
		This method will get the <code>ID</code> parameter from the page URL and
		try to retrieve the corresponding object from the {@link Application}.
		Then it checks that a user is logged in and that the logged in user has the 
		specified permission for item type. If {@link Permission#DENIED} is specified, 
		this method checks that the logged in user haven't been denied acces to
		those items.

		@param pageContext The JSP <code>pageContext</code> variable
		@param permission The permission to check for, use one of the codes
			defined by the {@link Permission} class
		@param itemType The code for the item type to check for the permission,
			use one of the values defined by the {@link Item} class
		@return A <code>SessionControl</code> object
		@throws PermissionDeniedException If the logged in user doesn't have the
			requested permission
		@throws NotLoggedInException If no user is logged in or a <code>SessionControl</code>
			object not is found
		@throws BaseException If there is another error
	*/
	public static final SessionControl getExistingSessionControl(PageContext pageContext, Permission permission, Item itemType)
		throws PermissionDeniedException, NotLoggedInException, BaseException
	{
		return getExistingSessionControl(pageContext.getRequest(), permission, itemType);
	}

	/**
		Get an existing {@link SessionControl} object given 
		a <code>ServletRequest</code> and check for generic (role-based)
		permission to an item type. 
	
		This method will get the <code>ID</code> parameter from the page URL and
		try to retrieve the corresponding object from the {@link Application}.
		Then it checks that a user is logged in and that the logged in user has the 
		specified permission for item type. If {@link Permission#DENIED} is specified, 
		this method checks that the logged in user haven't been denied acces to
		those items.
	
		@param request The Servlet <code>request</code> variable
		@param permission The permission to check for, use one of the codes
			defined by the {@link Permission} class
		@param itemType The code for the item type to check for the permission,
			use one of the values defined by the {@link Item} class
		@return A <code>SessionControl</code> object
		@throws PermissionDeniedException If the logged in user doesn't have the
			requested permission
		@throws NotLoggedInException If no user is logged in or a <code>SessionControl</code>
			object not is found
		@throws BaseException If there is another error
		@since 2.7
	*/
	public static final SessionControl getExistingSessionControl(ServletRequest request, 
		Permission permission, Item itemType)
		throws PermissionDeniedException, NotLoggedInException, BaseException
	{
		SessionControl sc = getExistingSessionControl(request, true);
		boolean hasPermission = sc.hasPermission(permission, itemType);
		boolean checkDenied = permission == Permission.DENIED;
		if (checkDenied && hasPermission || !checkDenied && !hasPermission)
		{
			throw new PermissionDeniedException(permission, itemType.toString());
		}
		return sc;
	}	
	
	/**
		Same as: <code>getAndSetCurrentContext(sc, itemType, "", pageContext, defaultContext, false)</code>.
		@see #getAndSetCurrentContext(SessionControl, Item, String, PageContext, ItemContext, boolean)
	*/
	public static ItemContext getAndSetCurrentContext(SessionControl sc, Item itemType, 
			PageContext pageContext, ItemContext defaultContext)
	{
		return getAndSetCurrentContext(sc, itemType, "", pageContext, defaultContext, false);
	}
	
	/**
		Same as: <code>getAndSetCurrentContext(sc, itemType, "", pageContext, defaultContext, resetTemporary)</code>.
		@see #getAndSetCurrentContext(SessionControl, Item, String, PageContext, ItemContext, boolean)
	*/
	public static ItemContext getAndSetCurrentContext(SessionControl sc, Item itemType, 
			PageContext pageContext, ItemContext defaultContext, boolean resetTemporary)
	{
		return getAndSetCurrentContext(sc, itemType, "", pageContext, defaultContext, resetTemporary);
	}
	
	/**
		Same as: <code>getAndSetCurrentContext(sc, itemType, subContext, pageContext, defaultContext, false)</code>.
		@see #getAndSetCurrentContext(SessionControl, Item, String, PageContext, ItemContext, boolean)
	*/
	public static ItemContext getAndSetCurrentContext(SessionControl sc, Item itemType, 
			String subContext, PageContext pageContext, ItemContext defaultContext)
	{
		return getAndSetCurrentContext(sc, itemType, subContext, pageContext, 
				defaultContext, false);
	}
	
	/**
		@see #getAndSetCurrentContext(SessionControl, Item, String, ServletRequest, ItemContext, boolean)
	*/
	public static ItemContext getAndSetCurrentContext(SessionControl sc, Item itemType, 
		String subContext, PageContext pageContext, ItemContext defaultContext, boolean resetTemporary)
	{
		ServletRequest request = pageContext == null ? null : pageContext.getRequest();
		return getAndSetCurrentContext(sc, itemType, subContext, request, 
				defaultContext, resetTemporary);
	}
	
	/**
		Get the current context and update it from request information if available.
		
		This methods updates the following settings on the context:
		<ul>
		<li>If the active project has changed since last time the context
			was retrieved, the {@link Include#IN_PROJECT} option is added
			to {@link ItemContext#getInclude()}
		<li>If a default context is specified, the <code>defaultColumns</code>
			object is copied to the current context.
			
		<li>The following settings are only updated if they are available in the
			request parameters:
			
			<table>
			<caption>Context settings</caption>
			<tr>
				<th>Setting</th>
				<th>Request parameter</th>
			</tr>
			<tr>
				<td>setSetting("columns")</td>
				<td>columns</td>
			</tr>
			<tr>
				<td>setRowsPerPage</td>
				<td>rowsperpage</td>
			</tr>
			<tr>
				<td>setPage</td>
				<td>page</td>
			</tr>
			<tr>
				<td>setSortProperty</td>
				<td>sortby</td>
			</tr>
			<tr>
				<td>setSortDirection</td>
				<td>direction</td>
			</tr>
			<tr>
				<td>Include.REMOVED</td>
				<td>removed</td>
			</tr>
			<tr>
				<td>Include.MINE</td>
				<td>owned</td>
			</tr>
			<tr>
				<td>Include.SHARED</td>
				<td>shared</td>
			</tr>
			<tr>
				<td>Include.IN_PROJECT</td>
				<td>inproject</td>
			</tr>
			<tr>
				<td>Include.OTHERS</td>
				<td>others</td>
			</tr>
			</table>
			
			<p>
			If a the name of a request parameter starts with "filter:" it is interpreted as a
			property filter. The name is formatted as follows: <code>filter:TYPE:property</code>
			where TYPE is one of {@link Type#name()} strings and property is the name of the
			property to filter on. This information is used to create a new {@link PropertyFilter}
			object. The operator and value parameter is taken from the request parameters value.
			<table>
			<caption>Operator mappings</caption>
			<tr>
				<th>Value starts with</th>
				<th>Operator</th>
			</tr>
			<tr>
				<td>=</td>
				<td>Operator.EQ</td>
			</tr>
			<tr>
				<td>!= or &lt;&gt;</td>
				<td>Operator.NEQ</td>
			</tr>
			<tr>
				<td>&gt;=</td>
				<td>Operator.GTEQ</td>
			</tr>
			<tr>
				<td>&gt;</td>
				<td>Operator.GT</td>
			</tr>
			<tr>
				<td>&lt;=</td>
				<td>Operator.LTEQ</td>
			</tr>
			<tr>
				<td>&lt;</td>
				<td>Operator.LT</td>
			</tr>
			<tr>
				<td>any other charcter</td>
				<td>Operator.EQ or Operator.LIKE (if value contains % sign which is wildcard match)</td>
			</tr>
			</table>
			
			<p>
			If the name of a request parameter is a number it is interpreted as the ID of
			a selected itema and is added to the {@link ItemContext#getSelected()} set.
			
			<ul>
			<li>
			</ul>
		</ul>
		
		@param sc The current session control
		@param itemType The type of item to get the context for
		@param subContext The name of the subcontext
		@param request A <code>ServletRequest</code> object to get request
			information from, or null if the current context shouldn't be updated
		@param defaultContext A default context to copy information from if a current context
			doesn't exist
		@param resetTemporary If information of temporary nature should be reset or not, for
			example the 'exclude' filter which is only used in popup windows
	 	@return an updated {@link net.sf.basedb.core.ItemContext} object.
	 	@since 2.7
	*/
	public static ItemContext getAndSetCurrentContext(SessionControl sc, Item itemType, 
		String subContext, ServletRequest request, ItemContext defaultContext, boolean resetTemporary)
	{
		ItemContext cc = sc.getCurrentContext(itemType, subContext, defaultContext);
		ItemContext dirContext = itemType == Item.FILE ? 
			sc.getCurrentContext(Item.DIRECTORY, subContext) : null; // Used in special case for file manager
		
		// Used in special case for analysis tree
		ItemContext transformationContext = null;
		ItemContext extraValueContext = null;
		
		if (itemType == Item.BIOASSAYSET)
		{
			transformationContext = sc.getCurrentContext(Item.TRANSFORMATION, subContext);
			extraValueContext = sc.getCurrentContext(Item.EXTRAVALUE, subContext);
		}
		
		if (resetTemporary || request != null && Values.getBoolean(request.getParameter("resetTemporary")))
		{
			// Clear the 'exclude' option
			cc.removeObject("exclude");
			cc.removeAllTemporaryFilters();
		}
		
		// Detect if active project has changed and then set the include option
		int projectId = sc.getActiveProjectId();
		if (projectId != 0)
		{
			Object lastActive = cc.getObject("activeProject");
			if (lastActive == null || (Integer)lastActive != projectId)
			{
				cc.getInclude().add(Include.IN_PROJECT);
			}
			cc.setObject("activeProject", projectId);
		}
		else
		{
			cc.removeObject("activeProject");
		}
		
		// Copy default columns so it is always available in other contexts
		if (defaultContext != null)
		{
			cc.setObject("defaultColumns", defaultContext.getObject("defaultColumns"));
		}
		if (request != null)
		{

			// If the filter has changed we must return to the first page
			boolean filterHasChanged = false;
			
			// Current item
			String itemId = Values.getStringOrNull(request.getParameter("item_id"));
			cc.setId(Values.getInt(itemId, cc.getId()));
			
			// Record this in the list of recently view items
			Class<?> itemClass = itemType.getItemClass();
			if (itemId != null && itemClass != null && Nameable.class.isAssignableFrom(itemClass) && itemClass != AnyToAny.class)
			{
				String recent = sc.getUserClientSetting("menu.mostRecent.viewed");
				List<String> recentItems = null;
				if (recent == null)
				{
					recentItems = new ArrayList<String>();
				}
				else
				{
					// Need a new list since Arrays.asList returns a read-only list
					recentItems = new ArrayList<String>(Arrays.asList(recent.split(":")));
				}
				String thisItem = itemType.name()+"="+itemId;
				recentItems.remove(thisItem);
				recentItems.add(0, thisItem);
				int maxRecent = Values.getInt(sc.getUserClientSetting("menu.mostRecent.maxViewed"), 6);
				if (recentItems.size() > maxRecent) recentItems = recentItems.subList(0, maxRecent);
				sc.setUserClientSetting("menu.mostRecent.viewed", StringUtil.join(recentItems, ":", true));
			}
			// Visible columns
			cc.setSetting("columns", Values.getString(Values.getStringOrNull(request.getParameter("columns")), cc.getSetting("columns")));

			// Page settings, ignore negative values, 'rowsperpage' setting may come as an
			// array since a navigation bar can be present both before and after the table
			int currentRowsPerPage = cc.getRowsPerPage();
			String[] rowsPerPageParams = request.getParameterValues("rowsperpage");
			if (rowsPerPageParams != null)
			{
				for (String param : rowsPerPageParams)
				{
					int rowsPerPage = Values.getInt(param, currentRowsPerPage);
					if (rowsPerPage != currentRowsPerPage && rowsPerPage > 0) cc.setRowsPerPage(rowsPerPage);
				}
			}
			int page = Values.getInt(request.getParameter("page"), cc.getPage());
			if (page >= 0) cc.setPage(page);
	
			// Sort property and direction
			cc.setSortProperty(Values.getString(Values.getStringOrNull(request.getParameter("sortby")), cc.getSortProperty()));
			String direction = Values.getStringOrNull(request.getParameter("direction"));
			if (direction != null)
			{
				cc.setSortDirection(ItemContext.SortDirection.valueOf(direction));
			}
	
			// Include options
			Set<Include> includes = cc.getInclude();
			includes.add(Include.NOT_REMOVED);
			filterHasChanged = setInclude(includes, request.getParameter("removed"), Include.REMOVED);
			filterHasChanged |= setInclude(includes, request.getParameter("owned"), Include.MINE);
			filterHasChanged |= setInclude(includes, request.getParameter("others"), Include.OTHERS);
			filterHasChanged |= setInclude(includes, request.getParameter("shared"), Include.SHARED);
			filterHasChanged |= setInclude(includes, request.getParameter("inproject"), Include.IN_PROJECT);
			
			// Item permission
			String permission = request.getParameter("permission");
			if (permission != null)
			{
				Permission currentPermission = cc.getItemPermission();
				Permission newPermission = "".equals(permission) ? null : Permission.valueOf(permission);
				if (newPermission != currentPermission)
				{
					cc.setItemPermission(newPermission);
					filterHasChanged = true;
				}
			}
			
			// The ID of excluded items in a query
			String exclude = Values.getStringOrNull(request.getParameter("exclude"));
			if (exclude == null)
			{
				cc.removeObject("exclude");
			}
			else
			{
				cc.setObject("exclude", exclude);
			}
			
			// Add filter rows
			int rowsToAdd = Values.getInt(request.getParameter("addFilterRows"));
			if (rowsToAdd > 0) cc.addFilterRows(rowsToAdd);
			
			if (Values.getBoolean(request.getParameter("filter:clearAll")))
			{
				cc.removeAllPropertyFilters();
				cc.setItemPermission(null);
				filterHasChanged = true;
			}
			else
			{
				// Property filters and selected items
				// Filter parameters have names like: filter:STRING:propertyName
				// where propertyName can start with an optional operator: <>, >=, <=, >, <, =
				Enumeration<String> names = request.getParameterNames();
				cc.getSelected().clear();
				if (dirContext != null) dirContext.getSelected().clear();
				if (transformationContext != null) transformationContext.getSelected().clear();
				if (extraValueContext != null) extraValueContext.getSelected().clear();

				if (itemId != null && ModeInfo.get(request.getParameter("mode")).hasRadio())
				{
					// Hack to let Delete, Restore and Share buttons work in popup 
					// window with radio buttons
					cc.getSelected().add(Values.getInt(itemId));
				}
				
				while (names.hasMoreElements())
				{
					String name = names.nextElement();
					boolean isFilter = name.startsWith("filter:");
					boolean isTempFilter = name.startsWith("tmpfilter:");
					if (isFilter || isTempFilter)
					{
						// filter:TYPE:(unit-id:)propertyName
						String value = Values.getStringOrNull(request.getParameter(name));
						String[] parts = name.split("\\:", 4);
						String property = parts[parts.length-1];
						int filterIndex = 0;
						if (property.endsWith("]"))
						{
							int split = property.lastIndexOf('[');
							filterIndex = Integer.parseInt(property.substring(split+1, property.length()-1));
							property = property.substring(0, split);
						}
						if (value != null)
						{
							boolean hasWildcard = value.indexOf('%') >= 0;
							boolean isList = value.indexOf('|') >= 0;
							Type valueType = Type.valueOf(parts[1]);
							int unitId = parts.length > 3 ? Values.getInt(parts[2]) : 0;
	
							PropertyFilter filter = cc.getPropertyFilter(property, filterIndex);
							if (filter == null || isTempFilter != filter.isTemporary()) 
							{
								filter = new PropertyFilter(property, filterIndex);
							}
							filter.setTemporary(isTempFilter);
							filter.setValueType(valueType);
							if (unitId != 0)
							{
								DbControl dc = sc.newDbControl();
								try
								{
									filter.setUnit(Unit.getById(dc, unitId));
								}
								finally 
								{
									if (dc != null) dc.close();
								}
							}
							
							Operator op = hasWildcard ? 
									(isList ? Operator.LIKE_IN : Operator.LIKE) : 
										(isList ? Operator.IN : Operator.EQ);
							if (value.startsWith("<>") || value.startsWith("!="))
							{
								op = hasWildcard ? 
										(isList ? Operator.NOTLIKE_IN : Operator.NOTLIKE) : 
											(isList ? Operator.NOTIN : Operator.NEQ);
								value = value.substring(2);
							}
							else if (value.startsWith("=="))
							{
								op = Operator.EQ;
								value = value.substring(2);
							}
							else if (value.startsWith(">="))
							{
								op = Operator.GTEQ;
								value = value.substring(2);
							}
							else if (value.startsWith("<="))
							{
								op = Operator.LTEQ;
								value = value.substring(2);
							}
							else if (value.startsWith("><"))
							{
								op = Operator.BETWEEN;
								value = value.substring(2);
							}
							else if (value.startsWith(">"))
							{
								op = Operator.GT;
								value = value.substring(1);
							}
							else if (value.startsWith("<"))
							{
								op = Operator.LT;
								value = value.substring(1);
							}
							else if (value.startsWith("="))
							{
								op = hasWildcard ? 
										(isList ? Operator.LIKE_IN : Operator.LIKE) : 
											(isList ? Operator.IN : Operator.EQ);
								value = value.substring(1);
							}							
							if ("".equals(value)) value = null;
							if (valueType.isTemporal() && value != null)
							{
								// Temporal values are stored as long timevalues to avoid problems if user
								// changes the date format. The date is converted back when displayed
								Formatter<Date> dateFormatter = FormatterFactory.getDateFormatter(sc);
								try
								{
									if (isList)
									{	
										String[] dates = value.split("\\" + filter.getValueSeparator());
										StringBuilder tmpValue = new StringBuilder();
										String separator = null;
										for (String s : dates)
										{
											Date d = dateFormatter != null ? dateFormatter.parseString(s) : DateUtil.parseString(s);											
											if (d != null)
											{
												if (separator != null)
												{
													tmpValue.append(separator);
												}
												else
												{
													separator = filter.getValueSeparator();
												}
												tmpValue.append(Long.toString(d.getTime()));
											}
										}
										value = tmpValue.length() == 0 ? null : tmpValue.toString();										
									}
									else
									{
										Date d = dateFormatter != null ? dateFormatter.parseString(value) : DateUtil.parseString(value);
										value = d == null ? null : Long.toString(d.getTime());
									}
								}
								catch (Throwable t)
								{}
							}				
							if ((filter.getValue() == null && value != null) ||
									(filter.getOperator() == null && op != null) ||
									(filter.getValue() != null && !filter.getValue().equals(value)) ||
									(filter.getOperator() != null && !filter.getOperator().equals(op)))
							{
								filterHasChanged = true;
							}							
							filter.setOperator(op);
							filter.setValue(value);
							cc.setPropertyFilter(filter);
						}
						else
						{
							cc.removePropertyFilter(property, filterIndex);
						}
					}
					else if (dirContext != null && name.startsWith("D:"))
					{
						// Special case for file manager which lists both files and directories in the
						// same list
						int id = Values.getInt(name.substring(2), -1);
						if (id != -1) dirContext.getSelected().add(id);
					}
					else if (transformationContext != null && name.startsWith("T:"))
					{
						// Special case for analysis tree which lists both bioassay sets, 
						// transformations and extra value in the same list
						int id = Values.getInt(name.substring(2), -1);
						if (id != -1) transformationContext.getSelected().add(id);
					}
					else if (extraValueContext != null && name.startsWith("X:"))
					{
						// Special case for analysis tree which lists both bioassay sets, 
						// transformations and extra value in the same list
						int id = Values.getInt(name.substring(2), -1);
						if (id != -1) extraValueContext.getSelected().add(id);
					}
					else
					{
						int id = Values.getInt(name, -1);
						if (id != -1) cc.getSelected().add(id);
					}
				}
			}
			
			// Remove filter rows
			Set<Integer> removedFilters = new HashSet<Integer>(Arrays.asList(Values.getInt(request.getParameterValues("removeFilterRow"))));
			for (int row : removedFilters)
			{
				cc.removeFilterRow(row);
				filterHasChanged = true;
			}
			
			// Return to first page if filter has changed
			if (filterHasChanged) cc.setPage(0);
		}
		return cc;
	}
	
	public static ItemContext loadContext(SessionControl sc, int contextId, ItemContext defaultContext)
	{
		ItemContext cc = sc.loadContext(contextId, true);
		// Copy default columns so it is always available in other contexts
		if (defaultContext != null && cc != null)
		{
			cc.setObject("defaultColumns", defaultContext.getObject("defaultColumns"));
		}
		return cc;
	}
	
	/**
		Convert a property filter to a displayable string with 
		operator + value. Ie. &gt;10000, =Nicklas
		@param filter The property filter to convert
	 	@param dateFormatter Describing how the date is format locally. 
	 		The format will be yyyy-MM-dd if this is null. 
		@return A displayable string (null is never returned)
	*/
	public static String getPropertyFilterString(PropertyFilter filter, Formatter<Date> dateFormatter)
	{
		String result = "";
		if (filter != null)
		{
			Operator op = filter.getOperator();
			String value = filter.getValue();
			if (filter.getValueType().isTemporal() && value != null)
			{
				try
				{
					// Dates are stored as long time values. Reformat according to date formatter
					// or as yyyy-MM-dd if no formatter is specified
					if (!op.isListOperator())
					{
						Date d = new Date(Long.parseLong(value));
						if (dateFormatter != null)
						{
							value = dateFormatter.format(d);
						}
						else
						{
							value = DateUtil.formatDate(d);
						}
					}
					else
					{
						String[] dates = value.split("\\" + filter.getValueSeparator());
						boolean addSeparator = false;
						for (String s : dates)
						{
							Date d = new Date(Long.parseLong(s));
							value = addSeparator ? value + filter.getValueSeparator() : "";
							if (dateFormatter != null)
							{
								value += dateFormatter.format(d);								
							}
							else
							{
								value += DateUtil.formatDate(d); 
							}
							addSeparator = true;
						}
					}
				}
				catch (Throwable t)
				{}
			}
			if (op == Operator.EQ && value == null)
			{
				result = "=";
			}
			else if (op == Operator.NEQ || op == Operator.NOTLIKE || 
					op == Operator.NOTLIKE_IN || op == Operator.NOTIN)
			{
				result = "<>";
			}
			else if (op == Operator.GT)
			{
				result = ">";
			}
			else if (op == Operator.GTEQ)
			{
				result = ">=";
			}
			else if (op == Operator.LT)
			{
				result = "<";
			}
			else if (op == Operator.LTEQ)
			{
				result = "<=";
			}
			else if (op == Operator.BETWEEN)
			{
				result = "><";
			}
			else if (op == Operator.EQ && (value.contains(filter.getValueSeparator()) || value.contains("%")))
			{
				result = "==";
			}
			result += value == null ? "" : value;
		}
		return result;
	}
	
	/**
		Set include option if option parameter isn't null.
		@return TRUE if the option was changed, FALSE if not
	*/
	private static boolean setInclude(Set<Include> includes, String option, Include i)
	{
		boolean changed = false;
		if (option != null)
		{
			if (Values.getBoolean(option))
			{
				changed = includes.add(i);
			}
			else
			{
				changed = includes.remove(i);
			}
		}
		return changed;
	}
	
	/**
		Create a default <code>ItemContext</code> with settings for default
		visible columns and sort property. 'MINE' and 'SHARED' objects are set to be included.
	 	@param sortProperty The property, as a String, to sort on. If it starts with
	 		a '-', the sort is made in descending order (and the minus is removed)
	 	@param defaultColumns The visible columns as a comma separated String.
	 	@return an {@link net.sf.basedb.core.ItemContext} object
	*/
	public static ItemContext createDefaultContext(String sortProperty, String defaultColumns)
	{
		SortDirection sortDirection = SortDirection.ASC;
		if (sortProperty.startsWith("-"))
		{
			sortDirection = SortDirection.DESC;
			sortProperty = sortProperty.substring(1);
		}
		ItemContext cc = new ItemContext(sortProperty, sortDirection, 30, 0);
		cc.setSetting("columns", defaultColumns);
		cc.setObject("defaultColumns", defaultColumns);
		cc.getInclude().add(Include.MINE);
		cc.getInclude().add(Include.SHARED);
		return cc;
	} 
	
	/**
		Configures a query. This method calls {@link ItemContext#configureQuery(DbControl, EntityQuery, boolean)}
		and then adds the following filter:
		<ul>
		<li>If <code>ItemContext.getObject("exclude")</code> contains a list
			of comma-separated ID:s those are added to the query as a filter
			that excludes item with those ID:s from the query result
		</ul>
	 	@param <T> The kind of BasicItem this method is used with. 
	 	@param dc The DbControl that will be used to execute the query
		@param cc The current context
		@param autoLeftJoin If associated items should be left joined or not when
			they appear in a filter
		@param query The query to configure
	 	@param mode Info about the mode the page has. 
		@return The same query, configured according to information in the current context
		@since 2.8
	*/
	public static <T extends BasicItem> ItemQuery<T> getConfiguredQuery(DbControl dc, ItemContext cc, boolean autoLeftJoin, ItemQuery<T> query, ModeInfo mode)
	{
		cc.configureQuery(dc, query, autoLeftJoin);
		final String exclude = (String)cc.getObject("exclude");
		if (exclude != null)
		{
			query.restrict(
				Restrictions.not(
					Restrictions.in(
						Hql.property("id"),
						Expressions.parameter("_excludes_")
					)
				)
			);
			query.setParameter("_excludes_", Arrays.asList(Values.getInt(exclude.split(","))), Type.INT);
		}
		if (mode != null && mode.forceItemPermission() != null && cc.getItemPermission() == null)
		{
			query.setItemPermission(mode.forceItemPermission());
		}
		query.setFailSafe(true);
		return query;
	}
	
	/**
		Get a query that returns all annotation types defined for a particular 
		item type. The query will include all items owned by or shared to the
		logged in user or the active project.
		@param itemType The item type
		@return An <code>ItemQuery</code> object
		@see #getAnnotationTypesQuery(Item, Boolean)
	*/
	public static ItemQuery<AnnotationType> getAnnotationTypesQuery(Item itemType)
	{
		return getAnnotationTypesQuery(itemType, null);
	}
	
	/**
		Get a query that returns all annotation types defined for a particular 
		item type and, optionally, with a specific protocol parameter setting. 
		The query will include all possible annotation types.
		@param itemType The item type
		@param isProtocolParameter Specify if only annotation types assigned as protocol
			parameters should be returned, a null value ignores the protocol parameter flag
		@return An <code>ItemQuery</code> object
	*/
	public static ItemQuery<AnnotationType> getAnnotationTypesQuery(Item itemType, Boolean isProtocolParameter)
	{
		ItemQuery<AnnotationType> q = AnnotationType.getQuery(itemType);
		q.order(Orders.asc(Hql.property("name")));
		q.include(Include.ALL);
		if (isProtocolParameter != null)
		{
			q.restrict(
				Restrictions.eq(
					Hql.property("protocolParameter"), 
					Expressions.parameter("isParameter", isProtocolParameter, Type.BOOLEAN)
				)
			);
		}
		q.setCacheResult(true);
		return q;
	}

	/**
		Get a query returning all annotation types that are needed to display
		inherited annotations for the given columns. The columns is a string with
		comma-separated values that define the visible columns in a table.
		Columns for inherited annotations are defined by 'ia'+&lt;id&gt;.
		@since 3.5
	*/
	public static ItemQuery<AnnotationType> getInheritedAnnotationColumns(String columns)
	{
		ItemQuery<AnnotationType> q = AnnotationType.getQuery(null);
		q.order(Orders.asc(Hql.property("name")));
		q.include(Include.ALL);
		q.restrict(Restrictions.in(Hql.property("id"), Expressions.parameter("ids")));
		q.restrict(Restrictions.eq(Hql.property("disableInheritance"), Expressions.bool(false)));
		List<Integer> ids = new ArrayList<Integer>();
		ids.add(0);
		if (columns != null)
		{
			for (String c : columns.split(","))
			{
				if (c.startsWith("ia")) ids.add(Integer.parseInt(c.substring(2)));
			}
		}
		q.setParameter("ids", ids, Type.INT);
		q.setCacheResult(true);
		return q;
	}
	
	/**
		Get a query returning all data file types that are needed to display
		file set members for the given columns. The columns is a string with
		comma-separated values that define the visible columns in a table.
		Columns for data file types annotations are defined by 'ft'+&lt;id&gt;.
		@since 3.5
	*/
	public static ItemQuery<DataFileType> getDataFileColumns(Item itemType, String columns)
	{
		ItemQuery<DataFileType> q = DataFileType.getQuery(null, null, itemType, null);
		q.order(Orders.asc(Hql.property("name")));
		q.include(Include.ALL);
		if (columns != null)
		{
			q.restrict(Restrictions.in(Hql.property("id"), Expressions.parameter("ids")));
			List<Integer> ids = new ArrayList<Integer>();
			ids.add(0);
			if (columns != null)
			{
				for (String c : columns.split(","))
				{
					if (c.startsWith("ft")) ids.add(Integer.parseInt(c.substring(2)));
				}
			}
			q.setParameter("ids", ids, Type.INT);
		}
		q.setCacheResult(true);
		return q;
	}
	
	/**
		Get a query that returns all subtypes for the given main item type.
		@param itemType The main item type
		@return An <code>ItemQuery</code> object
		@since 3.0
	*/
	public static ItemQuery<ItemSubtype> getSubtypesQuery(Item itemType)
	{
		ItemQuery<ItemSubtype> query = ItemSubtype.getQuery(itemType);
		query.order(Orders.asc(Hql.property("name")));
		query.setCacheResult(true);
		return query;
	}

	
	/**
		Get a query that returns all annotation types defined for a 
		{@link AnnotatableProxy} item.
		@param proxy The annotatable proxy
		@return An <code>ItemQuery</code> object
	*/
	public static ItemQuery<AnnotationType> getAnnotationTypesQuery(AnnotatableProxy proxy)
	{
		ItemQuery<AnnotationType> q = proxy.getAnnotationTypes();
		if (q == null)
		{
			// A query returning no annotations
			q = AnnotationType.getQuery(Item.SYSTEM);
		}
		q.order(Orders.asc(Hql.property("name")));
		q.include(Include.ALL);
		return q;
	}
	
	/**
		Get a query that returns all annotation types assigned as parameters to
		a specific protocol. The query will include all possible annotation types.
		@param protocol The protocol, or null
		@return An <code>ItemQuery</code> object, or null if the protocol is null
	*/
	public static ItemQuery<AnnotationType> getProtocolParametersQuery(Protocol protocol)
	{
		ItemQuery<AnnotationType> q = null;
		if (protocol != null)
		{
			q = protocol.getParameters();
			q.order(Orders.asc(Hql.property("name")));
			q.include(Include.ALL);
		}
		return q;
	}
	
	/**
		Get a query that returns all data file types for a specific
		platform/variant/subtype that can be used on a given item type. If an item
		is specified the query will also return data file types that
		already are present in the item's file set even if those 
		file types are not part of the platform / variant.
		@param itemType The item type
		@param item An optional item
		@param platform The platform to get file types for, or null
		@param variant The platform variant to get file types for, or null
		@param subtype The item subtype to get the file types for, or null
		@return An <code>ItemQuery</code> object
		@since 3.0
	*/
	public static ItemQuery<DataFileType> getDataFileTypes(Item itemType, FileStoreEnabled item, Platform platform, PlatformVariant variant, ItemSubtype subtype)
	{
		ItemQuery<DataFileType> q = DataFileType.getQuery();
		q.include(Include.ALL);
		// Restrict on itemType
		Restriction onItemType = Restrictions.eq(
				Hql.property("itemType"), 
				Expressions.integer(itemType.getValue())
			);
		
		// Restrict on platform
		Restriction onPlatform = null;
		if (platform != null)
		{
			onPlatform = Restrictions.eq(Hql.property("pft", "platform.id"), Hql.entity(platform));
			if (variant != null)
			{
				onPlatform = Restrictions.and(
					onPlatform, 
					Restrictions.or(
						Restrictions.eq(Hql.property("pft", "variant.id"), null),
						Restrictions.eq(Hql.property("pft", "variant.id"), Hql.entity(variant))
				));
			}
			else
			{
				onPlatform = Restrictions.and(
					onPlatform, 
					Restrictions.eq(Hql.property("pft", "variant.id"), null)
				);
			}
			q.join(Hql.leftJoin(null, "platforms", "pft", onPlatform, false));
		}
		
		// Restrict on item subtype
		Restriction onSubtype = null;
		if (subtype != null)
		{
			onSubtype = Restrictions.eq(Hql.property("its", "itemSubtype.id"), Hql.entity(subtype));
			q.join(Hql.leftJoin(null, "itemSubtypes", "its", onSubtype, false));
		}
		
		// Also include file types already present for item
		Restriction onItem = null;
		if (item != null && item.hasFileSet())
		{
			FileSet fs = item.getFileSet();
			onItem = Restrictions.eq(Hql.property("mbr", "fileSet.id"), Expressions.integer(fs.getId()));
			q.join(Hql.leftJoin(null, "fileSetMembers", "mbr", onItem, false));
		}
		
		// full = onItem || (onItemType && onPlatform && onSubtype)
		Restriction full = onItemType;
		if (onPlatform != null) full = Restrictions.and(full, onPlatform);
		if (onSubtype != null) full = Restrictions.and(full, onSubtype);
		if (onItem != null) full = Restrictions.or(full, onItem);
		
		q.restrict(full);
		q.order(Orders.asc(Hql.property("name")));
		q.setDistinct(true);
		return q;
	}
	
	public static net.sf.basedb.util.Enumeration<String, String> getReporterListsEnum(DbControl dc)
	{
		ItemQuery<ReporterList> query = ReporterList.getQuery();
		query.order(Orders.asc(Hql.property("name")));
		query.setCacheResult(true);
		query.include(Include.ALL);
		
		net.sf.basedb.util.Enumeration<String, String> lists = 
			new net.sf.basedb.util.Enumeration<String, String>();
		for (ReporterList list : query.list(dc))
		{
			lists.add(Integer.toString(list.getId()), HTML.encodeTags(list.getName()));
		}
		return lists;
	}

	/**
		Get available item lists as an Enumeration that is
		suitable to use in a column definition for a table.
		@param dc The DbControl to use
		@param memberType Only load lists with this member type, or null
			to load all lists
		@param include Include options
		@return The found lists as an enumeration, empty if no lists are
			found
		@since 3.5
	*/
	public static net.sf.basedb.util.Enumeration<String, String> 
		getItemListsEnum(DbControl dc, Item memberType, Collection<Include> include)
	{
		ItemQuery<ItemList> query = ItemList.getQuery();
		query.setIncludes(include == null ? Include.ALL : include);
		if (memberType != null)
		{
			query.restrict(
				Restrictions.eq(
					Hql.property("memberType"), 
					Expressions.integer(memberType.getValue())
				)
			);
		}
		query.order(Orders.asc(Hql.property("name")));
		query.setCacheResult(true);
		
		net.sf.basedb.util.Enumeration<String, String> lists = 
			new net.sf.basedb.util.Enumeration<String, String>();
		for (ItemList list : query.list(dc))
		{
			lists.add(Integer.toString(list.getId()), HTML.encodeTags(list.getName()));
		}
		return lists;
	}

	
	private static final Float defaultScale = 1.0f;
	/**
		Get the display scale setting. For normal font the scale is 1.0,
		for larger fonts the scale is higher and for smaller fonts the scale is
		lower. The scale is used for recalculating the size of popup windows
		depending on the font size setting.
	 	@param sc SessionControl of the active session 
	 	@return The current scale as a float.
	*/
	public static float getScale(SessionControl sc)
	{
		Float scale = defaultScale;
		if (sc != null) 
		{
			scale = (Float)sc.getSessionSetting("appearance.scale"); // In fractions
			if (scale == null)
			{
				int percent = Values.getInt(sc.getUserClientSetting("appearance.scale"), 100);  // In percent
				scale = new Float(percent / 100.0f);
				sc.setSessionSetting("appearance.scale", scale);
			}
		}
		return scale;
	}
	
	private static final Integer defaultMaxRecent = 4;
	/**
		Get the setting for the maximum number of recently used items 
		to save.
	 	@param sc SessionControl of the active session. 
	 	@return an int.
	*/
	public static int getMaxRecent(SessionControl sc)
	{
		Integer maxRecent = defaultMaxRecent;
		if (sc != null) 
		{
			maxRecent = (Integer)sc.getSessionSetting("appearance.recent");
			if (maxRecent == null)
			{
				maxRecent = Values.getInt(sc.getUserClientSetting("appearance.recent"), defaultMaxRecent);
				sc.setSessionSetting("appearance.recent", maxRecent);
			}
		}
		return maxRecent;
	}
	
	/**
		Get the cookie with the specified name.

		@param request The JSP <code>request</code> variable
		@param name The name of the cookie
		@return A <code>Cookie</code> object, or null if it is not found
	*/
	public static final Cookie getCookie(HttpServletRequest request, String name)
	{
		Cookie[] cookies = request.getCookies();
		for (int i = 0; i < cookies.length; i++)
		{
			if (cookies[i].getName().equals(name)) return cookies[i];
		}
		return null;
	}

	/**
		Get the value of the cookie with the specified name.

		@param request The JSP <code>request</code> variable
		@param name The name of the cookie
		@return The value of the cookie, or null if it is not found
	*/
	public static final String getCookieValue(HttpServletRequest request, String name)
	{
		Cookie cookie = getCookie(request, name);
		return cookie == null ? null : cookie.getValue();
	}

	/**
		Get the old item that is stored in the <code>SessionControl</code>:s 
		session setting <code>itemType.item</code>. If the setting is null
		compare the <code>newItem</code>:s version with the <code>oldVersion</code>,
		and if the match the new item is returned as the old item. If the
		versions doesn't match, an <code>ItemModifiedException</code> is thrown.
	 	@param <T> The kind of item that extends from BasicItem. 
		@param sc The <code>SessionControl</code> which keeps the old item
			in the session settings
		@param newItem The new item
		@param oldVersion The version of the old item
		@return The old item if it is found in the session setting, or the
			new item if the old version is the same
		@throws ItemModifiedException If the old item wasn't found in the session
			setting and the old version is different from the new item's version
	*/
	@SuppressWarnings({"unchecked"})
	public static final <T extends BasicItem> T getOldItem(SessionControl sc, T newItem, int oldVersion)
		throws ItemModifiedException
	{
		T oldItem = (T)sc.getSessionSetting(newItem.getType().name()+".item");
		if (oldItem == null)
		{
			if (oldVersion != newItem.getVersion())
			{
				throw new ItemModifiedException(newItem.toString());
			}
			oldItem = newItem;
		}
		return oldItem;
	}
	
	/**
		Update annotations on an item taking concurrent modifications into account. 
		The old item is the object that was used to present the annotations for the 
		user in the web form. The new item is the object that is loaded in the submit 
		request. The new item is used to update the annotations. While
		updating the annotations the version of each annotation on the old an new 
		item is compared. If they don't match an <code>ItemModifiedException</code>
		is thrown. We do not have to compare with the values in the request, since
		only modified annotations are submitted.
 
		@param dc A <code>DbControl</code> object used to read information from the
			database
		@param oldItem The old annotatable item, or null if no old item exists
		@param newItem The new item, or null if only an old item exists
		@param request The request object containing the modified annotations
		@throws ItemModifiedException If the new and old item don't match
	 	@throws BaseException If the update fails in some way.
	*/
	public static final void updateAnnotations(DbControl dc, Annotatable oldItem, Annotatable newItem, HttpServletRequest request)
		throws ItemModifiedException, BaseException
	{
		// Modified annotations
		String modified = Values.getStringOrNull(request.getParameter("modifiedAnnotations"));
		if (newItem == null) newItem = oldItem;
		if (modified != null)
		{
			JSONArray jsonModified = null;
			try
			{
				jsonModified = (JSONArray)new JSONParser().parse(modified);
			}
			catch (ParseException ex)
			{
				throw new BaseException(ex);
			}
			
			//String[] modifiedAnnotations = modified.split(",");
			//AnnotationSet oldAs = oldItem != null && oldItem.isAnnotated() ? 
			//	oldItem.getAnnotationSet() : null;
			AnnotationSet newAs = newItem.getAnnotationSet();
			if (oldItem == newItem)
			{
				// The items were not loaded in separate sessions, the protocol has to be 
				// reattached to the current session
				Protocol p = newItem.getProtocol();
				if (p != null) dc.reattachItem(p, false);
			}
			
			for (int entryNo = 0; entryNo < jsonModified.size(); entryNo++)
			{
				JSONObject jsonA = (JSONObject)jsonModified.get(entryNo);
				Number annotationId = (Number)jsonA.get("annotationId");
				Number inheritedId = (Number)jsonA.get("inheritedId");
				Number annotationTypeId = (Number)jsonA.get("annotationTypeId");
				JSONArray jsonValues = (JSONArray)jsonA.get("values");
				Annotation.Source source = Annotation.Source.valueOf((String)jsonA.get("source"));
				String modifyType = (String)jsonA.get("modified");
				
				AnnotationType at = AnnotationType.getById(dc, annotationTypeId.intValue());
				Annotation a = null;
				if (annotationId != null)
				{
					a = Annotation.getById(dc, annotationId.intValue());
					
					if ("DELETE".equals(modifyType))
					{
						if (source == Annotation.Source.PRIMARY)
						{
							newAs.removeAnnotation(at);
						}
						else
						{
							newAs.removeInheritedAnnotation(a);
						}
						continue;
					}
					else if ("SYNCED".equals(modifyType) && source == Annotation.Source.CLONED)
					{
						// Sync a cloned annotation
						a.resyncCloned();
						continue;
					}
					else if (a.getSource() != source)
					{
						// Convert between inherited<->cloned
						a = newAs.inheritAnnotation(a.getInheritedFrom(), source == Annotation.Source.CLONED);
						if ("CONVERTED".equals(modifyType)) continue; // Only converted between types
					}
				}
				else if (inheritedId != null)
				{
					if (!"DELETE".equals(modifyType))
					{
						Annotation inherited = Annotation.getById(dc, inheritedId.intValue());
						a = newAs.inheritAnnotation(inherited, source == Annotation.Source.CLONED);
					}
				}
				else if (jsonValues != null && jsonValues.size() > 0)
				{
					if (!"DELETE".equals(modifyType))
					{
						a = newAs.getAnnotation(at);
					}
				}
				
				if (a == null) continue;
				
				if (jsonValues != null && jsonValues.size() > 0)
				{
					Type valueType = at.getValueType();
					Number unitId = (Number)jsonA.get("unitId");
					Unit unit = unitId == null ? null : Unit.getById(dc, unitId.intValue());
					
					Object[] oValues = new Object[jsonValues.size()];
					Formatter<?> formatter = FormatterFactory.getTypeFormatter(dc.getSessionControl(), valueType);
					for (int j = 0; j < oValues.length; ++j)
					{
						oValues[j] = formatter.parseString((String)jsonValues.get(j));
					}
					a.setValuesIfDifferent(Arrays.asList(oValues), unit);
				}
			}
		}
		
	}
	
	/**
		Update files that are associated with a {@link FileStoreEnabled}
		item. The <code>request</code> parameter is checked for parameters
		with a name like <code>datafile.id</code> where <code>id</code> is
		the numeric id of a {@link DataFileType}. The parameter value is the
		path of the file to associated with the data file type, or null
		to remove an existing association.
		
		@param dc A DbControl to access the database
		@param item The item the files should be associated with
		@param request The request object containing the selected files
		@param validate If {@link FileSet#validate(DbControl)}
			should be called or not
	 	@param cc Save the current context information. Null is allowed.
	 	@param maxRecent The maximum of recent items to remember. 
		@since 3.0
	*/
	public static void updateFiles(DbControl dc, FileStoreEnabled item, 
		HttpServletRequest request, boolean validate,
		ItemContext cc, int maxRecent)
	{
		Enumeration<String> names = request.getParameterNames();
		FileSet fileSet = null;
		while (names.hasMoreElements())
		{
			String param = names.nextElement();
			if (param.startsWith("setfile."))
			{
				int dataFileTypeId = Values.getInt(param.substring(8));
				int fileId = Values.getInt(request.getParameter(param));
				DataFileType dft = DataFileType.getById(dc, dataFileTypeId);
				File file = fileId == 0 ? null : File.getById(dc, fileId);
				if (fileSet == null) fileSet = item.getFileSet();
				fileSet.setMember(file, dft);
				if (cc != null && file != null)
				{
					cc.setRecent(file, dft.getExternalId(), maxRecent);
				}
			}
			else if (param.startsWith("addfile."))
			{
				int dataFileTypeId = Values.getInt(param.substring(8));
				Integer[] fileId = Values.getInt(request.getParameterValues(param));
				DataFileType dft = DataFileType.getById(dc, dataFileTypeId);
				if (fileSet == null) fileSet = item.getFileSet();
				for (Integer id : fileId)
				{
					File file = File.getById(dc, id);
					fileSet.addMember(file, dft);
					if (cc != null)
					{
						cc.setRecent(file, dft.getExternalId(), maxRecent);
					}
				}
			}
			else if (param.startsWith("removefile."))
			{
				int dataFileTypeId = Values.getInt(param.substring(11));
				Integer[] fileId = Values.getInt(request.getParameterValues(param));
				DataFileType dft = DataFileType.getById(dc, dataFileTypeId);
				if (fileSet == null) fileSet = item.getFileSet();
				for (Integer id : fileId)
				{
					File file = File.getById(dc, id);
					fileSet.removeMember(file, dft);
				}
			}
		}
		if (validate)
		{
			if (fileSet == null) fileSet = item.getFileSet();
			fileSet.validate(dc);
		}
	}
	
	/**
		Encode the name of the nameable item using the {@link HTML#encodeTags(String)}
		method. If the item parameter is null, <code>- none -</code> or
		<code>- denied -</code> is returned, depending on if the denied
		parameter is true or false. 
		
		@param item The nameable item or null
		@param denied TRUE if the item is null because the current use
			doesn't have read permission

	*/
	public static String getEncodedName(Nameable item, boolean denied)
	{
		String name = "<i>- none -</i>";
		if (item != null)
		{
			name = HTML.encodeTags(item.getName());
		}
		else if (denied)
		{ 
			name = "<i>- denied -</i>";
		}
		return name;
	}
	

	public static String getLinkedName(String ID, Nameable item, boolean denied, boolean enableEditLink)
	{
		String link = "<i>- none -</i>";
		if (item != null)
		{
			link = getLink(ID, HTML.encodeTags(item.getName()), item.getType(), item.getId(), enableEditLink && item.hasPermission(Permission.WRITE));
		}
		else if (denied)
		{ 
			link = "<i>- denied -</i>";
		}
		return link;
	}
	
	public static String getLink(String ID, String name, Item itemType, int itemId, boolean enableEditLink)
	{
		return getLink(ID, name, itemType, itemId, enableEditLink, 
				"View this " + itemType.toString().toLowerCase());
	}

	/**
		@since 2.9
	*/
	public static String getLink(String ID, String name, Item itemType, int itemId, boolean enableEditLink, String tooltip)
	{
		StringBuilder sb = new StringBuilder();
		/*
			<span class="link auto-init" data-auto-init="item-link" 
				data-item-type="itemType" data-item-id="itemId"
				[data-no-edit="1"]
				title="View this item (use CTRL, ALT or SHIFT to edit">name</span>
		*/
		sb.append("<span class=\"link auto-init\" data-auto-init=\"item-link\"");
		sb.append(" data-item-type=\"").append(itemType.name()).append("\"");
		sb.append(" data-item-id=\"").append(itemId).append("\"");
		if (!enableEditLink) sb.append(" data-no-edit=\"1\"");
		if (tooltip != null)
		{
			sb.append(" title=\"").append(tooltip);
			if (enableEditLink) sb.append(" (use CTRL, ALT or SHIFT to edit)");
			sb.append("\"");
		}
		sb.append(">").append(name).append("</span>");
		return sb.toString();
	}

	
	public static String getLinkedFile(String ID, File file, boolean denied, boolean enableEditLink, boolean enableViewDownload, String root)
	{
		String link = "<i>- none -</i>";
		if (file != null)
		{
			link = getLink(ID, HTML.encodeTags(file.getName()), file.getType(), file.getId(), enableEditLink && file.hasPermission(Permission.WRITE));
			if (enableViewDownload)
			{
				link += getFileLinks(ID, file, root);
			}
		}
		else if (denied)
		{ 
			link = "<i>- denied -</i>";
		}
		return link;

	}
	
	/**
		Get view/download links for a file.
		@param ID The session ID
		@param file The file item
		@param root Relative path to root directory of the web server
		@return HTML code for the links
		@since 2.5
	*/
	public static String getFileLinks(String ID, File file, String root)
	{
		String link = "";
		// [view | download] - or - [offline] - or - [secondary]
		Location l = file.getLocation();
		if (l == Location.OFFLINE)
		{
			link += " [offline]";
		}
		else if (l == Location.SECONDARY)
		{
			link += " [secondary]";
		}
		else if (l.isDownloadable())
		{
			String viewIcon = "view";
			String downloadIcon = "download";
			String separator = " | ";
			String bytes = Values.formatBytes(file.getSize(), 1);
			if (root != null)
			{
				viewIcon = "<img src=\"" + root + "images/view.png" + "\" alt=\"view\">";
				downloadIcon = "<img src=\"" + root + "images/download.png" + "\" alt=\"download\">";
				separator = "&nbsp;";
			}
			link += "&nbsp;<span style=\"white-space: nowrap;\">[<span class=\"link auto-init\" data-auto-init=\"view-file\" data-file-id=\"" + file.getId() + "\"" + 
				" title=\"View the contents of this file (" + bytes + ")\">" + viewIcon + "</span>" + 
				separator + 
				"<span class=\"link auto-init\" data-auto-init=\"download-file\" data-file-id=\"" + file.getId() + "\"" +
				" title=\"Download the contents of this file (" + bytes + ")\">" + downloadIcon + "</span>]</span>";
		}
		return link;
	}
	
	public static <T extends BasicItem> T getFirstMatching(DbControl dc, ItemQuery<T> query, String property, PropertyFilter filter)
	{
		if (filter == null) return null;
		query.include(Include.MINE, Include.SHARED, Include.IN_PROJECT, Include.OTHERS);

		Expression parameter = filter.getValue() == null ? null : Expressions.parameter("context", filter.getValueAsObject());
		Restriction r = filter.getOperator().getRestriction(Hql.property(property), parameter);
		query.restrict(r);
		ItemResultIterator<T> result = query.iterate(dc);
		if (result.hasNext())
		{
			return result.next();
		}
		else
		{
			return null;
		}
	}
	
	public static Color[] getConfiguredColors(SessionControl sc)
	{
		Color[] colors = new Color[3];
		colors[0] = ColorGenerator.fromHex(Values.getString(sc.getUserClientSetting("ratiocolor.min"), "0000FF"));
		colors[1] = ColorGenerator.fromHex(Values.getString(sc.getUserClientSetting("ratiocolor.mid"), "FFFFFF"));
		colors[2] = ColorGenerator.fromHex(Values.getString(sc.getUserClientSetting("ratiocolor.max"), "FFFF00"));
		return colors;
	}
	
	public static ColorGenerator getColorGenerator(SessionControl sc, Coloring coloring)
	{
		Color[] colors = getConfiguredColors(sc);
		ColorGenerator colorGen = new ColorGenerator(colors[0], colors[1], colors[2], 
			coloring.getMinValue(), coloring.getMidValue(), coloring.getMaxValue(), 
				coloring.isLogarithmic());
		return colorGen;
	}
	
}

