/*
	$Id: Application.java 7262 2016-12-12 12:42:30Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.SchemaVersionData;
import net.sf.basedb.util.EmailUtil;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.RegexpFileFilter;
import net.sf.basedb.util.StaticCache;
import net.sf.basedb.util.extensions.Registry;
import net.sf.basedb.util.extensions.manager.ExtensionsManager;
import net.sf.basedb.util.extensions.manager.ProcessResults;
import net.sf.basedb.util.extensions.manager.filter.InstalledFilter;
import net.sf.basedb.util.extensions.manager.filter.WasProcessedFilter;
import net.sf.basedb.util.extensions.manager.processor.MarkAsProcessedProcessor;
import net.sf.basedb.util.extensions.manager.processor.RegisterExtensionsProcessor;
import net.sf.basedb.util.extensions.xml.ExtensionPointFilter;
import net.sf.basedb.util.extensions.xml.XmlLoader;
import net.sf.basedb.util.timer.Scheduler;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.Collections;
import java.util.Iterator;
import java.util.Random;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.net.InetAddress;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.SQLException;

import org.hibernate.jdbc.ReturningWork;


/**
	This class represents the BASE application and is the starting point for
	all client applications. This class is used to start and stop the
	application, create {@link SessionControl} objects used for
	user authentication, etc.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2016-12-12 13:42:30 +0100 (må, 12 dec 2016) $
*/
public final class Application
{

	/**
		Log core events.
	*/
	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(Application.class);

	/**
		If BASE is running or not.
	*/
	private static boolean isRunning = false;

	/**
		The name of the external authenication class, if any.
	*/
	//private static String authenticationDriver;
	
	/**
		The name of the secondary storage controller class, if any.
	*/
	private static String secondaryStorageDriver;

	/**
		A cache of SessionControl objects.
	*/
	private static Map<String,SessionControl> sessionCache;

	/**
		Number of minutes a SessionControl object can be unused before
		it is removed from the cache.
	*/
	private static int sessionCacheTimeout;

	/**
		The core timer which schedules events.
	*/
	private static volatile Timer timer;

	/**
		An interface for adding timed task to the core timer.
	*/
	private static volatile Scheduler scheduler;
	
	/**
		Number of minutes permission information can be kept in
		memory before it must be reloaded.
	*/
	private static int permissionTimeout;

	/**
		The directory where uploaded files are stored.
	*/
	private static java.io.File userFilesDirectory;
	
	/**
		The one and only place where plug-ins can be installed.
	*/
	private static java.io.File pluginsDirectory;
	
	/**
		Cache for static data.
	*/
	private static StaticCache staticCache;

	/**
		The path to the file containing predefined queries.
	*/
	private static String queryFile;

	/**
		The path to the file containing extended properties.
	*/
	private static String extendedPropertiesFile;

	/**
		The path to the file containing raw data types.
	*/
	private static String rawDataTypesFile;

	/**
		The name of the dynamic database.
	*/
	private static String dynamicCatalog;
	
	/**
		The name of the dynamic schema.
	*/
	private static String dynamicSchema;
	
	/**
		Allow automatic unload of plug-in JAR files or not.
	*/
	private static boolean autoUnloadPlugins;
	
	/**
		The host name of the server running BASE.
	*/
	private static String hostName;
	
	/**
		The internal job queue reference if it is enabled.
	*/
	private static InternalJobQueue internalJobQueue;
		
	/**
	 	The name and version of the database server
	*/
	private static String databaseVersionNumber;
	
	/**
		If auto-compression has been enabled or not.
	*/
	private static boolean noAutoCompression;
	
	/**
		Manager for the extensions system.
	*/
	private static ExtensionsManager xtManager;
	
	/**
		Get the major version.
		@since 3.0 (was returning a String in 2.x)
	*/
	public static int getMajorVersion()
	{
		return Version.getMajor();
	}
	
	/**
		Get the minor version.
		@since 3.0 (was returning a String in 2.x)
	*/
	public static int getMinorVersion()
	{
		return Version.getMinor();
	}

	
	/**
		Get the maintenance version.
		@since 3.0 (was returning a String in 2.x)
	*/
	public static int getMaintenanceVersion()
	{
		return Version.getMaintenance();
	}
	
	/**
		Get the build number.
	*/
	public static int getBuild()
	{
		return Version.getBuild();
	}

	/**
		Get the version suffix (eg. "-dev").
		@since 3.0
	*/
	public String getVersionSuffix()
	{
		return Version.getSuffix();
	}
	
	/**
		Get the BASE version string. It is created by combining
		the major, minor, maintenance, build and schema versions.
	*/
	public static String getVersionString()
	{
		return "BASE " + Version.getMajor() + "." + Version.getMinor() +
			"." + Version.getMaintenance() + Version.getSuffix() + " (build #" + Version.getBuild() +
			"; schema #" + Application.getSchemaVersion()+")";
	}
	
	private static int schemaVersion = -2;
	/**
		Get the current database schema version number.
		@return The schema version number, or -2 if a connection to the database couldn't be established, 
		or -1 if schemaVersion could not be loaded
	*/
	public static int getSchemaVersion()
	{
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		if (schemaVersion > 0) return schemaVersion;
		
		synchronized (Application.class)
		{
			schemaVersion = -2;
			try
			{
				session = HibernateUtil.newSession();
				schemaVersion++;
				tx = HibernateUtil.newTransaction(session);
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_SCHEMA_VERSION_FOR_APP");
				query.setString("appId", SchemaVersionData.BASE_APP_ID);
				schemaVersion = HibernateUtil.loadData(SchemaVersionData.class, query).getSchemaVersion();
				HibernateUtil.commit(tx);
			}
			catch (Throwable t)
			{
				if (tx != null) HibernateUtil.rollback(tx); 
				if (Install.NEW_SCHEMA_VERSION != 118)
				{
					// Exception is expected when updating to 118 (BASE 3.4) so we don't warn about this
					log.warn("getSchemaVersion", t);
				}
			}
			finally
			{
				if (session != null) HibernateUtil.close(session);
			}			
		}
		return schemaVersion;
	}
	
	
	/**
		Get the BASE core logger.
	*/
	static org.slf4j.Logger getLogger()
	{
		return log;
	}
	
	/**
		Get the host name of the server running BASE.
	*/
	public static String getHostName()
	{
		if (hostName == null)
		{
			try
			{
				InetAddress host =  InetAddress.getLocalHost();
				hostName = host == null ? "unknown" : host.getCanonicalHostName();
			}
			catch (Throwable ex)
			{
				log.error("Exception finding server name", ex);
				hostName = "unknown";
			}
		}
		return hostName;
	}
	
	/**
	 	Gets the version of the database server
	 	@return A String with the name and version number of the database server
	 	@since 2.3.1
	*/
	public static String getDatabaseVersionString()
	{
		org.hibernate.Session session = null;
		if (databaseVersionNumber == null)
		{
			try
			{
				session = HibernateUtil.newSession();
				databaseVersionNumber = 
					HibernateUtil.doReturningWork(session, 
						new ReturningWork<String>()
						{
							@Override
							public String execute(Connection connection) 
								throws SQLException
							{
								DatabaseMetaData meta = connection.getMetaData();
								return meta.getDatabaseProductName() + " " + meta.getDatabaseProductVersion();
							}
						});
			}
			catch (Throwable th)
			{
				databaseVersionNumber = "-N/A-";
			}
			finally
			{
				if (session != null) HibernateUtil.close(session);
			}
		}
		return databaseVersionNumber;
	}
	
	/**
		Check if BASE is running.
		@return TRUE if BASE is running, FALSE otherwise
		@see #start()
		@see #stop()
	*/
	public static boolean isRunning()
	{
		return isRunning;
	}

	/**
		Starts the BASE application. This method will read settings 
		from the configurations files and initialise all internal
		structures. Changes made to configuration files requires a
		restart to become effective. If BASE is already started this method
		does nothing.
		@throws BaseException If BASE cannot be started
		@see #stop()
		@see #isRunning()
	*/
	public static synchronized void start()
		throws BaseException
	{
		start(false, true, null);
	}
	
	/**
		Starts the BASE application. This method will read settings 
		from the configurations files except if the internal job queue
		should be enabled or not. 
		@param useInternalJobQueue If the internal job queue should be 
			enabled or not
		@throws BaseException If BASE cannot be started
		@see #start()
		@see #stop()
		@see #isRunning()
	*/
	public static synchronized void start(boolean useInternalJobQueue)
	{
		start(false, true, useInternalJobQueue);
	}

	/**
		Starts the BASE application.
		@param installation TRUE if we are starting the application for
			installation, FALSE for a normal start
		@throws BaseException If BASE cannot be started
		@see #start()
	*/
	static synchronized void start(boolean installation, boolean verifySchemaVersion, Boolean useInternalJobQueue)
		throws BaseException
	{
		if (isRunning()) return;
		log.info("Starting BASE");

		try
		{
			// Read settings from the configuration file
			Version.init();
			log.info("Version: " + Version.getVersion());
			
			Config.init();
			//authenticationDriver = Config.getString("auth.driver");
			//log.info("auth.driver = " + authenticationDriver);
			
			secondaryStorageDriver = Config.getString("secondary.storage.driver");
			log.info("secondary.storage.driver = " + secondaryStorageDriver);
			
			sessionCacheTimeout = Config.getInt("cache.timeout", 20); // minutes
			if (sessionCacheTimeout <= 0) sessionCacheTimeout = 1;
			log.info("cache.timeout = " + sessionCacheTimeout + " minutes");
			
			permissionTimeout = Config.getInt("permission.timeout", 10); // minutes
			log.info("permission.timeout = " + permissionTimeout + " minutes");
			
			userFilesDirectory = new java.io.File(Config.getRequiredString("userfiles", "No setting for 'userfiles' in base.config."));
			log.info("userfiles = " + userFilesDirectory);
			if (!userFilesDirectory.exists() || !userFilesDirectory.isDirectory()) 
			{
				throw new ConfigurationException("Directory 'userfiles' in base.config doesn't exists or is not a directory: " + userFilesDirectory);
			}

			pluginsDirectory = new java.io.File(Config.getRequiredString("plugins.dir", "No setting for 'plugins.dir' in base.config"));
			log.info("plugins.dir = " + pluginsDirectory);
			if (!pluginsDirectory.exists() || !pluginsDirectory.isDirectory()) 
			{
				throw new ConfigurationException("Plug-ins directory '" + pluginsDirectory + "' doesn't exists or is not a directory.");
			}
			
			log.info("db.dialect = " + Config.getString("db.dialect"));
			log.info("db.driver = " + Config.getString("db.driver"));
			log.info("db.url = " + Config.getString("db.url"));
			log.info("db.username = " + Config.getString("db.username"));
			log.debug("db.password = " + Config.getString("db.password"));
			
			queryFile = Config.getString("db.queries");
			log.info("db.queries = " + queryFile);
			
			extendedPropertiesFile = Config.getString("db.extended-properties", "/extended-properties.xml");
			log.info("db.extended-properties = " + extendedPropertiesFile);
			
			rawDataTypesFile = Config.getString("db.raw-data-types", "/raw-data-types.xml");
			log.info("db.raw-data-types = " + rawDataTypesFile);
			
			dynamicCatalog = Config.getString("db.dynamic.catalog");
			log.info("db.dynamic.catalog = " + dynamicCatalog);
			
			dynamicSchema = Config.getString("db.dynamic.schema");
			log.info("db.dynamic.schema = " + dynamicSchema);
			log.info("db.batch-size = " + Config.getString("db.batch-size"));
	
			autoUnloadPlugins = Config.getBoolean("plugins.autounload");
			log.info("plugins.autounload = " + autoUnloadPlugins);
			
			noAutoCompression = Config.getBoolean("autocompress.disable");
			log.info("autocompress.disable = " + noAutoCompression);
			
			// Create a cache for SessionControl objects
			sessionCache = Collections.synchronizedMap(new HashMap<String,SessionControl>());
			
			// Initialise other utility classes
			staticCache = new StaticCache(new java.io.File(userFilesDirectory, "static.cache"));
			staticCache.setDisabled(true);
			QueryExecutor.init();
			HibernateUtil.init1();
			ExtendedProperties.init();
			RawDataTypes.init();
			RawDataUtil.init();
			HibernateUtil.init2();
			QueryRuntimeFilterFactory.init();
			PredefinedQuery.init();
			BatchUtil.init();
			Metadata.init();
			EmailUtil.init();
	
			if (verifySchemaVersion)
			{
				int schemaVersion = getSchemaVersion();
				if (schemaVersion != Install.NEW_SCHEMA_VERSION)
				{
					String msg = null;
					if (schemaVersion == -2)
					{
						msg = "Couldn't connect to the database; Please check that the base.config file has the correct settings " +
								"and that the database server is running.";
					}
					else if(schemaVersion == -1)
					{
						msg = "Couldn't find any schema version in the database; Please check that BASE has been installed properly";
					}
					else
					{	
						msg = "Schema version mismatch: database schema version=" + 
							schemaVersion + "; expected=" + Install.NEW_SCHEMA_VERSION;
						
						if (schemaVersion < Install.NEW_SCHEMA_VERSION)
						{
							msg += "; Please update the database with the 'updatedb.sh' script before starting BASE.";
						}
						else
						{
							msg += "; Please update the BASE code before starting.";
						}
					}
					log.error(msg);
					throw new BaseException(msg);
				}
			}
			
			if (!installation)
			{
				SystemItems.init();
				Keyring.init();
				if (verifySchemaVersion) HibernateUtil.testTransactions();
				RawDataTypes.initPlatforms(null);
	
				// Initialize extensions system 
				boolean extensionsAreDisabled = Config.getBoolean("extensions.disabled");
				Registry xtRegistry = new Registry();
				java.io.File xtSettings = new java.io.File(userFilesDirectory, "extension-settings.xml");
				xtManager = new ExtensionsManager(xtRegistry, xtSettings);

				// Add core extension points and extensions
				xtManager.addURI(Application.class.getResource("/core-extensions.xml").toURI());
				// Add core plug-ins
				xtManager.addURI(Application.class.getResource("/core-plugins.xml").toURI());
				// Add plugins directory to the manager (unless disabled)
				if (!extensionsAreDisabled) xtManager.addDirectory(pluginsDirectory);
				
				// Register and load core extensions only
				XmlLoader loader = new XmlLoader();
				loader.setFilter(new ExtensionPointFilter("net\\.sf\\.basedb\\.(core|util)\\..*"));
				ProcessResults results = new ProcessResults();
				xtManager.processFiles(new RegisterExtensionsProcessor(loader, results), new InstalledFilter(xtManager.getSettings()));
				
				// Finalize registration
				xtManager.processFiles(new MarkAsProcessedProcessor(), new WasProcessedFilter(results));
				xtManager.setLastProcessResults(results);
				
				// Adding a task that cleans the session control cache at regular intervals
				long milliSeconds = 60000L * sessionCacheTimeout;  // minutes --> milliseconds
				getScheduler().schedule(new SessionControlCacheCleaner(), milliSeconds, milliSeconds, false);
				
				// Adding a task that cleans the database at regular intervals
				long cleanupInterval = 3600L * 1000L * Config.getInt("db.cleanup.interval", 24); // Hours --> milliseconds
				if (cleanupInterval > 0)
				{
					getScheduler().schedule(new DbCleaner(), cleanupInterval, cleanupInterval, false);
				}
				
				// Adding a task that cleans the static cache at regular intervals
				staticCache.setDisabled(Config.getBoolean("cache.static.disabled"));
				if (!staticCache.isDisabled())
				{
					long maxFileAge = 24L * 3600L * 1000L * Config.getLong("cache.static.max-age", 30); // Days-->milliseconds
					long checkInterval = 12L * 3600L * 1000L; // Once every twelve hours
					getScheduler().schedule(
						staticCache.cleanUpTask(staticCache.olderThan(maxFileAge)), 60000L, checkInterval, false);
				}
				
				if (useInternalJobQueue == null) useInternalJobQueue = Config.getBoolean("jobqueue.internal.enabled");
				log.info("jobqueue.internal.enabled = " + useInternalJobQueue);
				if (useInternalJobQueue)
				{
					long checkInterval = 1000L * Config.getInt("jobqueue.internal.checkinterval", 30); // Seconds --> milliseconds
					internalJobQueue = new InternalJobQueue();
					getScheduler().schedule(internalJobQueue, checkInterval, checkInterval, false);
				}
				
				if (hasSecondaryStorage())
				{
					long checkInterval = Config.getInt("secondary.storage.interval", 0);
					String timePoints = Config.getString("secondary.storage.time");
					TimerTask sss = new SecondaryStorageControllerTask(getSecondaryStorageController());
					Scheduler s = getScheduler();
					if (timePoints != null)
					{
						Calendar c = new GregorianCalendar();
						long EVERY_DAY = 24l * 3600l * 1000l;
						Pattern p = Pattern.compile("(\\d\\d):(\\d\\d)");
						for (String timePoint : timePoints.split(","))
						{
							Matcher m = p.matcher(timePoint);
							if (m.matches())
							{
								int hour = Integer.parseInt(m.group(1));
								int minute = Integer.parseInt(m.group(2));
								log.info("Scheduling secondary storage driver at " + timePoint + " every day");
								c.set(Calendar.HOUR_OF_DAY, hour);
								c.set(Calendar.MINUTE, minute);
								c.set(Calendar.SECOND, 0);
								s.scheduleAtFixedRate(sss, c.getTime(), EVERY_DAY, false);
							}
							else
							{
								log.warn("Time-point " + timePoint + " is not a valid value; must match hh:mm");
							}
						}
					}
					else if (checkInterval > 0)
					{
						log.info("Scheduling secondary storage driver every " + checkInterval + " seconds");
						s.schedule(sss, 1000 * checkInterval, 1000 * checkInterval, false);
					}
				}
				
				// Add shutdown hook so we can clean up if System.exit is called.
				Runtime.getRuntime().addShutdownHook(
					new Thread(
						new Runnable()
						{
							@Override
							public void run()
							{
								if (!Application.isRunning()) return;
								try
								{
									log.info("System is shutting down. Cleaning up as much as possible");
								}
								catch (Throwable t)
								{}
								Application.stop();
							}
						}
					)
				);
			}
			log.info("BASE is up and running");
			isRunning = true;
		}
		catch (Throwable t)
		{
			log.error("Exception when starting...", t);
			isRunning = true; // Or the stop() method won't do anything
			stop();
			throw new BaseException(t);
		}
	}

	/**
		Stops the BASE application. If BASE is not running this
		method does nothing.
		@throws BaseException If BASE cannot be stopped
		@see #start()
		@see #isRunning()
	*/
	public static synchronized void stop()
		throws BaseException
	{
		if (!isRunning()) return;

		log.info("Stopping BASE");
		// Stop timer and clean the cache of SessionControl objects
		if (timer != null) timer.cancel();
		if (internalJobQueue != null) internalJobQueue.close();
		internalJobQueue = null;
		timer = null;
		scheduler = null;
		cleanSessionControlCache(true);
		sessionCache = null;
		Keyring.unload();
		SystemItems.unload();
		Metadata.unload();
		EmailUtil.unload();
		BatchUtil.unload();
		PredefinedQuery.unload();
		QueryRuntimeFilterFactory.unload();
		HibernateUtil.unload();
		RawDataUtil.unload();
		RawDataTypes.unload();
		QueryExecutor.unload();
		ExtendedProperties.unload();
		Config.unload();

		xtManager = null;
		secondaryStorageDriver = null;
		userFilesDirectory = null;
		pluginsDirectory = null;
		extendedPropertiesFile = null;
		rawDataTypesFile = null;
		dynamicCatalog = null;
		dynamicSchema = null;
		hostName = null;
		databaseVersionNumber = null;
		
		schemaVersion = -2;
		log.info("BASE has been stopped");
		
		isRunning = false;
	}

	/**
		Check if BASE is using internal or external authentication.
		@return Always true
		@see #getAuthenticator()
		@deprecated In 3.3, no replacement
	*/
	@Deprecated
	public static boolean isUsingInternalAuthentication()
	{
		return true;
	}

	/**
		Get an object of the class that is used for external
		authentication. The authenticator is initialised before it is returned.
		If internal authentication is used, this method returns null.
		@return Always null
		@throws BaseException If the authenticator class could not be loaded
		@see #isUsingInternalAuthentication()
		@deprecated In 3.3, no replacement
	*/
	@Deprecated
	static net.sf.basedb.core.authentication.Authenticator getAuthenticator()
		throws BaseException
	{
		return null;
	}


	public static ExtensionsManager getExtensionsManager()
	{
		return xtManager;
	}
	
	/**
		Get the maximum number of minutes permission information for
		a logged in user is kept in memory before beeing reloaded.
		@return An integer with the number of minutes
	*/
	public static int permissionTimeout()
	{
		return permissionTimeout;
	}
	
	/**
	 	Gets the maximum number of minutes before an active session 
	 	is logged out.
	  	@return An integer with number of minutes
	  	@since 2.5
	 */	
	public static int sessionCacheTimeout()
	{
		return sessionCacheTimeout;
	}
	
	/**
		Checks if the server has configured a secondary storage or not. 
		
		@return TRUE if a secondary storage is available, FALSE otherwise
		@see File#setLocation(Location)
	*/
	public static boolean hasSecondaryStorage()
	{
		return secondaryStorageDriver != null;
	}

	/**
		Get an object of the class that is used for seconary storage handling.
		The controller is initialised before it is returned.
		If no controller has been specified, this method returns null.
		@return A <code>SecondaryStorageController</code> object, or null if
		@throws BaseException If the controller class could not be loaded
		@see #hasSecondaryStorage()
	*/
	static SecondaryStorageController getSecondaryStorageController()
	{
		if (!hasSecondaryStorage()) return null;
		SecondaryStorageController ssc = null;
		try
		{
			ssc = (SecondaryStorageController)Class.forName(secondaryStorageDriver).newInstance();
			ssc.init(Config.getString("secondary.storage.init"));
		}
		catch (Throwable t)
		{
			throw new BaseException("The storage controller '"+secondaryStorageDriver+
				"' could not be loaded. The error message is: "+t.getMessage());
		}
		return ssc;
	}
	
	/**
		Get the path to the directory where uploaded
		files are stored.
	*/
	public static java.io.File getUserFilesDirectory()
	{
		return userFilesDirectory;
	}

	/**
		Get the static cache mananger.
		@since 2.11
	*/
	public static StaticCache getStaticCache()
	{
		return staticCache;
	}
	
	/**
		Get the path to the XML file containing predefined
		queries for SQL and HQL.
		@see PredefinedQuery
	*/
	static String getQueryFile()
	{
		return queryFile;
	}

	/**
		Get the path to the XML file containing extended properties.
		@see ExtendedProperties
	*/
	static List<String> getExtendedPropertyFiles()
	{
		List<String> files = new ArrayList<String>();
		// Main extended properties file
		files.add(extendedPropertiesFile);  
		// Scan extended-properties directory
		String configDir = Config.getConfigDirectory().getAbsolutePath();
		java.io.File propConfigDir = new java.io.File(configDir, "extended-properties");
		if (propConfigDir.exists())
		{
			// Find all XML files
			List<java.io.File> xmlFiles = FileUtil.findFiles(propConfigDir, 
				new RegexpFileFilter(Pattern.compile(".*\\.xml"), null));
			if (xmlFiles != null)
			{
				for (java.io.File xmlFile : xmlFiles)
				{
					String path = xmlFile.getAbsolutePath().replace(configDir, "").
						replace(java.io.File.separatorChar, '/');
					files.add(path);
				}
			}
		}
		return files;	
	}

	/**
		Get the path to the XML file containing raw data type definitions.
		@see RawDataTypes
	*/
	static List<String> getRawDataTypeFiles()
	{
		List<String> files = new ArrayList<String>();
		// Main raw data types file
		files.add(rawDataTypesFile);  
		// Scan raw-data-types directory
		String configDir = Config.getConfigDirectory().getAbsolutePath();
		java.io.File rawConfigDir = new java.io.File(configDir, "raw-data-types");
		if (rawConfigDir.exists())
		{
			// Find all XML files
			List<java.io.File> xmlFiles = FileUtil.findFiles(rawConfigDir, new RegexpFileFilter(Pattern.compile(".*\\.xml"), null));
			if (xmlFiles != null)
			{
				for (java.io.File xmlFile : xmlFiles)
				{
					String path = xmlFile.getAbsolutePath().replace(configDir, "").
						replace(java.io.File.separatorChar, '/');;
					files.add(path);
				}
			}
		}
		return files;
	}

	/**
		Get the name of the dynamic catalog.
	*/
	static String getDynamicCatalog()
	{
		return dynamicCatalog;
	}
	
	/**
		Get the name of the dynamic schema.
	*/
	static String getDynamicSchema()
	{
		return dynamicSchema;
	}

	/**
		If external plug-in JAR files should automatically be unloaded if the are modified.
		@since 2.4
	*/
	static boolean autoUnloadPlugins()
	{
		return autoUnloadPlugins;
	}
	
	/**
	 	Get the directory where plug-ins and extensions must be installed.
	 	@return java.io.File A directory where the external plugins are located/installed
	 	@since 3.0
	*/
	public static java.io.File getPluginsDirectory()
	{
		return pluginsDirectory;
	}
	
	/**
		Check if auto-compression has been enabled or not. Files that
		are checked with the {@link File#setMimeTypeAuto(String, ItemSubtype)}
		and will automaically be compressed if auto-compression is enabled and
		the {@link MimeType#getAutoCompress()} returns TRUE.
		@return TRUE if auto-compression is enabled
		@since 2.5
	*/
	public static boolean autoCompressionEnabled()
	{
		return !noAutoCompression;
	}
	
	/**
		Create a new <code>SessionControl</code> object. If BASE is not 
		running it will be started.
		
		@param externalClientId The external id for a registered client
			application, or null if not important
		@param remoteId An optional parameter identifying the remote
			computer, for example the IP number. This should make it harder
			to hijack a login by guessing a <code>sessionControlId</code>.
		@param sessionControlId The id to give the new <code>SessionControl</code>
			object, or null to generate a new id
		@throws ItemNotFoundException If no client with the specified 
			external id exists
		@throws ItemAlreadyExistsException If another <code>SessionControl</code>
			with the given id already exists
		@throws BaseException If there is another error
		@see SessionControl#getId()
	*/
	public static synchronized SessionControl newSessionControl(
		String externalClientId, String remoteId, String sessionControlId)
		throws BaseException
	{
		if (!isRunning()) start();
		if (sessionControlId == null)
		{
			do
			{
				sessionControlId = generateRandomId(8);
			} while (sessionCache.containsKey(sessionControlId));
		}
		else if (sessionCache.containsKey(sessionControlId))
		{
			throw new ItemAlreadyExistsException("SessionControl[id="+sessionControlId+"]");
		}

		SessionControl sc = new SessionControl(sessionControlId, externalClientId, true, remoteId);
		sessionCache.put(sessionControlId, sc);
		return sc;
	}

	/**
		Create a new session control for executing plugins.
		@param parent The parent session control
		@param plugin The plugin to execute
		@param job The job that executes the plug-in, or null
		@return The new plugin session control
	*/
	static synchronized PluginSessionControl newPluginSessionControl(SessionControl parent, 
		PluginDefinition plugin, Job job)
	{
		String sessionControlId = null;
		do
		{
			sessionControlId = generateRandomId(8);
		} while (sessionCache.containsKey(sessionControlId));
		return new PluginSessionControl(sessionControlId, parent, plugin, job);
	}
	
	/**
		Get an existing <code>SessionControl</code> object if you know
		the id. 

		@param sessionControlId The id of the <code>SessionControl</code> object
		@param externalClientId The external id for a registered client
			application, or null if not important 
		@param remoteId If this parameter was passed to the <code>newSessionControl</code>
			method, the same value must be passed here. Use, for example, the IP address
			of the client computer. This should make it harder to hijack a login
			by guessing a <code>sessionControlId</code>
		@param checkLoggedIn TRUE to check if a user is logged in, FALSE to not check
		@return The <code>SessionControl</code> object
		@throws ItemNotFoundException If an object with the specified id could not be found
		@throws PermissionDeniedException If the remote or external client id is different from the 
			remote/external client id that was passed when the <code>SessionControl</code> object 
			was created
		@throws NotLoggedInException If a user is not logged in when a check was requested
			
		@throws BaseException If there is another error
		@see #newSessionControl(String,String,String)
		@see SessionControl#getId()
		@since 3.9
	*/
	public static SessionControl getSessionControl(String sessionControlId, String externalClientId, String remoteId, boolean checkLoggedIn)
		throws ItemNotFoundException, BaseException
	{
		if (!isRunning()) start();
		SessionControl sc = sessionCache.get(sessionControlId);
		if (sc == null) throw new ItemNotFoundException("SessionControl[ID="+sessionControlId+"]");
		if (!StringUtil.isEqualOrNull(remoteId, sc.getRemoteId()))
		{
			log.warn("getSessionControl: Invalid remoteId: "+remoteId+"; expected: "+sc.getRemoteId());
			throw new PermissionDeniedException("Invalid remoteId ("+remoteId+"; expected: "+sc.getRemoteId()+")");
		}
		if (externalClientId != null && !sc.isAllowedToUseClient(externalClientId))
		{
			log.warn("getSessionControl: Invalid externalClientId: "+externalClientId+"; expected: "+sc.getExternalClientId());
			throw new PermissionDeniedException(Permission.USE, "Client[externalId="+externalClientId+"]");
		}
		if (checkLoggedIn && !sc.isLoggedIn()) 
		{
			log.warn("getSessionControl: User not logged in. externalClientId: "+externalClientId+"; remoteId: "+remoteId);
			throw new NotLoggedInException();
		}
		sc.updateLastAccess();
		return sc;
	}
	
	/**
		Get an existing <code>SessionControl</code> object if you know
		the id. 
		
		NOTE! For maximum backwards compatibility calling this method 
		is equivalent to using "net.sf.basedb.clients.web" as the 
		external client id and not checking if a user is logged in. 
		Existing applications that use a different id MUST be updated to 
		use the new method or they may no longer work in BASE 3.9.
	
		@deprecated In 3.9, use {@link #getSessionControl(String, String, String, boolean)} instead
	*/
	@Deprecated
	public static SessionControl getSessionControl(String sessionControlId, String remoteId)
		throws ItemNotFoundException, BaseException
	{
		return getSessionControl(sessionControlId, "net.sf.basedb.clients.web", remoteId, false);
	}
	
	/**
		Clean the cache from unused <code>SessionControl</code> objects.
		@param force If TRUE all objects will be removed from the cache
			regardless of if they are in use or not
	*/
	private static void cleanSessionControlCache(boolean force)
	{
		if (sessionCache == null) return;
		if (!force)
		{
			// Ping all sessioncontrols in the pingers set
			for (Pinger p : pingers)
			{
				p.ping();
			}
		}
		
		synchronized(sessionCache)
		{
			Iterator<SessionControl> i = sessionCache.values().iterator();
			long now = System.currentTimeMillis();
			long timout = 60000L * sessionCacheTimeout; // minutes --> milliseconds
			while (i.hasNext())
			{
				SessionControl sc = i.next();
				if (force || ((now - sc.getLastAccess()) > timout))
				{
					log.info("cleanSessionControlCache: Remove[ID="+sc.getId()+"]");
					try
					{
						if (sc.isLoggedIn()) sc.logout();
					}
					catch (Throwable t)
					{
						log.error("Exception during cleanSessionControlCache", t);
					}
					finally
					{
						sc.close();
						i.remove();
					}
				}
			}
		}
	}

	/**
		Generate a random id.  
		@param bytes The number of bytes of the id
		@return The generated id represented as a hexadecimal string
	*/
	public static String generateRandomId(int bytes)
	{
		byte[] random = new byte[bytes];
		Random generator = new Random();
		StringBuilder sb = new StringBuilder(2*bytes);
		generator.nextBytes(random);
		for (int i = 0; i<random.length; i++)
		{
			int hex = random[i] & 0xFF;
			sb.append(hex < 16 ? "0": "").append(Integer.toHexString(hex));
		}
		return sb.toString();
	}
	
	static Timer getCoreTimer()
	{
		if (timer == null)
		{
			synchronized (Application.class)
			{
				if (timer == null) timer = new Timer("Base2Core.Timer", true);
			}
		}
		return timer;
	}
	
	/**
		Get a scheduler that can be used for registering {@link TimerTask}
		for executions. Each task that is added to the scheduler is executed in
		it's own thread.
		@return A scheduler
	*/
	public static Scheduler getScheduler()
	{
		if (scheduler == null)
		{
			synchronized (Application.class)
			{
				if (scheduler == null) scheduler = new Scheduler(getCoreTimer(), true);
			}
		}
		return scheduler;
	}
	
	private static class SessionControlCacheCleaner
		extends TimerTask
	{
		/*
			From the TimerTask class
			-------------------------------------------
		*/
		@Override
		public void run()
		{
			log.info("Cleaning session control cache");
			cleanSessionControlCache(false);
		}
		// -------------------------------------------
	}
	
	private static class DbCleaner
		extends TimerTask
	{
		/*
			From the TimerTask class
			-------------------------------------------
		*/
		@Override
		public void run()
		{
			log.info("Cleaning database");
			// Stray any-to-any links
			try
			{
				int numDeleted = AnyToAny.deleteStrayLinks(null);
				log.info("Found " + numDeleted + " stray any-to-any links");
			}
			catch (RuntimeException ex)
			{
				log.error("Could not delete stray any-to-any links", ex);
			}
			
			// Change history entries referencing deleted items
			try
			{
				int numDeleted = ChangeHistory.deleteStrayEntries(null);
				log.info("Found " + numDeleted + " stray history entries");
			}
			catch (RuntimeException ex)
			{
				log.error("Could not delete stray history entries", ex);
			}
			
			// Empty annotation sets
			try
			{
				int numDeleted = AnnotationSet.deleteEmptyAnnotationSets(null);
				log.info("Found " + numDeleted + " empty annotation sets");
			}
			catch (RuntimeException ex)
			{
				log.error("Could not delete unused annotation sets", ex);
			}
			
			// Project-specific annotions for deleted projects
			try
			{
				int numDeleted = ProjectSpecificAnnotationsManager.cleanupProjectSpecificAnnotations();
				log.info("Found " + numDeleted + " project-specific annotations belonging to deleted projects");
			}
			catch (RuntimeException ex)
			{
				log.error("Could not delete project-specific annotations belonging to deleted projects", ex);
			}
			
			// Item and project keys
			try
			{
				int numDeleted = ItemKey.deleteUnusedItemKeys();
				log.info("Found " + numDeleted + " unused item keys");
			}
			catch (RuntimeException ex)
			{
				log.error("Could not delete unused item keys", ex);
			}
			
			try
			{
				int numDeleted = ProjectKey.deleteUnusedProjectKeys();
				log.info("Found " + numDeleted + " unused project keys");
			}
			catch (RuntimeException ex)
			{
				log.error("Could not delete unused project keys", ex);
			}
			
			log.info("Finished cleaning of database");
		}
		// -------------------------------------------
	}

	
	private static class SecondaryStorageControllerTask
		extends TimerTask
	{
		
		private final SecondaryStorageController ssc;
		private boolean isRunning;
		
		private SecondaryStorageControllerTask(SecondaryStorageController ssc)
		{
			this.ssc = ssc;
			this.isRunning = false;
		}
		
		/*
			From the TimerTask class
			-------------------------------------------
		*/
		@Override
		public void run()
		{
			if (isRunning) return;
			isRunning = true;
			log.info("Executing secondary storage controller: " + ssc);
			ssc.run();
			isRunning = false;
		}
		// -------------------------------------------
	}
	
	private static Set<Pinger> pingers = 
		Collections.synchronizedSet(new HashSet<Pinger>());
	
	/**
		Create a new <code>Pinger</code> object that is used
		to keep a session control alive.
		
		@param sc The SessionControl that the pinger should protect
		@return A pinger object
		@since 2.4.5
	*/
	public static Pinger newPinger(SessionControl sc)
	{
		Pinger p = new Pinger(sc);
		pingers.add(p);
		return p;
	}
	
	private static void stopPinger(Pinger p)
	{
		pingers.remove(p);
	}
	
	/**
		A <code>Pinger</code> is an object that is used to protect another
		object from timing out by automatically pinging it. This can for
		example be used by a client to override the timeout for SessionControl
		items.
		
		@since 2.4.5
	*/
	public static class Pinger
	{
		private final SessionControl sc;
		private Pinger(SessionControl sc)
		{
			this.sc = sc;
		}
		/**
			Ping the SessionControl by updating the last access time:
			{@link SessionControl#updateLastAccess()}
		*/
		public void ping()
		{
			if (!sc.isClosed()) sc.updateLastAccess();
		}
		/**
			Stop this pinger.
		*/
		public void stop()
		{
			Application.stopPinger(this);
		}
	}
	
}
