/*
	$Id: PluginResponse.java 7264 2016-12-14 11:30:12Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Johan Enell

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.Job.ProgressReporterImpl;
import net.sf.basedb.core.plugin.ExportOutputStream;
import net.sf.basedb.core.plugin.ImmediateDownloadExporter;
import net.sf.basedb.core.plugin.Plugin;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalReceiver;
import net.sf.basedb.core.signal.SignalTarget;

import java.io.IOException;
import java.util.List;
import java.util.Collections;

/**
	Get the response from a configuration step or execution of a plugin.
	Use {@link #getStatus()} to find out the status of the last invokation.
	See {@link net.sf.basedb.core.PluginRequest} for more information.

	@author Nicklas
	@version 2.0
	@base.modified $Date$
*/
public final class PluginResponse
{

	private final PluginConfigurationRequest request;
	private final Response response;
	private final boolean allowImmediateExecution;
	private final boolean allowImmediateDownload;
	private Response.Status status;
	private Job.ExecutionTime estimatedExecutionTime;
	private String message;
	private boolean dryRun;
	private String suggestedJobName;
	private List<? extends Throwable> errorList;
	private String nextCommand;

	PluginResponse(PluginConfigurationRequest request, 
		boolean allowImmediateExection, boolean allowImmediateDownload, boolean dryRun)
	{
		this.request = request;
		this.response = new ResponseImpl();
		this.status = Response.Status.ERROR;
		this.allowImmediateDownload = allowImmediateDownload;
		this.allowImmediateExecution = allowImmediateExection;
		this.message = "A response has not been specified by the plugin";
		this.dryRun = dryRun;
	}

	/**
		Get the status of the response.
	*/
	public Response.Status getStatus()
	{
		return status;
	}
	
	/**
		Check if the job should be executed or was executed in
		dry-run mode.
		@since 2.8
	*/
	public boolean isDryRun()
	{
		return dryRun;
	}
	
	/**
		Get the job name that was suggested by the job.
	
		@return A job name or null if no name has been suggested
		@since 2.17
	*/
	public String getSuggestedJobName()
	{
		return suggestedJobName;
	}
	
	/**
		Get the return message from the plugin.
	*/
	public String getMessage()
	{
		return message;
	}
	
	/**
		Get the estimated execution time of the job.
		This value is only meaningful after a successful configuration
		sequence for a job.
	*/
	public Job.ExecutionTime getEstimatedExecutionTime()
	{
		return estimatedExecutionTime == null ? Job.ExecutionTime.LONG : estimatedExecutionTime;
	}
	
	/**
		Get a list containing detailed error messages if the status is 
		{@link net.sf.basedb.core.plugin.Response.Status#ERROR}.
		@see #getMessage()
	*/
	public List<? extends Throwable> getErrorList()
	{
		return errorList;
	}
	
	/**
		Get the next command sent via {@link net.sf.basedb.core.plugin.Response.Status#CONTINUE}.
		@return The next command or null
		@since 2.16
	*/
	public String getNextCommand()
	{
		return nextCommand;
	}
	
	/**
		Get a <code>PluginRequest</code> object that handles the next request, 
		if status is {@link net.sf.basedb.core.plugin.Response.Status#CONTINUE}.
	*/
	public PluginConfigurationRequest getNextRequest()
	{
		if (getStatus() != Response.Status.CONTINUE)
		{
			throw new PermissionDeniedException("Can't continue since response status is: " + getStatus());
		}
		request.setNextCommand(nextCommand);
		return request;
	}
	
	public PluginExecutionRequest getExecutionRequest(ProgressReporter progress)
		throws PermissionDeniedException
	{
		if (getStatus() != Response.Status.EXECUTE_IMMEDIATELY)
		{
			throw new PermissionDeniedException("Can't execute immediately since response status is: " + getStatus());
		}
		ProgressReporterImpl jobProgress = new ProgressReporterImpl(request.getJob(), Application.getHostName(), null, null, progress);
		PluginExecutionRequest executionRequest = new PluginExecutionRequest(request, Request.COMMAND_EXECUTE);
		executionRequest.setProgressReporter(jobProgress);
		return executionRequest;
	}
	
	/**
		Save all parameters that has been configured during the configuration sequence to
		the database.
	 	@param dc An open DbControl to use when accessing the database. 
	*/
	public void saveParameters(DbControl dc)
	{
		if (!getStatus().allowSaveParameters())
		{
			throw new PermissionDeniedException("Can't save parameters since response status is: " + getStatus());
		}
		request.saveParameters(dc, request.getGuiContext());
	}
	
	public void downloadImmediately(ExportOutputStream out, ProgressReporter progress)
		throws IOException
	{
		if (getStatus() != Response.Status.DOWNLOAD_IMMEDIATELY)
		{
			throw new PermissionDeniedException("Cannot export since response status is: " + getStatus());
		}
		Plugin plugin = request.getPlugin();
		if (plugin instanceof SignalTarget)
		{
			SignalHandler signalHandler = ((SignalTarget)plugin).getSignalHandler();
			if (progress instanceof SignalReceiver && signalHandler != null)
			{
				((SignalReceiver)progress).registerSignalHandler(signalHandler);
			}
		}
		try
		{
			ImmediateDownloadExporter exporter = (ImmediateDownloadExporter)plugin;
			exporter.doExport(out, progress);
		}
		finally
		{
			plugin.done();
		}
	}
	
	/**
		Get a new reponse object. Used by the {@link PluginRequest#invoke()}
		method.
	*/
	Response getResponseImpl()
	{
		return response;
	}
	
	/**
		Set a singe error message.
	*/
	void setError(String message, Throwable t)
	{
		reset(Response.Status.ERROR);
		this.message = message;
		errorList = Collections.singletonList(t);
	}
	
	private void reset(Response.Status status)
	{
		this.status = status;
		this.message = null;
		this.errorList = null;
		this.nextCommand = null;
	}
	
	/**
		An implementation of the Response interface used by the plugin
		to return information to the client.
	*/
	private class ResponseImpl
		implements Response
	{
		/*
			From the Response interface (plugin package)
			---------------------------------------------------
		*/
		@Override
		public void setContinue(String nextCommand)
			throws PermissionDeniedException
		{
			reset(Response.Status.CONTINUE);
			PluginResponse.this.nextCommand = nextCommand;
		}
		
		@Override
		public void setDone(String message, Job.ExecutionTime estimatedExecutionTime)
		{
			reset(Response.Status.DONE);
			PluginResponse.this.message = message;
			PluginResponse.this.estimatedExecutionTime = estimatedExecutionTime;
		}
		
		@Override
		public void setExecuteImmediately(String message, Job.ExecutionTime estimatedExecutionTime, boolean failIfDenied)
		{
			if (!allowImmediateExecution && failIfDenied)
			{
				String msg = "Immediate execution is not allowed for this plugin.";
				if (Request.COMMAND_EXECUTE.equals(request.getCommand()))
				{
					msg = "Immediate execution is not allows as a response to the command: " + request.getCommand();
				}
				throw new PermissionDeniedException(msg);
			}
			if (allowImmediateExecution)
			{
				reset(Response.Status.EXECUTE_IMMEDIATELY);	
				PluginResponse.this.estimatedExecutionTime = estimatedExecutionTime;
				PluginResponse.this.message = message;
			}
			else
			{
				setDone(message, estimatedExecutionTime);
			}
		}
		
		@Override
		public void setDownloadImmediately(String message, Job.ExecutionTime estimatedExecutionTime, boolean failIfDenied)
		{
			if (!allowImmediateDownload && failIfDenied)
			{
				String msg = "Immediate export is not supported by this plugin.";
				if (Request.COMMAND_EXECUTE.equals(request.getCommand()))
				{
					msg = "Immediate export is not allows as a response to the command: " + request.getCommand();
				}
				throw new PermissionDeniedException(msg);
			}
			if (allowImmediateDownload)
			{
				reset(Response.Status.DOWNLOAD_IMMEDIATELY);	
				PluginResponse.this.estimatedExecutionTime = estimatedExecutionTime;
				PluginResponse.this.message = message;
			}
			else
			{
				setDone(message, estimatedExecutionTime);
			}
		}
		
		@Override
		public void setDone(String message)
		{
			reset(Response.Status.DONE);
			PluginResponse.this.message = message;
		}
		
		@Override
		public void setError(String message, List<? extends Throwable> errorList)
		{
			reset(Response.Status.ERROR);
			PluginResponse.this.message = message;
			PluginResponse.this.errorList = errorList;
		}
		
		@Override
		public void setDryRun(boolean dryRun)
		{
			PluginResponse.this.dryRun = dryRun;
		}
		
		@Override
		public void setSuggestedJobName(String jobName)
		{
			PluginResponse.this.suggestedJobName = jobName;
		}
		// -------------------------------------------
	}
}
