/*
	$Id$

	Copyright (C) 2016 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.data.AnnotationData;
import net.sf.basedb.core.snapshot.SnapshotManager;

/**
	Internal class for keeping the "override_id" column in
	the "Annotations" table synchronized when annotations
	are creted and deleted.
	@since 3.10
	@author nicklas
*/
class ProjectSpecificAnnotationsManager 
	implements TransactionalAction
{
	
	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(ProjectSpecificAnnotationsManager.class);
	private static final boolean isDebugEnabled = log.isDebugEnabled();

	/**
		Delete project-specific annotations that belong to projects
		that no longer exists.
		@return The number of deleted annotation sets
	*/
	static synchronized int cleanupProjectSpecificAnnotations()
		throws BaseException
	{
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		int numDeleted = 0;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			
			ProjectSpecificAnnotationsManager psm = new ProjectSpecificAnnotationsManager(session);
			numDeleted = psm.deleteAnnotationsForNonExistingProjects();
			
			HibernateUtil.commit(tx);
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
		}
		return numDeleted;
	}
	
	private final org.hibernate.Session session;
	private final Set<Integer> deletedDefault;
	private final Set<Integer> deletedPrimary;
	private final List<AnnotationInfo> createdPrimary;
	private final List<AnnotationInfo> createdInherited;
	
	private org.hibernate.Query findProjectSpecific;

	ProjectSpecificAnnotationsManager(org.hibernate.Session session)
	{
		this.session = session;
		this.deletedDefault = new HashSet<Integer>();
		this.deletedPrimary = new HashSet<Integer>();
		this.createdPrimary = new ArrayList<AnnotationInfo>();
		this.createdInherited = new ArrayList<AnnotationInfo>();
	}
	
	/**
		Find a project-specific annotation for a given default annotation.
	*/
	Annotation findProjectSpecificAnnotation(Annotation defaultAnnotation, int projectId)
	{
		if (findProjectSpecific == null)
		{
			findProjectSpecific = HibernateUtil.getPredefinedQuery(session, 
				"PS_FIND_PROJECT_SPECIFIC_ANNOTATION");
			/*
				SELECT ad FROM AnnotationData
				WHERE ad.projectId = :activeProject
				AND ad.overrideId = :defaultAnnotation
				AND ad.annotationSet = :annotationSet
			*/
			findProjectSpecific.setInteger("activeProject", projectId);
		}
		
		findProjectSpecific.setInteger("defaultAnnotation", defaultAnnotation.getId());
		findProjectSpecific.setInteger("annotationSet", defaultAnnotation.getData().getAnnotationSet().getId());
		AnnotationData psAnnotation = HibernateUtil.loadData(AnnotationData.class, findProjectSpecific);
		return psAnnotation == null ? null : defaultAnnotation.getDbControl().getItem(Annotation.class, psAnnotation);
	}
	
	/**
		A PRIMARY annotation has been created in this transaction.
		If this is a DEFAULT annotation (projectId==0) we need to set 
		"override_id" for project-specific annotations to point to this annotation.
		If this is a PROJECT-SPECIFIC annotation (projectId!=0) we
		need to find the ID of the default annotation.

		@param id The annotation that was created
		@param annotationSet The annotation set the annotation belongs to
		@param annotationType The annotation type of the annotation
		@param project The id of the project, or 0 for default annotation
	*/
	void primaryCreated(int id, int annotationSet, int annotationType, int project)
	{
		createdPrimary.add(new AnnotationInfo(id, annotationSet, annotationType, project));
	}

	/**
		An INHERITED/CLONED annotation has been created in this transaction.
		If this is a DEFAULT annotation (projectId==0) we need to set
		"override_id" for project-specific annotations to point to this annotation.
		If this is a PROJECT-SPECIFIC annotation (projectId!=0) we
		need to find the ID of the default annotation.
		
		@param id The annotation that was created
		@param annotationSet The annotation set the annotation belongs to
		@param parent The id of the parent annotation
		@param parentSet The id of the parent annotation set
		@param parentOverride The "override_id" column from the parent annotation
		@param project The id of the project, or 0 for default annotation
	*/
	void inheritedCreated(int id, int annotationSet, int parent, int parentSet, int parentOverride, int project)
	{
		createdInherited.add(new AnnotationInfo(id, annotationSet, parent, parentSet, parentOverride, project));
	}
	
	/**
		A default annotation is being deleted in this transaction.
		We need to set override_id=0 on all other annotations that are 
		referencing this annotation. This method can be used for both
		primary, inherited and cloned annotations.
	*/
	void defaultDeleted(int id, boolean isPrimary)
	{
		deletedDefault.add(id);
		if (isPrimary) deletedPrimary.add(id);
	}

	@Override
	public void onBeforeCommit() 
	{
		if (isDebugEnabled) log.debug("deletedDefault: " + deletedDefault);

		if (deletedDefault.size() > 0)
		{
			if (deletedPrimary.size() == 0) deletedPrimary.add(-1);
			org.hibernate.Query query = HibernateUtil.getPredefinedSQLQuery(session, 
				"PS_RESET_OVERRIDE_ON_PROJECT_ANNOTATIONS");
			/*
				UPDATE [Annotations]  
				SET [override_id] = 0
				WHERE [override_id] IN (:deletedAnnotations)
				OR [inherited_id] IN (
					SELECT [id] FROM [Annotations] 
					WHERE [override_id] IN (:deletedPrimaryAnnotations)
				)
			*/
			query.setParameterList("deletedAnnotations", deletedDefault, Type.INT.getTypeWrapper().getHibernateType());
			query.setParameterList("deletedPrimaryAnnotations", deletedPrimary, Type.INT.getTypeWrapper().getHibernateType());
			int numUpdated = query.executeUpdate();
			
			if (isDebugEnabled) log.debug(numUpdated + " annotations updated -> override_id=0");
		}

		if (isDebugEnabled) log.debug("createdPrimary: " + createdPrimary);

		if (createdPrimary.size() > 0)
		{
			org.hibernate.Query defaultQuery = HibernateUtil.getPredefinedSQLQuery(session, 
				"PS_UPDATE_PROJECT_ANNOTATION_OVERRIDE_REF");
			/*
				UPDATE [Annotations] 
				SET [override_id] = :annotationId
				WHERE [annotationset_id] = :annotationSetId
				AND [annotationtype_id] = :annotationTypeId
				AND [source] = 0  AND [project_id] <> 0
			*/
			org.hibernate.Query projectQuery = HibernateUtil.getPredefinedSQLQuery(session, 
				"PS_FIND_OVERRIDE_ID_FOR_PROJECT_ANNOTATION");
			/*
				UPDATE [Annotations] 
				SET [override_id] = COALESCE(
					(SELECT [id] FROM [Annotations] 
					WHERE [annotationset_id] = :annotationSetId
					AND [annotationtype_id] = :annotationTypeId
					AND [source] = 0 AND [project_id] = 0)
					, 0)
				WHERE id = :annotationId
			 */
			int numUpdated = 0;
			for (AnnotationInfo ai : createdPrimary)
			{
				if (ai.project == 0)
				{
					// A PRIMARY+DEFAULT annotation -- update override_id on project-specific annotations
					defaultQuery.setInteger("annotationId", ai.id);
					defaultQuery.setInteger("annotationSetId", ai.annotationSet);
					defaultQuery.setInteger("annotationTypeId", ai.annotationType);
					numUpdated += defaultQuery.executeUpdate();
				}
				else
				{
					// A PRIMARY+PROJECT-SPECIFIC annotation -- find id of default annotation se we can set override_id
					projectQuery.setInteger("annotationId", ai.id);
					projectQuery.setInteger("annotationSetId", ai.annotationSet);
					projectQuery.setInteger("annotationTypeId", ai.annotationType);
					numUpdated += projectQuery.executeUpdate();
				}
			}
			if (isDebugEnabled) 
			{
				log.debug(numUpdated + " PRIMARY annotations updated -> override_id=<default-id>");
			}
		}
		
		if (isDebugEnabled) log.debug("createdInherited: " + createdInherited);
		
		if (createdInherited.size() > 0)
		{
			org.hibernate.Query defaultQuery = HibernateUtil.getPredefinedSQLQuery(session, 
				"PS_UPDATE_PROJECT_ANNOTATION_OVERRIDE_REF_INHERITED");
			/*
				UPDATE [Annotations] 
				SET [override_id] = :annotationId
				WHERE [annotationset_id] = :annotationSetId
				AND [inherited_id] IN (
					SELECT [id] FROM [Annotations]
					WHERE [annotationset_id] = :parentSetId
					AND [override_id] = :parentId
				)
			 */
			org.hibernate.Query projectQuery = HibernateUtil.getPredefinedSQLQuery(session, 
				"PS_FIND_OVERRIDE_ID_FOR_PROJECT_ANNOTATION_INHERITED");
			/*
				UPDATE [Annotations] 
				SET [override_id] = COALESCE(
					(SELECT [id] FROM [Annotations] 
					WHERE [annotationset_id] = :annotationSetId
					AND [inherited_id] = :parentOverrideId
					AND [source] <> 0 
					AND [project_id] = 0)
					, 0)
				WHERE id = :annotationId
			 */
			int numUpdated = 0;
			for (AnnotationInfo ai : createdInherited)
			{
				if (ai.project == 0)
				{
					// An INHERITED+DEFAULT annotation -- update override_id on project-specific annotations
					defaultQuery.setInteger("annotationId", ai.id);
					defaultQuery.setInteger("annotationSetId", ai.annotationSet);
					defaultQuery.setInteger("parentId", ai.parent);
					defaultQuery.setInteger("parentSetId", ai.parentSet);
					numUpdated += defaultQuery.executeUpdate();
				}
				else
				{
					// An INHERITED+PROJECT-SPECIFIC annotation -- find id of default annotation se we can set override_id
					projectQuery.setInteger("annotationId", ai.id);
					projectQuery.setInteger("annotationSetId", ai.annotationSet);
					projectQuery.setInteger("parentOverrideId", ai.parentOverride);
					numUpdated += projectQuery.executeUpdate();
				}
			}
			if (isDebugEnabled) 
			{
				log.debug(numUpdated + " PRIMARY annotations updated -> override_id=<default-id>");
			}
		}
	}

	@Override
	public void onAfterCommit() 
	{}

	@Override
	public void onRollback() 
	{}
	
	@SuppressWarnings({"unchecked", "rawtypes"})
	int deleteAnnotationsForNonExistingProjects()
	{
		// Lists for holding all value_id:s found for specific value types
		List<Integer>[] valuesPerType = new List[Type.values().length+1];
		for (int i = 1; i <= Type.values().length; i++)
		{
			valuesPerType[i] = new ArrayList<Integer>();
		}

		// All project_id values found in "Annotations" table with no entry in "Projects" table
		Set<Integer> deletedProjects = new HashSet<Integer>();

		// All annotationset_id values found
		Set<Integer> annotationSets = new HashSet<Integer>();

		// Load information about all annotations that should be deleted
		org.hibernate.Query loadAnnotationsToDelete = HibernateUtil.getPredefinedSQLQuery(session, 
			"PS_LOAD_ANNOTATIONS_IN_NON_EXISTING_PROJECTS");
			/*
				SELECT a.[id], a.[annotationset_id], a.[value_id], a.[project_id], at.[value_type] 
				FROM [Annotations] a
				INNER JOIN [AnnotationTypes] at ON at.[id] = a.[annotationtype_id]
				WHERE [project_id] <> 0 AND [project_id] NOT IN (SELECT [id] FROM [Projects])
			*/
		List<Object[]> annotations = (List<Object[]>)loadAnnotationsToDelete.list();
		if (isDebugEnabled) 
		{
			log.debug("deletedProjects: " + deletedProjects);
			log.debug("annotations.size: " + annotations.size());
		}

		if (annotations.size() == 0) return 0;
		
		for (Object[] data : annotations)
		{
			annotationSets.add((Integer)data[1]);
			deletedProjects.add((Integer)data[3]);
			Integer valueId = (Integer)data[2];
			if (valueId != null)
			{
				Integer valueType = (Integer)data[4];
				valuesPerType[valueType.intValue()].add(valueId);
			}
		}

		// Delete the "Annotations" entries -- The list of projects is expected to be relatively small
		org.hibernate.Query deleteFromAnnotations = HibernateUtil.getPredefinedSQLQuery(session, 
			"PS_DELETE_FROM_ANNOTATIONS_IN_PROJECTS");
			/*
				DELETE FROM [Annotations] WHERE [project_id] IN (:deletedProjects)
			*/
		deleteFromAnnotations.setParameterList("deletedProjects", deletedProjects, Type.INT.getTypeWrapper().getHibernateType());
		int numDeleted = deleteFromAnnotations.executeUpdate();

		// Delete from ***Values and ParameterValues tables
		org.hibernate.Query deleteFromParameterValues = HibernateUtil.getPredefinedSQLQuery(session, 
			"PS_DELETE_FROM_PARAMETERVALUES");
		/*
			DELETE FROM [ParameterValues] WHERE [id] IN (:deletedValues)
		*/

		// Divide the query into chunks since the database driver may not support
		// an arbitrary large number of parameters
		int maxParameters = HibernateUtil.getDbEngine().getMaxParametersInQuery();

		for (int i = 0; i < Type.values().length; i++)
		{
			Type valueType = Type.values()[i];
			List<Integer> valuesForType = valuesPerType[valueType.getValue()];
			if (isDebugEnabled) log.debug("valuesPerType[" + valueType.name() + "]:" + valuesForType);
			if (valuesForType.size() > 0)
			{
				
				org.hibernate.Query deleteFromValues = HibernateUtil.getPredefinedSQLQuery(session, 
					"PS_DELETE_FROM_VALUES", valueType.getTableName());
				/*
					DELETE FROM [{1}] WHERE [id] IN (:deletedValues)
				*/
				
				int startIndex = 0;
				int endIndex = Math.min(maxParameters, valuesForType.size());
				while (startIndex < valuesForType.size())
				{
					List<Integer> sublist = valuesForType.subList(startIndex, endIndex);
				
					deleteFromValues.setParameterList("deletedValues", sublist, Type.INT.getTypeWrapper().getHibernateType());
					deleteFromValues.executeUpdate();

					deleteFromParameterValues.setParameterList("deletedValues", sublist, Type.INT.getTypeWrapper().getHibernateType());
					deleteFromParameterValues.executeUpdate();
					startIndex = endIndex;
					endIndex = Math.min(startIndex + maxParameters, valuesForType.size());
				}
			}
		}

		if (isDebugEnabled) log.debug("annotationSets: " + annotationSets);
		SnapshotManager.removeSnapshots(annotationSets);
		
		return numDeleted;
	}
	
	/**
		Holds some information about the annotation that
		has been created.
	*/
	static class AnnotationInfo
	{
		final int id;
		final int annotationSet;
		final int annotationType;
		final int project;
		final int parent;
		final int parentSet;
		final int parentOverride;
		
		AnnotationInfo(int id, int annotationSet, int annotationType, int project)
		{
			this.id = id;
			this.annotationSet = annotationSet;
			this.annotationType = annotationType;
			this.project = project;
			this.parent = 0;
			this.parentSet = 0;
			this.parentOverride = 0;
		}
		
		AnnotationInfo(int id, int annotationSet, int parent, int parentSet, int parentOverride, int project)
		{
			this.id = id;
			this.annotationSet = annotationSet;
			this.annotationType = 0;
			this.project = project;
			this.parent = parent;
			this.parentSet = parentSet;
			this.parentOverride = parentOverride;
		}
		
		@Override
		public String toString()
		{
			return Integer.toString(id);
		}
	}
}
