/**
	$Id: TransactionDetails.java 6080 2012-08-07 11:28:40Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.log;

import java.util.Date;

import org.hibernate.Query;

import net.sf.basedb.core.LogControl;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.data.ChangeHistoryData;

/**
	Holds information about the curren transaction, such as the logged in user,
	the active project, the plug-in that is executing, etc.

	@author Nicklas
	@version 2.13
	@base.modified $Date: 2012-08-07 13:28:40 +0200 (ti, 07 aug 2012) $
*/
public class TransactionDetails
{
	private final LogControl logControl;
	private final int userId;
	private final int sessionId;
	private final int clientId;
	private final int projectId;
	private final int pluginId;
	private final int jobId;
	
	private String userName;
	private String projectName;
	private String pluginName;
	private String jobName;
	
	/**
		Creates a new transaction details object with information from 
		given session control
	*/
	public TransactionDetails(LogControl logControl, SessionControl sc)
	{
		this.logControl = logControl;
		this.userId = sc.getLoggedInUserId();
		this.sessionId = sc.getCurrentSessionId();
		this.clientId = sc.getClientId();
		this.projectId = sc.getActiveProjectId();
		this.jobId = sc.getJobId();
		this.pluginId = sc.getPluginId();
	}
	
	/**
		Get the ID of the currently logged in user, or 0 if no user is logged in.
	*/
	public int getUserId()
	{
		return userId;
	}
	
	/**
		Get the ID of the session or 0 if no user is logged in.
	*/
	public int getSessionId()
	{
		return sessionId;
	}

	/**
		Get the ID of the client appliction that is used, or 0 if not known.
	*/
	public int getClientId()
	{
		return clientId;
	}
	
	/**
		Get the ID of the currently active project, or 0 if no project is active.
	*/
	public int getProjectId()
	{
		return projectId;
	}
	
	/**
		Get the ID of the currently running plug-in or 0 if no plug-in is running.
	*/
	public int getPluginId()
	{
		return pluginId;
	}
	
	/**
		Get the ID of the currently running job or 0 if no job is running.
	*/
	public int getJobId()
	{
		return jobId;
	}

	/**
		Convert the information about the transaction to a
		{@link ChangeHistoryData} object.
		@return A ChangeHistoryData object
	*/
	public ChangeHistoryData toChangeHistoryData()
	{
		ChangeHistoryData history = new ChangeHistoryData();
		history.setTime(new Date(System.currentTimeMillis()));
		history.setUserId(userId);
		history.setSessionId(sessionId);
		history.setClientId(clientId == 0 ? null : clientId);
		history.setProjectId(projectId == 0 ? null : projectId);
		history.setPluginId(pluginId == 0 ? null : pluginId);
		history.setJobId(jobId == 0 ? null : jobId);
		return history;
	}
	
	/**
		Utility method for loading the name of the currently logged in user.
		The first call to this method will lookup the name in the database
		but is then cached.
	 	@return The name of the currently logged in user
		@since 3.2
	*/
	public String getUserName()
	{
		if (userName == null && userId != 0)
		{
			Query query = logControl.createHqlQuery("select usr.name from UserData usr where usr.id = " + userId);
			userName = (String)query.uniqueResult();
		}
		return userName;
	}
	
	/**
		Utility method for loading the name of the currently active project.
		The first call to this method will lookup the name in the database
		but is then cached.
	 	@return The name of the currently active project, or null if no project is active
		@since 3.2
	*/
	public String getProjectName()
	{
		if (projectName == null && projectId != 0)
		{
			Query query = logControl.createHqlQuery("select prj.name from ProjectData prj where prj.id = " + projectId);
			projectName = (String)query.uniqueResult();
		}
		return projectName;
	}

	/**
		Utility method for loading the name of the currently executing plugin.
		The first call to this method will lookup the name in the database
		but is then cached.
	 	@return The name of the currently executing plugin, or null if no plugin is executing
		@since 3.2
	*/
	public String getPluginName()
	{
		if (pluginName == null && pluginId != 0)
		{
			Query query = logControl.createHqlQuery("select plg.name from PluginDefinitionData plg where plg.id = " + pluginId);
			pluginName = (String)query.uniqueResult();
		}
		return pluginName;
	}

	/**
		Utility method for loading the name of the currently executing job.
		The first call to this method will lookup the name in the database
		but is then cached.
	 	@return The name of the currently executing job, or null if no job is executing
		@since 3.2
	*/
	public String getJobName()
	{
		if (jobName == null && jobId != 0)
		{
			Query query = logControl.createHqlQuery("select job.name from JobData job where job.id = " + jobId);
			jobName = (String)query.uniqueResult();
		}
		return jobName;
	}

}
