/**
	$Id $

	Copyright (C) 2013 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.log.db;

import java.util.List;

import org.hibernate.Query;

import net.sf.basedb.core.Item;
import net.sf.basedb.core.LogControl;
import net.sf.basedb.core.PredefinedQuery;
import net.sf.basedb.core.data.AnyToAnyData;
import net.sf.basedb.core.data.ChangeHistoryDetailData;
import net.sf.basedb.core.data.LoggableData;
import net.sf.basedb.core.data.NameableData;
import net.sf.basedb.core.log.ChangeType;
import net.sf.basedb.core.log.EntityDetails;
import net.sf.basedb.core.log.EntityLogger;
import net.sf.basedb.core.log.LogManager;

/**
	Logger for any-to-any links that logs changes as
	updates to the 'from' item.
	
	@author Nicklas
	@since 3.3
	@base.modified $Date$
*/
public class AnyToAnyLogger
	implements EntityLogger
{

	private final DbLogManagerFactory factory;

	public AnyToAnyLogger(DbLogManagerFactory factory)
	{
		this.factory = factory;
	}
	
	/*
		From the EntityLogger interface
		-------------------------------
	*/
	@Override
	public void logChanges(LogManager logManager, EntityDetails details)
	{
		LogControl logControl = logManager.getLogControl();
		AnyToAnyData link = (AnyToAnyData)details.getEntity();

		// Get the type of the parent item
		int parentId = link.getFromId();
		Item parentType = Item.fromValue(link.getFromType());
		
		Class<?> parentClass = parentType.getDataClass();
		// If the parent item isn't implementing the LoggableData interface
		// we don't log the changes
		if (!LoggableData.class.isAssignableFrom(parentClass)) return;

		DbLogManager dbLogManager = ((DbLogManager)logManager);
		String propertyPrefix = "Link["+link.getName()+"]";
		ChangeType changeType = details.getChangeType();
		if (changeType == ChangeType.CREATE) 
		{
			changeType = ChangeType.ADD;
		}
		else if (changeType == ChangeType.DELETE)
		{
			changeType = ChangeType.REMOVE;
		}

		boolean delegateLogging = false;
		if (!factory.isLoggingDetailedProperties())
		{
			// If not logging detailed properties, we simply delegate logging
			delegateLogging = true;
		}
		else
		{
			// Check if the 'to' property is dirty
			int toIdIndex = details.getIndex("toId");
			int toTypeIndex = details.getIndex("toType");
			boolean isToDirty = details.getDirty(toIdIndex) == 1 || details.getDirty(toTypeIndex) == 1;

			// If CREATE or DELETE or 'to' is dirty, we create a log entry here
			if (changeType != ChangeType.UPDATE || isToDirty)
			{
				ChangeHistoryDetailData change = new ChangeHistoryDetailData();
				change.setChangeType(changeType.getValue());
				change.setItemId(parentId);
				change.setItemType(parentType.getValue());
				change.setChangeInfo(propertyPrefix);
				
				if (factory.isLoggingOldPropertyValues())
				{
					Item toType = Item.fromValue(link.getToType());
					int toId = link.getToId();
					
					if (changeType == ChangeType.ADD)
					{
						change.setNewValue(createLoggableString(logControl, toType, toId));
					}
					else if (changeType == ChangeType.REMOVE)
					{
						change.setOldValue(createLoggableString(logControl, toType, toId));
					}
					else if (isToDirty)
					{
						change.setNewValue(createLoggableString(logControl, toType, toId));
						change.setOldValue(createLoggableString(logControl, Item.fromValue((Integer)details.getPreviousState()[toTypeIndex]), (Integer)details.getPreviousState()[toIdIndex]));
					}
				}
				
				dbLogManager.logChangeDetails(change, details);
			}
			
			// If it's an update delegate logging for the remaining properties
			if (changeType == ChangeType.UPDATE)
			{
				delegateLogging = true;
				details.setDirty(toIdIndex, -1);
				details.setDirty(toTypeIndex, -1);
			}
		}
		
		if (delegateLogging)
		{
			List<ChangeHistoryDetailData> changes = details.toChangeDetails(parentType, parentId, propertyPrefix, changeType, factory.isLoggingDetailedProperties(), factory.isLoggingOldPropertyValues());
			dbLogManager.logChangeDetails(changes, details);
		}
		
	}
	// -------------------------------
	
	private String createLoggableString(LogControl logControl, Item itemType, int itemId)
	{
		String log = itemType.getDataClass().getSimpleName().replace("Data", "");
		if (NameableData.class.isAssignableFrom(itemType.getDataClass()))
		{
			// Query the database for the name
			Query q = logControl.createHqlQuery(
				PredefinedQuery.getQueryString("DBLOG_GET_NAME_FOR_ITEM", itemType.getDataClass().getName()));
			q.setInteger("id", itemId);
			String name = (String)q.uniqueResult();
			log += "[" + name + "]";
		}
		else
		{
			log += "[" + itemId + "]";
		}
		return log;
	}
	
}
