/**
	$Id$

	Copyright (C) 2016 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions.manager.processor;

import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.util.extensions.manager.ExtensionsFile;
import net.sf.basedb.util.extensions.manager.ExtensionsFileProcessor;
import net.sf.basedb.util.extensions.manager.ExtensionsManager;
import net.sf.basedb.util.extensions.manager.ProcessResults;
import net.sf.basedb.util.extensions.manager.ProcessResults.FileResults;
import net.sf.basedb.util.extensions.manager.ExtensionsFile.WriteableExtensionsFile;

/**
	Processor implementation for ignoring files.
	This processor need to be done in two steps due to
	ordering issues when an already installed extension
	has been selected for both uninstallation and ignoring.
	<p>
	The first step is executed via the manager: 
	{@link ExtensionsManager#processFiles(ExtensionsFileProcessor)}
	and should be done before the {@link net.sf.basedb.clients.web.extensions.UninstalledFileProcessor}
	is exectued.
	<p>
	The second step is either automatically executed when all files
	have been processed or by calling {@link #finalizeRegistration(ExtensionsManager)} 
	and should be done as the last step of the installation procedure.
	
	@author Nicklas
	@since 3.10
*/
public class IgnoreFileProcessor
	implements ExtensionsFileProcessor
{

	private final ProcessResults results;
	private Set<WriteableExtensionsFile> filesToIgnore;
	private int numFiles;
	private boolean delayRegistration;

	public IgnoreFileProcessor(ProcessResults results)
	{
		this.results = results;
	}
	
	/*
		From the RegisterExtensionsProcessor class
		------------------------------------------
	*/
	@Override
	public void begin(ExtensionsManager manager, int numFiles)
	{
		this.numFiles = 0;
		this.filesToIgnore = new HashSet<WriteableExtensionsFile>();
	}

	@Override
	public void processFile(ExtensionsManager manager, WriteableExtensionsFile wFile)
	{
		ExtensionsFile xtFile = wFile.getExtensionsFile();
		filesToIgnore.add(wFile);
		results.forgetOldResults(xtFile);
		numFiles++;
	}
	@Override
	public void done(ExtensionsManager manager)
	{
		if (!delayRegistration) finalizeRegistration(manager);
	}

	@Override
	public void done(ExtensionsManager manager, Throwable t)
	{}
	// ---------------------------------------------

	/**
		Set a flag to indicate if the actual registration of the found
		extensions and extension points should be delayed or not.
		The default is to not delay registration, which means that
		it happend when the {@link #done(ExtensionsManager)} method is
		called by the manager. If this flag is set the 
		{@link #finalizeRegistration(ExtensionsManager, boolean)}
		method must be explicitely called.
	*/
	public void setDelayRegistration(boolean delayRegistration)
	{
		this.delayRegistration = delayRegistration;
	}
	
	/**
		Is the actual registration delayed or not?
	*/
	public boolean isRegistrationDelayed()
	{
		return delayRegistration;
	}

	/**
		Get the number of files that was processed.
	*/
	public int getNumFiles()
	{
		return numFiles;
	}
	
	/**
		Finalize the ignore registration.
	*/
	public void finalizeRegistration(ExtensionsManager manager)
	{
		for (WriteableExtensionsFile wFile : filesToIgnore)
		{
			ExtensionsFile xtFile = wFile.getExtensionsFile();
			wFile.setError(false);
			manager.addIgnore(xtFile.getFile());
			FileResults currentResults = results.getResults(xtFile);
			String status = "Ignored";
			if (currentResults != null && currentResults.getStatus() != null)
			{
				status = currentResults.getStatus() + " + " + status;
			}
			results.setStatus(xtFile, status);
		}
	}


}
