/**
	$Id$

	Copyright (C) 2013 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.ma;

/**
	Calculates M and A values for 2-channel data. For "normal" intensity 
	values M and A are defined as follows:
	
	M = log2(ch1/ch2)
	A = log10(ch*ch2)/2
	
	Channel intensity corrections are used to calculate new
	values for channel 1 and 2. For corrections that use M+A
	values the corrections are usually calculated in log2-space
	and the resulting equations are:

	log2(newCh1) = log2(oldCh1) - factor
	log2(newCh2) = log2(oldCh2) + factor
	
	Source data that has been transformed (eg. log2-transformed) must use
	different equations to get the correct results.
	
	@author nicklas
	@since 3.2.4
*/
public class MACalculator 
{
	static final double INV_LN2 = 1 / Math.log(2);
	static final double INV_LN10x2 = 1 / (Math.log(10) * 2);

	
	public MACalculator()
	{}
	
	/**
		Calculate the M and A values from the ch1 and ch2 intensity values.
		See class description for information about the definitions of M and A.
		@return An array with [0] = M, [1] = A
	*/
	public double[] MA(double ch1, double ch2) 
	{
		double lnCh1 = Math.log(ch1);
		double lnCh2 = Math.log(ch2);
		// M = log2(ch1)-log2(ch2) = (ln(ch1)-ln(ch2)) / ln(2)
		// A = (log10(ch1)+log10(ch2))/2 = (ln(ch1)+ln(ch2))/(ln(10)*2)
		double[] MA = { (lnCh1-lnCh2)*INV_LN2, (lnCh1+lnCh2)*INV_LN10x2 };
		return MA;
	}

	/**
		Apply a correction factor to the ch1 and ch2 intensity values.
		@return An array with [0] = newCh1, [1] = newCh2
	*/
	public double[] correct(double ch1, double ch2, double factor) 
	{
		factor = Math.pow(2, factor);
		double[] result = {ch1 / factor , ch2 * factor};
		return result;
	}
	

}
