/**
	$Id: PhysicalBioAssayValidator.java 6799 2015-03-25 07:13:39Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DerivedBioAssay;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.PhysicalBioAssay;
import net.sf.basedb.core.Item;
import net.sf.basedb.util.listable.ExtractToParentExtractTransformer;
import net.sf.basedb.util.listable.SourceItemTransformer;
import net.sf.basedb.util.listable.TransformContext;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.Node.Type;

/**
	Validator implementation for physical bioassays. Validation rules:
	<ul>
	<li>Missing item: {@link Validator#MISSING_PHYSICALBIOASSAY}
	<li>Access denied: {@link Validator#DENIED_PHYSICALBIOASSAY}
	<li>Subtype of parent extract: {@link Validator#INCORRECT_EXTRACT_TYPE}
	</ul>

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2015-03-25 08:13:39 +0100 (on, 25 mar 2015) $
*/
public class PhysicalBioAssayValidator
	extends NameableNodeValidator<PhysicalBioAssay>
{
	
	public PhysicalBioAssayValidator()
	{
		super(Validator.MISSING_PHYSICALBIOASSAY, Validator.DENIED_PHYSICALBIOASSAY);
	}

	/* 
		From BasicValidator class
		-------------------------
	*/	
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode) 
	{
		super.postValidate(dc, context, node, parentNode);
		
		if (parentNode != null)
		{
			Node parentItemNode = parentNode.getNodeType() == Type.ITEM ? parentNode : parentNode.getParent();
			Item parentType = parentItemNode.getItemType();

			if (parentType == Item.EXTRACT)
			{
				checkExtractSubtype(dc, context, node, parentItemNode);
			}
			else if (parentType == Item.DERIVEDBIOASSAY)
			{
				DerivedBioAssayValidator.checkPhysicalBioAssaySubtype(dc, context, parentItemNode, node);
			}
		}
	}

	/**
		@return Always null (since it can never be missing)
	*/
	@Override
	protected Fix getMissingItemFix(DbControl dc, Node parentNode)
	{
		return null;
	}
	// ----------------------------

	
	/**
		Check the subtype of the extract that is the parent to a physical bioassay. Given that
		the bioassay has a non-null subtype, which has a related EXTRACT subtype, this
		method checks if the parent extract has the same subtype. If there is no related
		subtype for the bioassay, any subtype for the parent is accepted.
	*/
	public static void checkExtractSubtype(DbControl dc, OverviewContext context, Node bioAssayNode, Node extractNode)
	{
		PhysicalBioAssay bioAssay = (PhysicalBioAssay)bioAssayNode.getItem(dc);
		Extract extract = (Extract)extractNode.getItem(dc);
		
		try
		{
			ItemSubtype expectedSubtype = ItemSubtype.getRelatedSubtype(dc, bioAssay, Item.EXTRACT, 0);
			ItemSubtype subtype = extract.getItemSubtype();
			
			if (expectedSubtype != null && !expectedSubtype.equals(subtype))
			{
				ItemSubtype mySubtype = bioAssay.getItemSubtype();
				Fix subtypeFix = subtype == null ? null : 
					new Fix("Change related extract subtype of '" + mySubtype.getName() + 
						"' to '" + subtype.getName() + "'", mySubtype);

				context.createFailure(Validator.INCORRECT_EXTRACT_TYPE, bioAssayNode, 
					"Expected an extract with subtype: " + expectedSubtype.getName(), 
					new Fix("Change subtype of '" + extract.getName() + "' to '" + expectedSubtype.getName() + "'", extract),
					new Fix("Change subtype or parent of '" + bioAssayNode.getTitle() + "'", bioAssay),
					subtypeFix
				);
			}
		}
		catch (PermissionDeniedException ex)
		{}
	}

	/**
		Get all physical bioassays that are parents to the given derived bioassay. This is a 
		utility method that provides caching support so that queries don't have to be repeated 
		for the same derived bioassay.
		@since 3.2
	*/
	@SuppressWarnings("unchecked")
	public static List<PhysicalBioAssay> getPhysicalBioAssays(DbControl dc, OverviewContext context, DerivedBioAssay bioAssay)
	{
		String cacheKey = "derivedbioassay." + bioAssay.getId() + ".physicalbioassays";
		List<PhysicalBioAssay> physicalBioAssays = (List<PhysicalBioAssay>)context.getCachedObject(cacheKey);
		
		if (physicalBioAssays == null)
		{
			ItemQuery<PhysicalBioAssay> query = bioAssay.getPhysicalBioAssays();
			query.include(Include.ALL);
			physicalBioAssays = new ArrayList<PhysicalBioAssay>(query.list(dc));
			context.setCachedObject(cacheKey, physicalBioAssays);
		}
		return physicalBioAssays;
	}

	/**
		Get all parent extracts that are associated with the given physical bioassay.
		@since 3.2
	*/
	@SuppressWarnings("unchecked")
	public static Set<Extract> getAllParentExtracts(DbControl dc, OverviewContext context, PhysicalBioAssay bioAssay)
	{
		String cacheKey = "physicalbioassay." + bioAssay.getId() + ".extracts";
		Set<Extract> allExtracts = (Set<Extract>)context.getCachedObject(cacheKey);
		
		if (allExtracts == null)
		{
			ItemQuery<Extract> query = bioAssay.getExtracts(0);
			query.include(Include.ALL);
			
			SourceItemTransformer transformer = new ExtractToParentExtractTransformer(true);
			Set<Integer> extractIds = transformer.transform(new TransformContext(dc), new HashSet<Integer>(query.idList(dc)));
			
			allExtracts = new HashSet<Extract>(extractIds.size());
			for (int id : extractIds)
			{
				allExtracts.add(Extract.getById(dc, id));
			}
			context.setCachedObject(cacheKey, allExtracts);
		}
		return allExtracts;
	}

}
