/*
	$Id: HistogramPlot.java 6898 2015-05-12 09:27:08Z nicklas $

	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.plot;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.query.SqlResult;
import net.sf.basedb.core.query.SqlResultIterator;
import net.sf.basedb.core.signal.ThreadSignalHandler;

import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.plot.DatasetRenderingOrder;
import org.jfree.chart.plot.SeriesRenderingOrder;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYBarRenderer;
import org.jfree.chart.renderer.xy.XYItemRenderer;
import org.jfree.chart.renderer.xy.YIntervalRenderer;
import org.jfree.data.DomainOrder;
import org.jfree.data.Range;
import org.jfree.data.general.AbstractSeriesDataset;
import org.jfree.data.xy.IntervalXYDataset;

import java.awt.geom.Rectangle2D;
import java.io.Serializable;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


/**
	A simple plot utility for generating histogram plots without
	the need to delve deep into the JFreePlot package. Note! This class
	is experimental and it is likely that the implementation will change 
	in the future.
	
	@author Nicklas
	@version 2.0
	@base.modified $Date: 2015-05-12 11:27:08 +0200 (ti, 12 maj 2015) $
*/
public class HistogramPlot
{
	private Range domainRange;
	private Range rangeRange;
	private XYPlot plot;
	private HistogramDataset histogram;
	private HistogramDataset hilo;
	private JFreeChart chart;

	/**
		Create a new HistogramPlot.
		@param nameX The label on the X-axis
		@param nameY The label on the Y-axis
		@param renderer The renderer to use or null for the default (XYBarRenderer)
	*/
	public HistogramPlot(String nameX, String nameY, XYItemRenderer renderer)
	{
		
		histogram = new HistogramDataset();
		hilo = new HistogramDataset();
		NumberAxis domainAxis = new NumberAxis(nameX);
		domainAxis.setAutoRangeIncludesZero(false);
		domainRange = new Range(0.0, 0.0);
		rangeRange = new Range(0.0, 0.0);
		
      NumberAxis rangeAxis = new NumberAxis(nameY);
		rangeAxis.setAutoRangeIncludesZero(true);
		if (renderer == null) renderer = new XYBarRenderer();
      plot = new XYPlot(histogram, domainAxis, rangeAxis, renderer);
      plot.setDataset(1, hilo);
      YIntervalRenderer yRenderer = new YIntervalRenderer();
      yRenderer.setBaseShape(new Rectangle2D.Float(-3, -1, 7, 2));
      yRenderer.setBaseSeriesVisibleInLegend(false, false);
      plot.setRenderer(1, yRenderer, false);
      
      plot.setSeriesRenderingOrder(SeriesRenderingOrder.FORWARD);
      plot.setDatasetRenderingOrder(DatasetRenderingOrder.FORWARD);
      chart = new JFreeChart(plot);
		chart.setAntiAlias(false);
	}

	/**
		Add data to the plot. The <code>SqlResultIterator</code> should return
		the x value as a float in the first position, ie. <code>data.getFloat(1)</code>
		and, if used, the y value in the second position.
		<p>
		The y value is not used if the yAggregate parameter is {@link YAggregate#COUNT}.
	
		@param data The data to plot
		@param name The name of the data series
		@param binSize The size (= range on x-axis) of each histogram bin
		@param yAggregate What to plot on the y-axis
		@param hiloAggregate Optional parameter specifying if high/low values shold be drawn for each bin
	*/
	public void addData(SqlResultIterator data, String name, float binSize, YAggregate yAggregate, YAggregate hiloAggregate)
		throws SQLException
	{
		final boolean isCount = yAggregate == YAggregate.COUNT;
		float[] xValues = new float[(int)data.getTotalCount()];
		float[] yValues = isCount ? null : new float[(int)data.getTotalCount()];
		int i = 0;

		while (data.hasNext())
		{
			ThreadSignalHandler.checkInterrupted();
			SqlResult r = data.next();
			if (!isCount) yValues[i] = r.getFloat(2);
			xValues[i++] = r.getFloat(1);
		}
		HistogramSeries main = new HistogramSeries(name, xValues, yValues, binSize, yAggregate, Float.NaN, Float.NaN);
		histogram.addSeries(main);
		if (hiloAggregate != null)
		{
			hilo.addSeries(new HistogramSeries(name, main, hiloAggregate));
		}
		adjustRanges();
	}
	
	/**
		Auto-size of the x and y range doesn't seem to work, so we have to do it 
		manually.
	*/
	private void adjustRanges()
	{
		domainRange = Range.combine(domainRange, plot.getRenderer().findDomainBounds(histogram));
		if (hilo.getSeriesCount() > 0)
		{
			rangeRange = Range.combine(rangeRange, plot.getRenderer(1).findRangeBounds(hilo));
		}
		else
		{
			rangeRange = Range.combine(rangeRange, plot.getRenderer(0).findRangeBounds(histogram));
		}
		rangeRange = Range.expand(rangeRange, 0, 0.02);
		plot.getDomainAxis().setRange(domainRange);
		plot.getRangeAxis().setRange(rangeRange);
	}

	/**
		Add data to the plot. The <code>SqlResultIterator</code> should return
		the annotation value in the first position, ie. <code>data.getInt(1)</code>,
		the x value as a float in the second position, and, if used, the y value in the 
		third position.
		<p>
		The y value is not used if the yAggregate parameter is {@link YAggregate#COUNT}.
		
		@param data The data to plot
		@param annotations Annotation information that maps each data spot to 
			an annotation (see {@link ScatterPlot#addData(SqlResultIterator, Collection)}
		@param binSize The size (= range on x-axis) of each histogram bin
		@param yAggregate What to plot on the y-axis
		@param hiloAggregate Optional parameter specifying if high/low values shold be drawn for each bin
	 */
	public void addData(SqlResultIterator data, Collection<PlotAnnotation> annotations, float binSize, YAggregate yAggregate, YAggregate hiloAggregate)
		throws SQLException
	{
		final boolean isCount = yAggregate == YAggregate.COUNT;
		Map<Integer, TempSeries> series = new HashMap<Integer, TempSeries>(annotations.size());
		List<TempSeries> tempSeries = new ArrayList<TempSeries>(annotations.size());
		int meanCount = (int)(data.getTotalCount() / annotations.size());
		for (PlotAnnotation pa : annotations)
		{
			TempSeries ts = new TempSeries(pa.getName(), meanCount, !isCount);
			tempSeries.add(ts);
			for (Integer value : pa.getValues())
			{
				if (series.containsKey(value))
				{
					throw new BaseException("Value " + value + " has already been mapped to another series");
				}
				series.put(value, ts);
			}
		}
		float maxX = Float.MIN_VALUE;
		float minX = Float.MAX_VALUE;
		while (data.hasNext())
		{
			ThreadSignalHandler.checkInterrupted();
			SqlResult r = data.next();
			TempSeries ts = series.get(r.getInt(1));
			float x = r.getFloat(2);
			if (maxX < x) maxX = x;
			if (minX > x) minX = x;
			ts.x.add(x);
			if (!isCount) ts.y.add(r.getFloat(3));
		}
		
		for (TempSeries ts : tempSeries)
		{
			HistogramSeries main = new HistogramSeries(ts.name, ts.getX(), ts.getY(), binSize, yAggregate, minX, maxX);
			histogram.addSeries(main);
			if (hiloAggregate != null)
			{
				hilo.addSeries(new HistogramSeries(ts.name, main, hiloAggregate));
			}
			plot.getRenderer().setSeriesVisibleInLegend(histogram.getSeriesCount()-1, !ts.name.equals(""));
		}
		adjustRanges();
	}
	
	/**
	 	Get the underlying JFreeChar object. Use this method to polish up 
	 	everything and save the image.
	*/
	public JFreeChart getChart()
	{
		return chart;
	}
	
	/**
		Helper class for storing histogram data that needs to split into
		different annotations. For each annotation we create one TempSeries object
		and fills it with the data for that annotation.
	*/
	private static class TempSeries
	{
		final String name;
		final List<Float> x;
		final List<Float> y;
		
		TempSeries(String name, int size, boolean hasY)
		{
			this.name = name;
			this.x = new ArrayList<Float>(size);
			this.y = hasY ? new ArrayList<Float>(size) : null;
		}
		
		public float[] getX()
		{
			if (x == null) return null;
			float[] xf = new float[x.size()];
			int i = 0;
			for (Float f : x)
			{
				xf[i++] = f.floatValue();
			}
			return xf;
		}
		
		public float[] getY()
		{
			if (y == null) return null;
			float[] yf = new float[y.size()];
			int i = 0;
			for (Float f : y)
			{
				yf[i++] = f.floatValue();
			}
			return yf;
		}
		
	}
	
	/**
		An dataset implementation for histogram data that allows
		us to send data to JFreeChart the way we want. 
	*/
	private static class HistogramDataset
		extends AbstractSeriesDataset
		implements IntervalXYDataset
	{
		/**
			
		*/
		private static final long serialVersionUID = -515696571729019122L;
		private final List<HistogramSeries> all;
		
		private HistogramDataset()
		{
			this.all = new ArrayList<HistogramSeries>();
		}
		/*
			From the AbstractSeriesDataset class
			-------------------------------------------
		*/
		@Override
		public int getSeriesCount()
		{
			return all.size();
		}
		@SuppressWarnings("rawtypes")
		@Override
		public Comparable getSeriesKey(int series)
		{
			return all.get(series).getName();
		}
		// -------------------------------------------

		/*
			From the XYDataset interface
			-------------------------------------------
		*/
		@Override
		public DomainOrder getDomainOrder()
		{
			return DomainOrder.ASCENDING;
		}
		@Override
		public int getItemCount(int series)
		{
			return all.get(series).getItemCount();
		}
		@Override
		public Number getX(int series, int item)
		{
			return getXValue(series, item);
		}
		@Override
		public double getXValue(int series, int item)
		{
			HistogramBin bin = all.get(series).getBin(item);
			return (bin.getStartX() + bin.getEndX()) / 2;
		}

		@Override
		public Number getY(int series, int item)
		{
			return getYValue(series, item);
		}
		@Override
		public double getYValue(int series, int item)
		{
			return all.get(series).getMaxY(item);
		}
		// -------------------------------------------
		/*
			From the IntervalXYDataset interface
			-------------------------------------------
		*/
		@Override
		public Number getEndX(int series, int item)
		{
			return getEndXValue(series, item);
		}
		@Override
		public double getEndXValue(int series, int item)
		{
			return all.get(series).getBin(item).getEndX();
		}
		@Override
		public Number getEndY(int series, int item)
		{
			return getEndYValue(series, item);
		}
		@Override
		public double getEndYValue(int series, int item)
		{
			return getYValue(series, item);
		}
		@Override
		public Number getStartX(int series, int item)
		{
			return getStartXValue(series, item);
		}
		@Override
		public double getStartXValue(int series, int item)
		{
			return all.get(series).getBin(item).getStartX();
		}
		@Override
		public Number getStartY(int series, int item)
		{
			return getStartYValue(series, item);
		}
		@Override
		public double getStartYValue(int series, int item)
		{
			return all.get(series).getMinY(item);
		}
		// -------------------------------------------
		
		public void addSeries(HistogramSeries s)
		{
			all.add(s);
		}
	}

	/**
		Collects information about one histogram series. Multiple series may be
		added to {@link HistogramDataset} which is then added to JFreeChart.
	*/
	private static class HistogramSeries
		implements Serializable
	{

		private static final long serialVersionUID = 8540300441782223733L;
		private final String name;
		private final HistogramBin[] bins;
		private final YAggregate yAggregate;
		
		/**
			Create a new HistogramSeries. 
			@param name The name of the seris
			@param xValues The x values
			@param yValues The y values or null if no y values are used
			@param binSize The bin size on the x axis
			@param yAggregate What to show on the y axis
			@param xMin The minimum x axis value or {@link Float#NaN} to let this method calculate it
			@param xMax The maximum x axis value or {@link Float#NaN} to let this method calculate it
		 */
		private HistogramSeries(String name, float[] xValues, float[] yValues, float binSize, YAggregate yAggregate, float xMin, float xMax)
		{
			this.name = name;
			this.yAggregate = yAggregate;
			
			// Find max and min --> number of bins
			if (Float.isNaN(xMin) || Float.isNaN(xMax))
			{
				xMin = Float.MAX_VALUE;
				xMax = Float.MIN_VALUE;
				for (float x : xValues)
				{
					if (xMax < x) xMax = x;
					if (xMin > x) xMin = x;
				}
			}
			int numBins = 1 + (int)((xMax - xMin) / binSize);
			bins = new HistogramBin[numBins];
			
			// Create the bins
			for (int i = 0; i < numBins; ++i)
			{
				float startX = xMin + i * binSize;
				bins[i] = new HistogramBin(startX, startX + binSize);
			}
			
			// Add values to the bins
			for (int i = 0; i < xValues.length; ++i)
			{
				int bin = (int)((xValues[i] - xMin) / binSize);
				if (yValues == null)
				{
					bins[bin].increaseCount();
				}
				else
				{
					bins[bin].addMeasurement(yValues[i]);
				}
			}
		}
		
		/**
			Create a new HistogramSeries that is using data from another series.
			This is used by for example the high/low series which take it's data from
			the main series.

			@param name The name of the series
			@param other The other series
			@param yAggregate {@link YAggregate#HILOMAXMIN} or {@link YAggregate#HILOSTDEV}
		*/
		private HistogramSeries(String name, HistogramSeries other, YAggregate yAggregate)
		{
			this.name = name;
			this.bins = other.bins;
			this.yAggregate = yAggregate;
		}
		
		public String getName()
		{
			return name;
		}
		
		public int getItemCount()
		{
			return bins.length;
		}
		
		public HistogramBin getBin(int index)
		{
			return bins[index];
		}
		
		public float getMaxY(int index)
		{
			return yAggregate.getMaxY(bins[index]);
		}
		public float getMinY(int index)
		{
			return yAggregate.getMinY(bins[index]);
		}
	}
	
	/**
		Represents a single bin in a histogram.
	*/
	private static class HistogramBin
		implements Serializable
	{

		private static final long serialVersionUID = 513573632731346456L;
		private final double startX;
		private final double endX;
		
		private int count = 0;
		private float sum = 0.0f;
		private float squaredSum = 0.0f;
		private float min = Float.MAX_VALUE;
		private float max = Float.MIN_VALUE;
		
		/**
			Create a new HistogramBin.
			@param startX The min x-axis values for this bin
			@param endX The max x-axis value for this bin
		 */
		private HistogramBin(float startX, float endX)
		{
			this.startX = startX;
			this.endX = endX;
		}
		
		public double getStartX()
		{
			return startX;
		}
		
		public double getEndX()
		{
			return endX;
		}
		
		/**
			Increase the count value in this bin. Used if plotting
			{@link YAggregate#COUNT} values.
		*/
		public void increaseCount()
		{
			count++;
		}
		
		/**
			Add a measurement to this bin. Used if plotting values other than 
			{@link YAggregate#COUNT}. This method increases the count and updates
			the sum, max and min values as required.
			@param y The y value to add
		*/
		public void addMeasurement(float y)
		{
			count++;
			sum += y;
			squaredSum += y * y;
			if (y < min) min = y;
			if (y > max) max = y;
		}
		
		/**
		 	Get the number of values in this bin.
		*/
		public float getCount()
		{
			return count;
		}
		
		/**
		 	Get the sum of all values in this bin. The sum is not available if 
		 	plotting the {@link YAggregate#COUNT}.
		*/
		public float getSum()
		{
			return sum;
		}
		
		/**
		 	Get the mean of all values in this bin. The mean is not available if 
		 	plotting the {@link YAggregate#COUNT}.
		*/
		public float getMean()
		{
			return count == 0 ? 0.0f : sum / count;
		}
		
		/**
		 	Get the minumum of all values in this bin. The value is not available if 
		 	plotting the {@link YAggregate#COUNT}.
		*/
		public float getMin()
		{
			return min;
		}
		/**
		 	Get the maximum of all values in this bin. The value is not available if 
		 	plotting the {@link YAggregate#COUNT}.
		*/
		public float getMax()
		{
			return max;
		}
		/**
		 	Get the standard deviation of all values in this bin. The value is not available if 
		 	plotting the {@link YAggregate#COUNT}.
		*/
		public float getStdev()
		{
			return (float)Math.sqrt((squaredSum - sum * sum / count) / (count - 1));
		}
	}
	
	/**
	 	Defines what to to plot on the y axis in a histogram plot.
	*/
	public enum YAggregate
	{
		/**
			Plot the number of spots falling into each bin.
		*/
		COUNT
		{
			@Override
			public float getMaxY(HistogramBin bin)
			{
				return bin.getCount();
			}
		},
		/**
			Plot the sum of all y-values for spots falling into each bin.
		*/
		SUM
		{
			@Override
			public float getMaxY(HistogramBin bin)
			{
				return bin.getSum();
			}
		},
		/**
			Plot the mean of all y-values for spots falling into each bin.
		*/
		MEAN
		{
			@Override
			public float getMaxY(HistogramBin bin)
			{
				return bin.getMean();
			}
		},
		/**
			Plot the maximum of all y-values for spots falling into each bin.
		*/
		MAX
		{
			@Override
			public float getMaxY(HistogramBin bin)
			{
				return bin.getMax();
			}
		},
		/**
			Plot the minimum of all y-values for spots falling into each bin.
		*/
		MIN
		{
			@Override
			public float getMaxY(HistogramBin bin)
			{
				return bin.getMin();
			}
		},
		/**
			Plot the standard deviation of all y-values for spots falling into each bin.
		*/
		STDEV
		{
			@Override
			public float getMaxY(HistogramBin bin)
			{
				return bin.getStdev();
			}
			@Override
			public float getMinY(HistogramBin bin)
			{
				return bin.getMean() - bin.getStdev();
			}
		},
		/**
			Plot a high/low bar for each bin with the mean +/- 1 standard deviation.
		*/
		HILOSTDEV
		{
			@Override
			public float getMaxY(HistogramBin bin)
			{
				return bin.getMean() + bin.getStdev();
			}
			@Override
			public float getMinY(HistogramBin bin)
			{
				return bin.getMean() - bin.getStdev();
			}
			
		},
		/**
			Plot a high/low bar for each bin with the maximum and minimum values.
		*/
		HILOMAXMIN
		{
			@Override
			public float getMaxY(HistogramBin bin)
			{
				return bin.getMax();
			}
			@Override
			public float getMinY(HistogramBin bin)
			{
				return bin.getMin();
			}
		}
		;
		
		public abstract float getMaxY(HistogramBin bin);
		public float getMinY(HistogramBin bin)
		{
			return 0;
		}
		
	}
	
}
