/**
	$Id: AbstractItemImporter.java 7203 2016-10-18 16:30:54Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.batchimport;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

import net.sf.basedb.core.AnnotationTypeCategory;
import net.sf.basedb.core.ArrayBatch;
import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.ArraySlide;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.BioPlate;
import net.sf.basedb.core.BioPlateType;
import net.sf.basedb.core.BioSource;
import net.sf.basedb.core.BioWell;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DerivedBioAssay;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileServer;
import net.sf.basedb.core.FileSet;
import net.sf.basedb.core.FileSetMember;
import net.sf.basedb.core.FileStoreEnabled;
import net.sf.basedb.core.Hardware;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.Nameable;
import net.sf.basedb.core.PhysicalBioAssay;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.IntegerUtil;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemAlreadyExistsException;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.ItemKey;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Kit;
import net.sf.basedb.core.NumberOutOfRangeException;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.PermissionTemplate;
import net.sf.basedb.core.PlateGeometry;
import net.sf.basedb.core.Platform;
import net.sf.basedb.core.PlatformVariant;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.Project;
import net.sf.basedb.core.ProjectKey;
import net.sf.basedb.core.Protocol;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.RawDataTypes;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.Sample;
import net.sf.basedb.core.Shareable;
import net.sf.basedb.core.ShareableUtil;
import net.sf.basedb.core.Software;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.StringTooLongException;
import net.sf.basedb.core.StringUtil;
import net.sf.basedb.core.Subtypable;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.core.Tag;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Unit;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.ParameterValues;
import net.sf.basedb.core.plugin.ParameterValuesWrapper;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.plugins.AbstractFlatFileImporter;
import net.sf.basedb.plugins.util.Parameters;
import net.sf.basedb.util.Coordinate;
import net.sf.basedb.util.Enumeration;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.error.ClassMapErrorHandler;
import net.sf.basedb.util.error.ErrorHandler;
import net.sf.basedb.util.error.SimpleErrorHandler;
import net.sf.basedb.util.formatter.DateFormatter;
import net.sf.basedb.util.parser.ConfigureByExample;
import net.sf.basedb.util.parser.DefaultItemFunction;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;
import net.sf.basedb.util.parser.FlatFileParser.Data;

/**
	An abstract base class for plug-ins that import items in batch. This class implements
	job configuration, file parsing, logging, error handling, etc. Actual implementations
	have to provide item-specific parts, such as:
	<ul>
	<li>The context the plug-in can be used = the type of item it can import.
		Eg, implement the {@link InteractivePlugin#getGuiContexts()} method.
	<li>The methods that can be used for item identification: 
		{@link #getIdMethods()}
	<li>Job configuration parameters for column mappings to item-specific
		properties: {@link #addMoreColumnMappingParameters(List)}
	<li>Query used to find items of the specific type: {@link #createItemQuery()}
	<li>Create actual column mappers used to extract data from the file
		when running the plug-in: {@link #createColumnMappers(FlatFileParser, boolean)}
	<li>Create a new item and populate it with data from the file: 
		{@link #createItem(DbControl, FlatFileParser.Data)}
	<li>Update an item with data from the file: {@link #updateItem(DbControl, BasicItem, FlatFileParser.Data)}
		and {@link #updateMultiLineItem(DbControl, BasicItem, FlatFileParser.Data, int)}
	</ul>

	@author nicklas
	@version 2.8
	@base.modified $Date: 2016-10-18 18:30:54 +0200 (ti, 18 okt 2016) $
*/
public abstract class AbstractItemImporter<I extends BasicItem>
	extends AbstractFlatFileImporter
	implements InteractivePlugin, ConfigureByExample
{
	
	// second step in the wizard
	private static final String CONFIGURE_OPTIONS = "configure_options";
	
	protected static final StringParameterType optionalColumnMapping = new StringParameterType(255, null, false);
	protected static final StringParameterType requiredColumnMapping = new StringParameterType(255, null, true);

	protected static final PluginParameter<String> defaultsSection = new PluginParameter<String>(
			"defaultsSection",
			"Default settings",
			"Set default values for some runtime parameters.", 
			null
		);
	
	protected static final PluginParameter<String> identificationSection = new PluginParameter<String>(
			"idSection",
			"Item identification",
			"Options related to how find existing items.", 
			null
		);
	
	protected static final PluginParameter<String> internalIdColumnMapping = new PluginParameter<String>(
			"internalIdColumnMapping",
			"Internal ID",
			"Mapping that picks the BASE internal ID of the item from the data columns. This " +
			"parameter is required if the 'Internal ID' identification method is selected, " +
			"otherwise it is optional. Example: \\BASE ID\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> nameColumnMapping = new PluginParameter<String>(
			"nameColumnMapping",
			"Name",
			"Mapping that picks the name of the item from the data columns. This " +
			"parameter is required if the 'Name' identification method is selected, " +
			"otherwise it is optional. Example: \\Name\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> subtypeColumnMapping = new PluginParameter<String>(
			"subtypeColumnMapping",
			"Item subtype",
			"Mapping that picks the subtype of the item from the data columns. " +
			"Example: \\Subtype\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> descriptionColumnMapping = new PluginParameter<String>(
			"descriptionColumnMapping",
			"Description",
			"Mapping that picks the description of the item from the data columns." +
			"Example: \\Description\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> externalIdColumnMapping = new PluginParameter<String>(
			"externalIdColumnMapping",
			"External ID",
			"Mapping that picks the external ID of the item from the data columns. This " +
			"parameter is required if the 'External ID' identification method is selected, " +
			"otherwise it is optional. Example: \\External ID\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> protocolColumnMapping = new PluginParameter<String>(
			"protocolColumnMapping",
			"Protocol",
			"Mapping that picks the name, external or internal ID of a protocol from the data columns. The plug-in will first " +
			"try to find a protocol with the given name or external id. If none is found and the value is numeric it will " +
			"try to load by internal ID. " +
			"Example: \\Protocol\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> kitColumnMapping = new PluginParameter<String>(
			"kitColumnMapping",
			"Kit",
			"Mapping that picks the name or internal ID of a kit from the data columns. The plug-in will first " +
			"try to find a kit with the given name. If none is found and the value is numeric it will " +
			"try to load by internal ID. " +
			"Example: \\Kit\\",
			optionalColumnMapping
			);

	protected static final PluginParameter<String> permissionTemplateColumnMapping = new PluginParameter<String>(
			"permissionTemplateColumnMapping",
			"Permissions",
			"Mapping that picks the name or ID of a premission template that should be used " +
			"to set access permissions on the item. Current permissions on already existing items " +
			"are merged with the permission from the template. " +
			"The plug-in will first try to find a template with the given name. If none is found and " +
			"the value is numeric it will try to load by internal ID. " +
			"Example: \\Permissions\\",
			optionalColumnMapping
			);
	
	private static final PluginParameter<String> itemNotFoundErrorParameter = new PluginParameter<String>(
			"itemNotFoundError",
			"Item not found",
			"What to do if the plug-in can't find an item with the name or ID from the file.\n\n"+
			"skip = Skip the current data line and continue\n"+
			"fail = Stop with an error message",
			new StringParameterType(255, null, false, 1, 0, 0, 
				Arrays.asList( new String[] { "skip", "fail"} ))
		);
	
	private static final PluginParameter<String> itemExistsErrorParameter = new PluginParameter<String>(
			"itemExistsError",
			"Item already exists",
			"What to do if the plug-in finds an existing item with the name or ID from the file.\n\n"+
			"skip = Skip the current data line and continue\n"+
			"fail = Stop with an error message",
			new StringParameterType(255, null, false, 1, 0, 0, 
				Arrays.asList( new String[] { "skip", "fail"} ))
		);

	
	private static final PluginParameter<String> multipleItemsFoundErrorParameter = new PluginParameter<String>(
			"multipleItemsFoundError",
			"Multiple items found",
			"What to do if the plug-in finds more than one item with the same name or ID.\n\n"+
			"skip = Skip the current data line and continue\n"+
			"fail = Stop with an error message",
			new StringParameterType(255, null, false, 1, 0, 0, 
				Arrays.asList( new String[] { "skip", "fail"} ))
		);

	
	private static final PluginParameter<String> referenceNotFoundErrorParameter = new PluginParameter<String>(
			"referenceNotFoundError",
			"Referenced item not found",
			"What to do if the plug-in can't find a referenced item, eg. a protocol, etc.\n\n"+
			"ignore = Do not set/update the reference\n" +
			"fail = Stop with an error message",
			new StringParameterType(255, null, false, 1, 0, 0, 
				Arrays.asList( new String[] { "ignore", "fail"} ))
		);

	private static final PluginParameter<String> multipleReferencesFoundErrorParameter = new PluginParameter<String>(
			"multipleReferencesFoundError",
			"Multiple referenced items found",
			"What to do if the plug-in finds more than one referenced item (eg. protocol) with the same name or ID.\n\n"+
			"ignore = Do not set/update the reference\n" +
			"fail = Stop with an error message",
			new StringParameterType(255, null, false, 1, 0, 0, 
				Arrays.asList( new String[] { "ignore", "fail"} ))
		);
	
	private static final PluginParameter<String> noWritePermissionToItemErrorParameter = new PluginParameter<String>(
			"noWritePermissionToItemError",
			"Not granted write permission to item",
			"What to do if the plug-in finds an existing item but has no write permission to it.\n\n"+
			"skip = Skip the current data line and continue.\n"+
			"fail = Stop with an error message",
			new StringParameterType(255, null, false, 1, 0, 0, 
				Arrays.asList( new String[] { "skip", "fail"} ))
		);
	
	private static final PluginParameter<String> noUsePermissionToReferenceErrorParameter = new PluginParameter<String>(
			"noUsePermissionToReferenceError",
			"Not granted use permission to reference",
			"What to do if the plug-in finds a reference but has no permission to use it.\n\n"+
			"ignore = Do not set/update the reference\n" +
			"fail = Stop with an error message",
			new StringParameterType(255, null, false, 1, 0, 0, 
				Arrays.asList( new String[] { "ignore", "fail"} ))
		);
	
	public AbstractItemImporter()
	{}
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	@Override
	public boolean requiresConfiguration()
	{
		return false;
	}
	@Override
	public boolean supportsConfigurations()
	{
		return true;
	}
	// --------------------------------------------
	/*
		From the InteractivePlugin interface
		------------------------------------
	*/
	@Override
	public String isInContext(GuiContext context, Object item)
	{
		String message = getGuiContexts().contains(context) ? 
			null : "This plug-in can't be used in context: " + context;
		return message;
	}
	
	@Override
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN) || command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				boolean forJob = command.equals(Request.COMMAND_CONFIGURE_JOB);
				RequestInformation ri = getConfigureParserParameters(context, true);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}

				ParameterValues wrapper = new ParameterValuesWrapper(request, job, configuration, forJob);
				if (forJob)
				{
					// Check that file can be parsed with given regular expressions
					dc = sc.newDbControl();
					FlatFileParser ffp = getInitializedFlatFileParser(wrapper);
					File f = File.getById(dc, ((File)request.getParameterValue("file")).getId());
					InputStream in = f.getDownloadStream(0);
					ffp.setInputStream(in, getCharset(request));
					ffp.setDefaultNumberFormat(getNumberFormat());
					FlatFileParser.LineType result = ffp.parseHeaders();
					in.close();
					if (result == FlatFileParser.LineType.UNKNOWN)
					{
						response.setError("The file could not be parsed with the given settings, " +
							"no data found after " + ffp.getLineCount() + " lines.", null);
						return;
					}
				}
				
				// Check that a mapping has been provided for the selected idMethod
				IdMethod idMethod = getIdMethod((String)request.getParameterValue("idMethod"));
				if (idMethod == null)
				{
					response.setError("Unknown method for item identification: " + 
							request.getParameterValue("idMethod"), null);
					return;
				}
				PluginParameter<?> idMapping = ri.getParameter(idMethod.getColumnMappingParameterName());
				if (request.getParameterValue(idMapping.getName()) == null)
				{
					response.setError("No colum mapping provided for item identification: " + 
							idMapping.getLabel(), null);
					return;
				}

				// Check the mapping expressions
				boolean allowComplex = "allow".equals(request.getParameterValue(complexMappings.getName()));
				for (PluginParameter<?> pp : ri.getParameters())
				{
					if (pp.getName().contains("ColumnMapping"))
					{
						checkColumnMapping((String)request.getParameterValue(pp.getName()), allowComplex, pp.getLabel());
					}
				}
				
				// Everything is ok, save values
				for (PluginParameter<?> pp : ri.getParameters())
				{
					if (pp.getParameterType() != null)
					{
						storeValue(wrapper, request, ri.getParameter(pp.getName()));
					}
				}
				if (forJob)
				{
					response.setContinue(CONFIGURE_OPTIONS);
				}
				else
				{
					response.setDone("Plug-in configuration completed");
				}
			}
			else if (command.equals(CONFIGURE_OPTIONS))
			{
				RequestInformation ri = getConfigureOptionsParameters(context);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}
			
				// Error handling
				response.setDryRun(Boolean.TRUE.equals(request.getParameterValue(Parameters.DRY_RUN_PARAMETER)));
				for (PluginParameter<?> pp : ri.getParameters())
				{
					if (pp.getParameterType() != null)
					{
						storeValue(job, request, ri.getParameter(pp.getName()));
					}
				}

				File file = (File)job.getValue("file");
				response.setSuggestedJobName("Import " + context.getItem().toString().toLowerCase() 
						+ " items from file '" + file.getName() + "'");
				
				// Save context info
				response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);		
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}	
	
	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
		{
			requestInformation = getConfigureParserParameters(context, false);
		}
		else if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureParserParameters(context, true);
		}
		else if (command.equals(CONFIGURE_OPTIONS))
		{
			requestInformation = getConfigureOptionsParameters(context);
		}
		return requestInformation;
	}
	// ------------------------------------
	/*
		From the AbstractFlatFileImporter class
		---------------------------------------
	*/
	private DbControl dc;
	private Project activeProject;
	private FlatFileParser ffp;
	private DateFormatter dateFormatter;
	
	private IdMethod idMethod;
	private String idMapping;
	private Mapper idMapper;
	private Mapper permissionTemplateMapper;
	private Mapper subtypeMapper;
	private ItemSubtype selectedSubtype;
	private ItemQuery<I> itemQuery;

	private String lastIdentifier;
	private I lastItem;
	private int numMultiLines;
	private boolean skippedLast;
	
	// Error handling
	private boolean dryRun;
	private ClassMapErrorHandler errorHandler;
	private boolean cropStrings;
	private boolean createNotFoundItems;
	private boolean failIfNotFoundItems;
	private boolean updateExistingItems;
	private boolean failIfItemExists;
	private boolean failIfMultipleFoundItems;
	private boolean failIfNotFoundReference;
	private boolean failIfMultipleFoundReferences;
	private boolean failIfNoWritePermissionToItem;
	private boolean failIfNoUsePermissionToReference;
	
	// Status report
	private int numUpdated;
	private int numCreated;
	private int numError;
	private int numNotFound;
	private int numExisted;
	
	@Override
	/**
		The flat file parser is initialised with settings from either the job or configuration.
	*/
	protected FlatFileParser getInitializedFlatFileParser()
		throws BaseException
	{
		ParameterValuesWrapper wrapper = new ParameterValuesWrapper(null, job, configuration, true);
		return getInitializedFlatFileParser(wrapper);
	}
	
	/**
		Don't use AbstractFlatFileImporter to handle errors.
	*/
	@Override
	protected void setUpErrorHandling()
	{}
	
	/**
		Setup error handling and pre-load some of the configuration options.
	*/
	@Override
	protected void begin(FlatFileParser ffp)
	{
		this.ffp = ffp;
		String dateFormat = Values.getString((String)job.getValue(Parameters.DATE_FORMAT_PARAMETER), "yyyy-MM-dd");
		this.dateFormatter = new DateFormatter(dateFormat);
		this.dateFormatter.getDateFormat().setLenient(false);
		this.idMethod = getIdMethod((String)job.getValue("idMethod"));
		this.idMapping = (String)job.getValue(idMethod.getColumnMappingParameterName());
		this.itemCache = new HashMap<String, BasicItem>();
		String mode = (String)job.getValue("mode");
		boolean updateMode = mode == null || mode.contains("update");
		boolean createMode = mode == null || mode.contains("create");
		
		// Error handling, we don't use the error handling in AbstractFlatFileImporter
		this.dryRun = job.getJob().isDryRun();
		this.createNotFoundItems = createMode || "create".equals(getErrorOption("itemNotFoundError"));
		this.failIfNotFoundItems = !createMode && "fail".equals(getErrorOption("itemNotFoundError"));
		this.updateExistingItems = updateMode || "update".equals(getErrorOption("itemExistsError"));
		this.failIfItemExists = !updateMode && "fail".equals(getErrorOption("itemExistsError"));
		this.failIfMultipleFoundItems = "fail".equals(getErrorOption("multipleItemsFoundError"));
		this.failIfNotFoundReference = "fail".equals(getErrorOption("referenceNotFoundError"));
		this.failIfMultipleFoundReferences = "fail".equals(getErrorOption("multipleReferencesFoundError"));
		this.failIfNoWritePermissionToItem = "fail".equals(getErrorOption("noWritePermissionToItemError"));
		this.failIfNoUsePermissionToReference = "fail".equals(getErrorOption("noUsePermissionToReferenceError"));
		
		this.cropStrings = "crop".equals(getErrorOption("stringTooLongError"));

		ErrorHandler defaultErrorHandler = new SimpleErrorHandler("skip".equals(job.getValue("defaultError")));
		errorHandler = new ClassMapErrorHandler(defaultErrorHandler);
		errorHandler.addErrorHandler(StringTooLongException.class, 
				getErrorHandler((String)job.getValue("stringTooLongError"), defaultErrorHandler));
		errorHandler.addErrorHandler(NumberOutOfRangeException.class, 
				getErrorHandler((String)job.getValue("numberOutOfRangeError"), defaultErrorHandler));
		errorHandler.addErrorHandler(NumberFormatException.class,
				getErrorHandler((String)job.getValue("numberFormatError"), defaultErrorHandler));
		errorHandler.addErrorHandler(InvalidUseOfNullException.class, 
				getErrorHandler((String)job.getValue(Parameters.INVALID_USE_OF_NULL_ERROR), defaultErrorHandler));

	}
	
	/**
		Setup column mapping. Creates DbControl and query to find items.
	*/
	@SuppressWarnings("unchecked")
	@Override
	protected void beginData()
	{
		this.dc = sc.newDbControl();
		if (sc.getActiveProjectId() != 0) 
		{
			activeProject = Project.getById(dc, sc.getActiveProjectId());
		}
		
		// Mapper to get name or external ID
		this.idMapper = getMapper(ffp, idMapping, null, null);
		this.permissionTemplateMapper = getMapper(ffp, (String)job.getValue("permissionTemplateColumnMapping"), null, null);
		this.subtypeMapper = getMapper(ffp, (String)job.getValue("subtypeColumnMapping"), null, null);
		createColumnMappers(ffp, cropStrings);
		
		// Include options
		Set<Include> includes = EnumSet.noneOf(Include.class);
		if (Boolean.TRUE.equals((Boolean)job.getValue("includeMine"))) includes.add(Include.MINE);
		if (Boolean.TRUE.equals((Boolean)job.getValue("includeShared"))) includes.add(Include.SHARED);
		if (Boolean.TRUE.equals((Boolean)job.getValue("includeInProject"))) includes.add(Include.IN_PROJECT);
		if (Boolean.TRUE.equals((Boolean)job.getValue("includeOthers"))) includes.add(Include.OTHERS);
		itemQuery = idMethod.prepareQuery(dc, createItemQuery());
		itemQuery.setIncludes(includes);
		
		List<ItemSubtype> subtypes = (List<ItemSubtype>)job.getValues("itemSubtypes");
		if (subtypes != null && subtypes.size() > 0)
		{
			Expression subtypeParameter = Hql.entityParameter("subtypes", Item.ITEMSUBTYPE);
			if (subtypes.size() == 1)
			{
				selectedSubtype = ItemSubtype.getById(dc, subtypes.get(0).getId());
				itemQuery.restrict(Restrictions.eq(Hql.property("itemSubtype"), subtypeParameter));
				itemQuery.setEntityParameter("subtypes", selectedSubtype);
			}
			else
			{
				itemQuery.restrict(Restrictions.in(Hql.property("itemSubtype"), subtypeParameter));
				itemQuery.setParameter("subtypes", subtypes, null);
			}
		}
	}
	
	@SuppressWarnings("unchecked")
	@Override
	protected void handleData(Data data)
		throws BaseException
	{
		// Find the item(s) to update. Mapper gives us name, internal ID, or some other identifer
		String identifier = idMapper.getValue(data);
		String cacheKey = itemQuery.getItemType().name() + ":" + identifier;
		boolean sameAsLast = StringUtil.isEqualOrNull(identifier, lastIdentifier);
		if (!sameAsLast)
		{
			if (lastItem != null) doneWithItem(dc, lastItem, numMultiLines);
			numMultiLines = 0;
			lastIdentifier = identifier;
		}
		else
		{
			numMultiLines++;
			if (skippedLast)
			{
				// If the first line in a multi-line entry was skipped, skip the remaining lines also
				log("Skipped: " + idMethod + "=" + identifier, data);
				return;
			}
		}
		lastItem = null;
		I item = null;
		if (identifier != null)
		{
			// Try to find an existing item
			// We look in the cache first
			if (itemCache.containsKey(cacheKey))
			{
				item = (I)itemCache.get(cacheKey);
			}
			else
			{
				// and only query the database if not found
				List<I> items = idMethod.find(dc, itemQuery, identifier);
				// pick out items on which write permission is granted
				List<I> writableItems = new ArrayList<I>();
				if (updateExistingItems)
				{
					// Only check for WRITE permission if we are going to update
					for (I i : items)
					{
						if (i.hasPermission(Permission.WRITE))
						{
							writableItems.add(i);
						}
					}
				}
				else
				{
					// Pretend we have WRITE permission; we handle the !updateExistingItems later
					writableItems.addAll(items);
				}
				if (writableItems.size() == 1)
				{
					item = writableItems.get(0);
					itemCache.put(cacheKey, item);
				}
				else if (writableItems.size() > 1)
				{
					// If we find multiple items, we either throw an error or do nothing
					if (failIfMultipleFoundItems)
					{
						throw new BaseException("Found " + writableItems.size() + " items [" + idMethod + "=" + identifier + "]");
					}
					log("Found " + writableItems.size() + " items: " + idMethod + "=" + identifier, data);
					return;
				}
				else if (items.size() > 0) 
				{
					// Found item(s) but none with write permission to. Either throw an exception or do nothing
					if (failIfNoWritePermissionToItem)
					{
						throw new PermissionDeniedException(Permission.WRITE, 
								itemQuery.getItemType()+"[" + idMethod + "=" + identifier + "]");
					}
					log("Not granted write permission to " + itemQuery.getItemType()+"[" + idMethod + "=" + identifier + "]", data);
					return;
				}
			}
		}

		if (item == null)
		{
			// No item was found
			if (failIfNotFoundItems)
			{
				throw new ItemNotFoundException(itemQuery.getItemType()+"[" + idMethod + "=" + identifier + "]");
			}
			else if (createNotFoundItems)
			{
				// Create a new item
				try
				{
					item = createItem(dc, data);
					if (permissionTemplateMapper != null) updatePermissions(dc, (Shareable)item, data, true);
					dc.saveItem(item);
					itemCache.put(cacheKey, item);
					lastItem = item;
					log("Created: " + idMethod + "=" + identifier, data);
					skippedLast = false;
					numCreated++;
				}
				catch (Throwable t)
				{
					try
					{
						errorHandler.handleError(t);
						log("Skipped: " + idMethod + "=" + identifier, data, t);
						skippedLast = true;
						numError++;
						// The error should be ignored if we get passed the above line
					}
					catch (Throwable t2)
					{
						// handleData doesn't allow us to throw any Throwable
						throw t2 instanceof RuntimeException ?
							(RuntimeException)t2 : new BaseException(t);
					}
				}
			}
			else
			{
				log("Not found: " + idMethod + "=" + identifier, data);
				skippedLast = true;
				numNotFound++;
			}
		}
		else
		{
			// Found a single item
			itemCache.put(cacheKey, item);
			lastItem = item;
			if (failIfItemExists && !sameAsLast)
			{
				// Fail if this is the first time we see this item
				throw new ItemAlreadyExistsException(itemQuery.getItemType()+"[" + idMethod + "=" + identifier + "]");
			}
			else if (updateExistingItems || (sameAsLast && !item.isInDatabase()))
			{
				// Update existing items or second line of new items in 'create' mode
				try
				{
					if (sameAsLast)
					{
						updateMultiLineItem(dc, item, data, numMultiLines);
					}
					else
					{
						updateItem(dc, item, data);
						if (permissionTemplateMapper != null) updatePermissions(dc, (Shareable)item, data, true);
						numUpdated++;
					}
					log("Updated: " + idMethod + "=" + identifier, data);
					skippedLast = false;
				}
				catch (Throwable t)
				{
					try
					{
						errorHandler.handleError(t);
						log("Skipped: " + idMethod + "=" + identifier, data);
						skippedLast = true;
						numError++;
						// The error should be ignored if we get passed the above line
					}
					catch (Throwable t2)
					{
						// handleData doesn't allow us to throw any Throwable
						throw t2 instanceof RuntimeException ?
								(RuntimeException)t2 : new BaseException(t);
					}
				}
			}
			else
			{
				// Ignore existing items
				log("Exists: " + idMethod + "=" + identifier, data);
				skippedLast = true;
				numExisted++;
			}
		}
	}
	
	@Override
	protected void end(boolean success)
	{
		try
		{
			if (success)
			{
				if (lastItem != null) doneWithItem(dc, lastItem, numMultiLines);
				if (!dryRun) dc.commit();
			}
		}
		catch (BaseException ex)
		{
			success = false;
			throw ex;
		}
		finally
		{
			if (dc != null) dc.close();
			super.end(success);
		}
	}
	
	@Override
	protected String getSuccessMessage(int skippedLines)
	{
		String msg = "Parsed " + ffp.getParsedLines() + " lines";
		if (numCreated > 0) msg += "; " + numCreated + " items created";
		if (numUpdated > 0) msg += "; " + numUpdated + " items updated";
		if (numExisted > 0) msg += "; " + numExisted + " items skipped because the already exists";
		if (numError > 0) msg += "; " + numError + " items skipped because of invalid values";
		if (numNotFound > 0) msg += "; " + numNotFound + " items skipped because they could not be found";
		return msg;
	}
	
	// ------------------------------------

	/*
		Methods to be implemented by subclasses
		---------------------------------------
	*/
	/**
		Get an array with all identification methods that can be used to find
		items. The default implementation returns an array with the {@link
		PropertyIdMethod#NAME}, {@link InternalIdMethod#INTERNAL_ID} and 
		{@link FallbackIdMethod#NAME_OR_ID} ID methods. Subclasses that import 
		items that can use other ID methods, should override this method.
	*/
	protected IdMethod[] getIdMethods()
	{
		return new IdMethod[] { PropertyIdMethod.NAME, InternalIdMethod.INTERNAL_ID, FallbackIdMethod.NAME_OR_ID };
	}
	
	/**
		Add extra job parameters that the subclass needs. This
		method is called during configuration of the plug-in and
		the default implementation does nothing.
		
		@param parameters The list of parameters to add the column
			mapping parameters to
	*/
	protected void addMoreJobParameters(List<PluginParameter<?>> parameters)
	{}
	
	/**
		Add extra parameters to the "Parser settings" section that the subclass 
		needs. This method is called during configuration of the plug-in and
		the default implementation does nothing.
		
		@param parameters The list of parameters to add the column
			mapping parameters to
	*/
	protected void addMoreParsingParameters(List<PluginParameter<?>> parameters)
	{}

	/**
		Add column mapping parameters for item-specific properties. A subclass must
		add parameters for all properties, including name and description. This class
		has already created some useful parameters for this purpose, for example:
		{@link #nameColumnMapping} and {@link #descriptionColumnMapping}. This
		method is called during configuration of the plug-in.
		
		@param parameters The list of parameters to add the column
			mapping parameters to
	*/
	protected void addMoreColumnMappingParameters(List<PluginParameter<?>> parameters)
	{}
	
	/**
		Add extra parameters to the "Error handling" section that the subclass 
		needs. This method is called during configuration of the plug-in and
		the default implementation does nothing.
		
		@param parameters The list of parameters to add the column
			mapping parameters to
	*/
	protected void addMoreErrorParameters(List<PluginParameter<?>> parameters)
	{}
	
	/**
		Create a new query that returns items of the importable type. The
		query should be a plain query not including any restrictions or 
		other elements. This will automatically be added as needed by the
		core implementation. This method is called during execution of the 
		plug-in. Typically, the result from <code>getQuery()</code> should
		be returned, eg. <code>Sample.getQuery()</code>.
		@return A query
	*/
	protected abstract ItemQuery<I> createItemQuery();

	/**
		Does the importer want to use item subtypes when importing?
		If this method returns a non-null, this class will add a parameter
		in the "Item identification" section that makes it possible to
		select which subtypes the importer should use to limit the 
		search for existing items. The default implementation returns null.
		@return An Item or null
	*/
	protected Item getItemForSubtypes()
	{
		return null;
	}
	
	/**
		Create mappers that can be used to extract data values from the file
		that is currently being parsed by the <code>ffp</code> {@link FlatFileParser}.
		This method is called during initialisation of the plug-in when the 
		{@link FlatFileParser#parseHeaders()} has successfully found the start of 
		data. Use the {@link #getMapper(FlatFileParser, String, Integer, Mapper)}
		to create column mappers. Example:
		
		<pre class="code">
nameMapper = getMapper(ffp, job.getValue("nameColumnMapping"), 
   cropStrings ? Nameable.MAX_NAME_LENGTH : null, null);
</pre>
		
		
		@param ffp The parser that is parsing the data file
		@param cropStrings If strings that are too long to fit in the database
			should be cropped or not (=generate an error)
	*/
	protected void createColumnMappers(FlatFileParser ffp, boolean cropStrings)
	{}

	/**
		Create a new item and populate it with data from the current line. Do
		not call {@link DbControl#saveItem(BasicItem)}. This is done automatically
		by the core implementation. 
		
		<pre class="code">
BioSource bs = BioSource.getNew(dc);
// Set properties. See example on updateItem method
return bs;
</pre>
		<p>
		If the current data line doesn't have data for all required 
		properties, this method should throw an {@link InvalidUseOfNullException}
		exception.
		
		<p>
		This method is called during import when current item can be found
		and if the user has selected to create missing item. 
		
		@param dc The DbControl to use for database access
		@param data The current data line
		@return A new item
	*/
	protected abstract I createItem(DbControl dc, FlatFileParser.Data data);
	
	/**
		Update an item with data from the current line. Example:
		<pre class="code">
if (nameMapper != null) item.setName(nameMapper.getValue(data));
if (descriptionMapper != null) item.setDescription(descriptionMapper.getValue(data));
...
</pre>

		This method is called during import to update an existing item.

		@param dc The DbControl to use for database access
		@param item The item to update
		@param data The current data line
	*/
	protected abstract void updateItem(DbControl dc, I item, FlatFileParser.Data data);
	
	/**
		Update an item with data from a multi-line entry. Items which can have multiple
		parents should have data for the other parents on lines following
		the first line.
		
		@param dc The DbControl to use for database access
		@param item The item to update
		@param data The current data line
		@param multiLineNum The number of the multi-line entry, starting with 1
			for the first multi-line
	*/
	protected void updateMultiLineItem(DbControl dc, I item, FlatFileParser.Data data, int multiLineNum)
	{}
	
	/**
		Called when the importer is finished with an item. This method is useful when working
		with multi-line entries, when the subclass needs to do some additional work after
		all data has been set. NOTE! This assumes that the lines in the data file are sorted
		in the correct order and the no lines for the same item will follow later on.
		@param dc The DbControl to use for database access
		@param item The item to update
		@param numMultiLines The number of the multi-line entries, starting with 1
			for the first multi-line (0 for single-line entries)
	 */
	protected void doneWithItem(DbControl dc, I item, int numMultiLines)
	{}
	
	/**
		If a mapping for a permission template has been specified for the job, this
		method is automatically called. It will merge or replace the permissions on the
		item with the permissions from the template.
		@param dc A DbControl used to lookup the permission template
		@param item The item to update
		@param data Data from the file
		@param merge TRUE to merge the permissions, FALSE to replace the permissions
		@since 2.16
	*/
	protected void updatePermissions(DbControl dc, Shareable item, FlatFileParser.Data data, boolean merge)
	{
		String nameOrId = permissionTemplateMapper.getValue(data);
		PermissionTemplate template = findPermissionTemplate(dc, FallbackIdMethod.NAME_OR_ID, nameOrId);
		if (template != null) 
		{
			ItemKey itemKey = template.getItemKey();
			ProjectKey projectKey = template.getProjectKey();
			if (merge)
			{
				itemKey = ShareableUtil.merge(dc, itemKey, item.getItemKey());
				projectKey = ShareableUtil.merge(dc, projectKey, item.getProjectKey());
			}
			item.setItemKey(itemKey);
			item.setProjectKey(projectKey);
		}
	}
	
	/**
		If a mapping for item subtype has been specified or if a single subtype was selected 
		for item identification in this job, this method is will update the subtype on the item.
		@param dc A DbControl used to lookup the subtype
		@param item The item to update
		@param data Data from the file
		@since 3.0
	*/
	protected void updateItemSubtype(DbControl dc, Subtypable item, FlatFileParser.Data data)
	{
		if (subtypeMapper != null)
		{
			String nameOrId = subtypeMapper.getValue(data);
			ItemSubtype subtype = findItemSubtype(dc, FallbackIdMethod.NAME_OR_SYSTEMID_OR_ID, item.getType(), nameOrId);
			if (nameOrId == null || subtype != null) item.setItemSubtype(subtype);
		}
		else if (selectedSubtype != null)
		{
			item.setItemSubtype(selectedSubtype);
		}
	}
	
	/**
		Try to find subtype from the given file data. If a single subtype was selected
		for item identification in this job, that subtype is used if no mapped subtype
		can be found.
		@param dc A DbControl used to lookup the subtype
		@param mainItemType The main item type of the subtype
		@param data Data from the file
		@return A subtype of null if no subtype could be found
		@since 3.0
	*/
	protected ItemSubtype findItemSubtype(DbControl dc, Item mainItemType, FlatFileParser.Data data)
	{
		ItemSubtype subtype = null;
		if (subtypeMapper != null)
		{
			String nameOrId = subtypeMapper.getValue(data);
			subtype = findItemSubtype(dc, FallbackIdMethod.NAME_OR_SYSTEMID_OR_ID, mainItemType, nameOrId);
		}
		if (subtype == null) subtype = selectedSubtype;
		return subtype;
	}
	// ---------------------------------------
	
	/**
		Create a {@link DefaultItemFunction} for the given item type.
		The currently active project (if any) is used.
		@param itemType The type of items the function should look for
		@since 3.1
	*/
	protected <T extends Nameable> DefaultItemFunction<T> getDefaultItemFunction(Item itemType)
	{
		DefaultItemFunction<T> fun = new DefaultItemFunction<T>(dc, activeProject, itemType);
		return fun;
	}
		
	
	/**
		Get the id method to use for finding items. This implementation 
		searches the methods returned by {@link #getIdMethods()} for a method
		with the given name. If no method can be found {@link PropertyIdMethod#NAME}
		is returned.
	*/
	protected IdMethod getIdMethod(String method)
	{
		for (IdMethod im : getIdMethods())
		{
			if (im.getMethod().equals(method))
			{
				return im;
			}
		}
		return PropertyIdMethod.NAME;
	}
	
	/**
		Parse a string to a date using the date format specified by the 
		'Date format' parameter. This is a utility method that subclasses can
		use to extract date values, since the {@link Mapper} interface doesn't
		supports it.
	*/
	protected Date parseDate(String date)
	{
		return dateFormatter.parseString(date);
	}
	
	private ItemQuery<ItemSubtype> subtypeQuery;
	
	/**
		Find an item subtype with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier for the subtype
		@param mainType The main item type for the subtype
		@return A subtype, or null if no item could be found
		@since 3.0
	*/
	protected ItemSubtype findItemSubtype(DbControl dc, IdMethod idMethod, Item mainType, String identifier)
	{
		if (identifier == null) return null;
		if (subtypeQuery == null)
		{
			subtypeQuery = initReferenceQuery(dc, idMethod, ItemSubtype.getQuery(null), false);
			subtypeQuery.restrictPermanent(
				Restrictions.eq(
					Hql.property("itemType"), 
					Expressions.parameter("itemType", Type.INT)
				)
			);			
		}
		subtypeQuery.setParameter("itemType", mainType.getValue(), Type.INT);
		return findReferencedItem(dc, idMethod, subtypeQuery, identifier, 0, false);
	}
	
	private ItemQuery<Protocol> protocolQuery;
	/**
		Same as {@link #findProtocol(DbControl, IdMethod, String, ItemSubtype)} with
		a null item subtype.
	*/
	protected Protocol findProtocol(DbControl dc, IdMethod idMethod, String identifier)
	{
		return findProtocol(dc, idMethod, identifier, null);
	}
	
	/**
		Find a protocol with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		<p>
		NOTE! If this method is called with a non-null item subtype parameter and 
		no protocol is found the query will be retried without any subtype.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@param subtype The protocol subtype, or null if the type doesn't matter
		@return A protocol, or null if no item could be found
		@since 3.0
	*/
	protected Protocol findProtocol(DbControl dc, IdMethod idMethod, String identifier, ItemSubtype subtype)
	{
		if (identifier == null) return null;
		if (protocolQuery == null) 
		{
			protocolQuery = initReferenceQuery(dc, idMethod, Protocol.getQuery(), true);
		}
		return findReferencedItemWithSubtype(dc, idMethod, protocolQuery, identifier, subtype, false);
	}

	private ItemQuery<Kit> kitQuery;
	/**
		Find a kit  with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		<p>
		NOTE! If this method is called with a non-null item subtype parameter and 
		no kit is found the query will be retried without any subtype.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@param subtype The kit subtype, or null if the type doesn't matter
		@return A kit, or null if no item could be found
		@since 3.7
	*/
	protected Kit findKit(DbControl dc, IdMethod idMethod, String identifier, ItemSubtype subtype)
	{
		if (identifier == null) return null;
		if (kitQuery == null) 
		{
			kitQuery = initReferenceQuery(dc, idMethod, Kit.getQuery(), true);
		}
		return findReferencedItemWithSubtype(dc, idMethod, kitQuery, identifier, subtype, false);
	}

	private ItemQuery<Tag> tagQuery;
	/**
		Same as {@link #findTag(DbControl, IdMethod, String, ItemSubtype)} with
		a null item subtype.
	*/
	protected Tag findTag(DbControl dc, IdMethod idMethod, String identifier)
	{
		return findTag(dc, idMethod, identifier, null);
	}
	
	/**
		Find a tag with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		<p>
		NOTE! If this method is called with a non-null item subtype parameter and 
		no hardware is found the query will be retried without any subtype.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@param subtype The tag subtype, or null if the type doesn't matter
		@return A tag, or null if no item could be found
		@since 3.0
	*/
	protected Tag findTag(DbControl dc, IdMethod idMethod, String identifier, ItemSubtype subtype)
	{
		if (identifier == null) return null;
		if (tagQuery == null) 
		{
			tagQuery = initReferenceQuery(dc, idMethod, Tag.getQuery(), true);
		}
		return findReferencedItemWithSubtype(dc, idMethod, tagQuery, identifier, subtype, false);
	}
	
	
	private ItemQuery<Hardware> hardwareQuery;
	/**
		Same as {@link #findHardware(DbControl, IdMethod, String, ItemSubtype)} with
		a null item subtype.
	*/
	protected Hardware findHardware(DbControl dc, IdMethod idMethod, String identifier)
	{
		return findHardware(dc, idMethod, identifier, null);
	}
	
	/**
		Find a hardware with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		<p>
		NOTE! If this method is called with a non-null item subtype parameter and 
		no hardware is found the query will be retried without any subtype.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@param subtype The hardware subtype, or null if the type doesn't matter
		@return A hardware, or null if no item could be found
		@since 3.0
	*/
	protected Hardware findHardware(DbControl dc, IdMethod idMethod, String identifier, ItemSubtype subtype)
	{
		if (identifier == null) return null;
		if (hardwareQuery == null) 
		{
			hardwareQuery = initReferenceQuery(dc, idMethod, Hardware.getQuery(), true);
		}
		return findReferencedItemWithSubtype(dc, idMethod, hardwareQuery, identifier, subtype, false);
	}

	private ItemQuery<Software> softwareQuery;
	/**
		Same as {@link #findSoftware(DbControl, IdMethod, String, ItemSubtype)} with
		a null item subtype.
	*/
	protected Software findSoftware(DbControl dc, IdMethod idMethod, String identifier)
	{
		return findSoftware(dc, idMethod, identifier, null);
	}
	
	/**
		Find a software with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		<p>
		NOTE! If this method is called with a non-null item subtype parameter and 
		no software is found the query will be retried without any subtype.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@param subtype The software subtype, or null if the type doesn't matter
		@return A software, or null if no item could be found
		@since 3.0
	*/
	protected Software findSoftware(DbControl dc, IdMethod idMethod, String identifier, ItemSubtype subtype)
	{
		if (identifier == null) return null;
		if (softwareQuery == null) 
		{
			softwareQuery = initReferenceQuery(dc, idMethod, Software.getQuery(), true);
		}
		return findReferencedItemWithSubtype(dc, idMethod, softwareQuery, identifier, subtype, false);
	}
	
	private ItemQuery<AnnotationTypeCategory> annotationTypeCategoryQuery;
	/**
		Find an annotation type category  with a given identifier. This is a utility 
		method that	subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return An annotation type category, or null if no item could be found
		@since 3.8
	*/
	protected AnnotationTypeCategory findAnnotationTypeCategory(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (annotationTypeCategoryQuery == null) 
		{
			annotationTypeCategoryQuery = initReferenceQuery(dc, idMethod, AnnotationTypeCategory.getQuery(), false);
		}
		return findReferencedItem(dc, idMethod, annotationTypeCategoryQuery, identifier, 0, false);
	}
	
	private ItemQuery<Unit> unitQuery;
	/**
		Find a unit with a given identifier. This is a utility 
		method that	subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A unit, or null if no item could be found
		@since 3.8
	*/
	protected Unit findUnit(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (unitQuery == null) 
		{
			unitQuery = initReferenceQuery(dc, idMethod, Unit.getQuery(), false);
		}
		return findReferencedItem(dc, idMethod, unitQuery, identifier, 0, false);
	}

	private ItemQuery<Platform> platformQuery;
	/**
		Find a platform with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A platform, or null if no item could be found
	*/
	protected Platform findPlatform(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (platformQuery == null) 
		{
			platformQuery = initReferenceQuery(dc, idMethod, Platform.getQuery(), false);
		}
		return findReferencedItem(dc, idMethod, platformQuery, identifier, 0, false);
	}
	/**
		Get the default platform of the currently active project. If there
		is more than one default platform, the first one is selected.
		@return The active platform, or null if no project is active or
			the project doesn't specify a platform
		@since 2.9
	*/
	protected Platform getProjectDefaultPlatform(DbControl dc)
	{
		if (sc.getActiveProjectId() == 0) return null;
		Project activeProject = Project.getById(dc, sc.getActiveProjectId());
		List<? extends BasicItem> platforms = activeProject.findDefaultItems(dc, Item.PLATFORM, true);
		return platforms == null || platforms.size() == 0 ? null : (Platform)platforms.get(0);
	}
	
	/**
		Get the default platform variant of the currently active project.
		@return The active platform variant, or null if no project is active or
			the project doesn't specify a variant
		@since 2.9
	*/
	protected PlatformVariant getProjectDefaultVariant(DbControl dc)
	{
		if (sc.getActiveProjectId() == 0) return null;
		Project activeProject = Project.getById(dc, sc.getActiveProjectId());
		List<? extends BasicItem> platforms = activeProject.findDefaultItems(dc, Item.PLATFORMVARIANT, true);
		return platforms == null || platforms.size() == 0 ? null : (PlatformVariant)platforms.get(0);
	}
	
	private ItemQuery<PlatformVariant> variantQuery;
	/**
		Find a platform variant with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param platform If given, searched only variants of this platform
		@param identifier The identifier protocol
		@return A platform, or null if no item could be found
	*/
	protected PlatformVariant findVariant(DbControl dc, Platform platform, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (variantQuery == null) 
		{
			variantQuery = initReferenceQuery(dc, idMethod, PlatformVariant.getQuery(), false);
			// variant.platform = :platform or 1 = :anyplatform
			variantQuery.restrict(
				Restrictions.or(
					Restrictions.eq(Hql.property("platform.id"), Expressions.parameter("platform")),
					Restrictions.eq(Expressions.integer(1), Expressions.parameter("anyplatform"))
				)
			);
		}
		variantQuery.setParameter("platform", platform == null ? 0 : platform.getId(), Type.INT);
		variantQuery.setParameter("anyplatform", platform == null ? 1 : 0, Type.INT);
		return findReferencedItem(dc, idMethod, variantQuery, identifier, 0, false);
	}
	
	/**
		Get the default raw data type of the currently active project.
		@return The active raw data type, or null if no project is active or
			the project doesn't specify raw data type
		@since 2.9
	*/
	protected RawDataType getProjectDefaultRawDataType(DbControl dc)
	{
		if (sc.getActiveProjectId() == 0) return null;
		Project activeProject = Project.getById(dc, sc.getActiveProjectId());
		return activeProject.getDefaultRawDataType();
	}

	
	private ItemQuery<DataFileType> fileTypeQuery;
	/**
		Find a data file type with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A data file, or null if no item could be found
	*/
	protected DataFileType findDataFileType(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (fileTypeQuery == null) 
		{
			fileTypeQuery = initReferenceQuery(dc, idMethod, DataFileType.getQuery(), false);
		}
		return findReferencedItem(dc, idMethod, fileTypeQuery, identifier, 0, false);
	}
	
	private ItemQuery<PermissionTemplate> permissionTemplateQuery;
	/**
		Find a permission template with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier for the template
		@return A permission template, or null if no item could be found
	*/
	protected PermissionTemplate findPermissionTemplate(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (permissionTemplateQuery == null) 
		{
			permissionTemplateQuery = initReferenceQuery(dc, idMethod, PermissionTemplate.getQuery(), false);
		}
		return findReferencedItem(dc, idMethod, permissionTemplateQuery, identifier, 0, false);
	}

	
	private ItemQuery<BioSource> bioSourceQuery;
	/**
		Same as {@link #findBioSource(DbControl, IdMethod, String, ItemSubtype)} with
		a null item subtype.
	*/
	protected BioSource findBioSource(DbControl dc, IdMethod idMethod, String identifier)
	{
		return findBioSource(dc, idMethod, identifier, null);
	}
		
	/**
		Find a biosource with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		<p>
		NOTE! If this method is called with a non-null item subtype parameter and 
		no biosource is found the query will be retried without any subtype.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@param subtype The biosource subtype, or null if the type doesn't matter
		@return A biosource, or null if no item could be found
		@since 3.0
	*/
	protected BioSource findBioSource(DbControl dc, IdMethod idMethod, String identifier, ItemSubtype subtype)
	{
		if (identifier == null) return null;
		if (bioSourceQuery == null) 
		{
			// where ... (itemSubtype = :subtype AND checkType = 1) || checkType = 0
			bioSourceQuery = initReferenceQuery(dc, idMethod, BioSource.getQuery(), true);
		}
		return findReferencedItemWithSubtype(dc, idMethod, bioSourceQuery, identifier, subtype, false);	}

	private ItemQuery<Sample> sampleQuery;
	/**
		Same as {@link #findSample(DbControl, IdMethod, String, ItemSubtype)}
		with a null subtype.
	*/
	protected Sample findSample(DbControl dc, IdMethod idMethod, String identifier)
	{
		return findSample(dc, idMethod, identifier, null);
	}

	/**
		Find a sample with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		<p>
		NOTE! If this method is called with a non-null item subtype parameter and 
		no sample is found the query will be retried without any subtype.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@param subtype The sample subtype, or null if the type doesn't matter
		@return A sample, or null if no item could be found
		@since 3.0
	*/
	protected Sample findSample(DbControl dc, IdMethod idMethod, String identifier, ItemSubtype subtype)
	{
		if (identifier == null) return null;
		if (sampleQuery == null) 
		{
			sampleQuery = initReferenceQuery(dc, idMethod, Sample.getQuery(), true);
		}
		return findReferencedItemWithSubtype(dc, idMethod, sampleQuery, identifier, subtype, false);
	}
	
	private ItemQuery<Extract> extractQuery;
	/**
		Same as {@link #findSample(DbControl, IdMethod, String, ItemSubtype)}
		with a null subtype.
	*/
	protected Extract findExtract(DbControl dc, IdMethod idMethod, String identifier)
	{
		return findExtract(dc, idMethod, identifier, null);
	}

	/**
		Find an extract with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		<p>
		NOTE! If this method is called with a non-null item subtype parameter and 
		no extract is found the query will be retried without any subtype.
	
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@param subtype The extract subtype, or null if the type doesn't matter
		@return An extract, or null if no item could be found
		@since 3.0
	*/
	protected Extract findExtract(DbControl dc, IdMethod idMethod, String identifier, ItemSubtype subtype)
	{
		if (identifier == null) return null;
		if (extractQuery == null) 
		{
			extractQuery = initReferenceQuery(dc, idMethod, Extract.getQuery(), true);
		}
		return findReferencedItemWithSubtype(dc, idMethod, extractQuery, identifier, subtype, false);
	}

	private ItemQuery<Extract> labeledExtractQuery;
	/**
		Find a labled extract with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		<p>
		NOTE! A labeled extract is an extract with a subtype {@link Extract#LABELED}
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A labeled extract, or null if no item could be found
	*/
	protected Extract findLabeledExtract(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (labeledExtractQuery == null) 
		{
			labeledExtractQuery = initReferenceQuery(dc, idMethod, Extract.getQuery(), false);
			labeledExtractQuery.restrictPermanent(
				Restrictions.eq(
					Hql.property("itemSubtype"), 
					Expressions.integer(SystemItems.getId(Extract.LABELED))
				)
			);
		}
		return findReferencedItem(dc, idMethod, labeledExtractQuery, identifier, SystemItems.getId(Extract.LABELED), false);
	}
	
	private ItemQuery<BioPlate> bioPlateQuery;
	/**
	 	Find a bioplate with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A bioplate, or null if no item could be found
	 */
	protected BioPlate findBioPlate(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (bioPlateQuery == null)
		{
			bioPlateQuery = initReferenceQuery(dc, idMethod, BioPlate.getQuery(), false);
		}
		return findReferencedItem(dc, idMethod, bioPlateQuery, identifier, 0, false);
	}
	
	/**
		Find a well on a bioplate. Null is allowed
		for all parameters but will always result in a null
		return value.
		
		@param dc The DbControl to use for database access
		@param plate The plate to look on
		@param row The row coordinate in alphabetic or 1-based values
		@param column The column coordinate in alphabetic or 1-based values
		@return A biowell, or null if no item could be found
		@since 2.16
	*/
	protected BioWell findBioWell(DbControl dc, BioPlate plate, String row, String column)
	{
		if (plate == null || row == null || column == null) return null;

		// Convert alphanumeric (1-based) to 0-based numerical coordinate
		int r = Coordinate.alphaToNumeric(row) - 1;
		int c = Coordinate.alphaToNumeric(column) - 1;
		BioWell well = plate.getBioWell(r, c);
		if (failIfNotFoundReference && well == null) 
		{
			throw new ItemNotFoundException("BioWell[" + row + "," + column + "]");
		}
		return well;
	}
	
	private ItemQuery<ArraySlide> arraySlideQuery;
	/**
		Find an array slide with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return An array slide, or null if no item could be found
	*/
	protected ArraySlide findArraySlide(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (arraySlideQuery == null) 
		{
			arraySlideQuery = initReferenceQuery(dc, idMethod, ArraySlide.getQuery(), false);
		}
		return findReferencedItem(dc, idMethod, arraySlideQuery, identifier, 0, false);
	}
	
	private ItemQuery<PhysicalBioAssay> hybQuery;
	/**
		Find a hybridization with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		<p>
		NOTE! A hybridization is a physical bioassay with a subtype {@link 
		PhysicalBioAssay#HYBRIDIZATION}
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A hybridization, or null if no item could be found
	*/
	protected PhysicalBioAssay findHybridization(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (hybQuery == null) 
		{
			hybQuery = initReferenceQuery(dc, idMethod, PhysicalBioAssay.getQuery(), false);
			hybQuery.restrictPermanent(
					Restrictions.eq(
						Hql.property("itemSubtype"), 
						Expressions.integer(SystemItems.getId(PhysicalBioAssay.HYBRIDIZATION))
					)
				);
		}
		return findReferencedItem(dc, idMethod, hybQuery, identifier, SystemItems.getId(PhysicalBioAssay.HYBRIDIZATION), false);
	}

	private ItemQuery<PhysicalBioAssay> physicalBioAssayQuery;
	/**
		Find a physical bioassay with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		<p>
		NOTE! If this method is called with a non-null item subtype parameter and 
		no bioassay is found the query will be retried without any subtype.
	
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@param subtype The bioassay subtype, or null if the type doesn't matter
		@return A physical bioassay, or null if no item could be found
		@since 3.0
	*/
	protected PhysicalBioAssay findPhysicalBioAssay(DbControl dc, IdMethod idMethod, String identifier, ItemSubtype subtype)
	{
		if (identifier == null) return null;
		if (physicalBioAssayQuery == null) 
		{
			physicalBioAssayQuery = initReferenceQuery(dc, idMethod, PhysicalBioAssay.getQuery(), true);
		}
		return findReferencedItemWithSubtype(dc, idMethod, physicalBioAssayQuery, identifier, subtype, false);
	}

	private ItemQuery<DerivedBioAssay> derivedBioAssayQuery;
	/**
		Find a derived bioassay with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		<p>
		NOTE! If this method is called with a non-null item subtype parameter and 
		no bioassay is found the query will be retried without any subtype.
	
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@param subtype The bioassay subtype, or null if the type doesn't matter
		@return A derived bioassay, or null if no item could be found
		@since 3.0
	*/
	protected DerivedBioAssay findDerivedBioAssay(DbControl dc, IdMethod idMethod, String identifier, ItemSubtype subtype)
	{
		if (identifier == null) return null;
		if (derivedBioAssayQuery == null) 
		{
			derivedBioAssayQuery = initReferenceQuery(dc, idMethod, DerivedBioAssay.getQuery(), true);
		}
		return findReferencedItemWithSubtype(dc, idMethod, derivedBioAssayQuery, identifier, subtype, false);
	}
	
	private ItemQuery<ArrayDesign> designQuery;
	/**
		Find an array design with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return An array design, or null if no item could be found
	*/
	protected ArrayDesign findArrayDesign(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (designQuery == null) 
		{
			designQuery = initReferenceQuery(dc, idMethod, ArrayDesign.getQuery(), false);
		}
		return findReferencedItem(dc, idMethod, designQuery, identifier, 0, false);
	}

	private ItemQuery<ArrayBatch> batchQuery;
	/**
		Find an array batch with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return An array batch, or null if no item could be found
	*/
	protected ArrayBatch findArrayBatch(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (batchQuery == null) 
		{
			batchQuery = initReferenceQuery(dc, idMethod, ArrayBatch.getQuery(), false);
		}
		return findReferencedItem(dc, idMethod, batchQuery, identifier, 0, false);
	}

	
	private ItemQuery<PlateGeometry> geometryQuery;
	/**
		Find a plate geometry with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A plate geometry, or null if no item could be found
		@since 2.16
	*/
	protected PlateGeometry findPlateGeometry(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (geometryQuery == null) 
		{
			geometryQuery = initReferenceQuery(dc, idMethod, PlateGeometry.getQuery(), false);
		}
		return findReferencedItem(dc, idMethod, geometryQuery, identifier, 0, false);
	}

	private ItemQuery<BioPlateType> bioPlateTypeQuery;
	/**
		Find a bioplate type with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A bioplate type, or null if no item could be found
		@since 2.16
	*/
	protected BioPlateType findBioPlateType(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (bioPlateTypeQuery == null) 
		{
			bioPlateTypeQuery = initReferenceQuery(dc, idMethod, BioPlateType.getQuery(), false);
		}
		return findReferencedItem(dc, idMethod, bioPlateTypeQuery, identifier, 0, false);
	}
	
	/**
		Find a matching raw data type. This method will look in the following order:
		<ol>
		<li>The raw data type locked to the platform variant:
			{@link PlatformVariant#getRawDataType()}
		<li>The raw data type locked to the platform:
			{@link Platform#getRawDataType()}
		<li>Any registered raw data type with the given 'id':
			{@link RawDataTypes#getRawDataType(String)}
		<li>Any registered raw data type with the given 'name':
			{@link RawDataTypes#getRawDataTypes()}
		<li>The project default raw data type:
			{@link #getProjectDefaultRawDataType(DbControl)}
		</ol>
		If a raw data type is found by a check, the remaining checks are 
		ignored. If no item is found or if multiple items are found (in the 'name' lookup)
		the settings ({@link #failIfMultipleFoundReferences}
		and {@link #failIfNotFoundReference}) determine if an exception is thrown
		or not.
		@param platform The current platform or null to not use a platform for lookup
		@param variant The current platform variant or null to not use a variant for lookup
		@param identifier The 'id' or 'name' of a raw data type, or null to not use this
			for lookup
		@return A raw data type or null if not found
		@since 2.15
	*/
	protected RawDataType findRawDataType(DbControl dc, Platform platform, PlatformVariant variant, String identifier)
	{
		RawDataType rdt = null;
		
		// If a variant/platform is given, check them first
		if (variant != null) rdt = variant.getRawDataType();
		if (rdt == null && platform != null) rdt = platform.getRawDataType();
		
		// Find by 'id'
		if (rdt == null && identifier != null) rdt = RawDataTypes.getRawDataType(identifier);
			
		// Find by 'name'
		if (rdt == null && identifier != null)
		{
			for (RawDataType tmp : RawDataTypes.getRawDataTypes())
			{
				if (identifier.equals(tmp.getName()))
				{
					if (rdt == null)
					{
						rdt = tmp;
						if (!failIfMultipleFoundReferences)
						{
							// It is safe to ignore the rest
							break;
						}
					}
					else 
					{
						// This is the second raw data type with the same name that we find
						throw new BaseException("Fount >1 raw data type [name=" + identifier + "]");
					}
				}
			}
		}
		
		// Check the project default raw data type
		if (rdt == null) rdt = getProjectDefaultRawDataType(dc);
		if (rdt == null && failIfNotFoundReference)
		{
			throw new ItemNotFoundException("RawDataType[id or name=" + identifier + "]");
		}
		return rdt;
	}
	
	private ItemQuery<FileServer> fileServerQuery;
	/**
		Find a file server with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A file server, or null if no item could be found
		@since 3.3
	*/
	protected FileServer findFileServer(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (fileServerQuery == null) 
		{
			fileServerQuery = initReferenceQuery(dc, idMethod, FileServer.getQuery(), false);
		}
		return findReferencedItem(dc, idMethod, fileServerQuery, identifier, 0, false);
	}

	private Set<String> fileCache;
	/**
		Set or add a file to a file set. The first time this method is called for
		a given item and file type combination {@link FileSet#setMember(File, DataFileType)}
		is used to add the file to the file set (which will replace all other files of
		the same file type. If this method is called more times then 
		{@link FileSet#addMember(File, DataFileType)} is used so that already added
		files are preserved. 
		
		If the file is null the first time, all files of the given file types are removed.
		If the file is null after the first time, this call is ignored.
		
		@param item The item to set/add a file to
		@param fileType The type of the file
		@param file The file, or null
		@return The new member entry, or null if the call didn't result in
			an addition
		@since 3.0
	*/
	protected FileSetMember setOrAddFile(FileStoreEnabled item, DataFileType fileType, File file)
	{
		if (fileCache == null) fileCache = new HashSet<String>();
		String cacheId = item.getType().name() + "#" + item.getId() + "#" + fileType.getId();
		FileSetMember added = null;
		if (!fileCache.contains(cacheId))
		{
			// Has been called at least one time before
			if (file != null) 
			{
				added = item.getFileSet().addMember(file, fileType);
			}
		}
		else
		{
			added = item.getFileSet().setMember(file, fileType);
			fileCache.add(cacheId);
		}
		return added;
	}
	
	/**
		Cache that holds loaded/created items. The cache key is item type + identifier,
		eq. SAMPLE:My sample, EXTRACT:An extract. The same item may be cached under more 
		than one key depending on which identifier has been used to look it up,
		eg name, id, etc.: SAMPLE:My sample, SAMPLE:10.
	*/
	private Map<String, BasicItem> itemCache;
	
	/**
		Initialise a query that is used to find referenced items. This method
		delegates to {@link IdMethod#prepareQuery(DbControl, ItemQuery)} and
		then adds {@link Include} options: MINE, IN_PROJECT, SHARED and OTHERS.
				
		@param dc A DbControl used for database access
		@param idMethod The identification method to use
		@param query The query to initialised
		@param subtype If TRUE a restriction for subtype is added to the query
		@return The initialised query
	*/
	protected <T extends BasicItem> ItemQuery<T> initReferenceQuery(DbControl dc, IdMethod idMethod, 
		ItemQuery<T> query, boolean subtype)
	{
		query = idMethod.prepareQuery(dc, query);
		query.include(Include.MINE, Include.IN_PROJECT, Include.SHARED, Include.OTHERS);
		if (subtype)
		{
			// where ... (itemSubtype = :subtype AND checkType = 1) || checkType = 0
			Expression checkType = Expressions.parameter("checkType");
			query.restrictPermanent(
					Restrictions.or(
						Restrictions.and(
							Restrictions.eq(Hql.property("itemSubtype"), Expressions.parameter("subtype")),
							Restrictions.eq(checkType, Expressions.integer(1))
						),
						Restrictions.eq(checkType, Expressions.integer(0))
					)
				);
		}
		return query;
	}
	
	/**
		Find a referenced item with a subtype. This method may 
		call the {@link #findReferencedItem(DbControl, IdMethod, ItemQuery, String, int, boolean)}
		method two times. First, with a query limited to return items of the given subtype,
		then a second time without the limit. The query must have been created by
		the {@link #initReferenceQuery(DbControl, IdMethod, ItemQuery, boolean)} method with
		TRUE for the last parameter. If the subtype parameter is null the first query is
		skipped.
	*/
	protected <T extends BasicItem> T findReferencedItemWithSubtype(DbControl dc, IdMethod idMethod, 
			ItemQuery<T> query, String identifier, ItemSubtype subtype, boolean ignoreNotFound)
	{
		T item = null;
		if (subtype != null)
		{
			query.setParameter("subtype", subtype.getId(), Type.INT);
			query.setParameter("checkType", 1, Type.INT);
			item = findReferencedItem(dc, idMethod, query, identifier, subtype.getId(), true);
		}
		if (item == null)
		{
			query.setParameter("subtype", 0, Type.INT);
			query.setParameter("checkType", 0, Type.INT);
			item = findReferencedItem(dc, idMethod, query, identifier, 0, ignoreNotFound);
		}
		return item;
	}
	
	/**
		Find a referenced item. This method will first look in the {@link #itemCache}.
		If no item is found in the cache it uses the query and id method 
		({@link IdMethod#find(DbControl, ItemQuery, String)}) to look for the item.
		The result of the lookup is stored in the cache (even if no item is found).

		@param dc A DbControl used for database access
		@param idMethod The identification method 
		@param ignoreNotFound TRUE to ignore not found errors
		@return The item, or null if none is found
		@throws ItemNotFoundException If no item is found and the error handling is set to
			fail for this condition
		@throws BaseException If multiple items are found and the error handling is set to
			fail for this condition
	*/
	@SuppressWarnings("unchecked")
	protected <T extends BasicItem> T findReferencedItem(DbControl dc, IdMethod idMethod, 
		ItemQuery<T> query, String identifier, int subtypeId, boolean ignoreNotFound)
	{
		T item = null;
		Item itemType = query.getItemType();
		String cacheKey1 = itemType.name() + ":" + identifier;
		String cacheKey2 = cacheKey1 + ":" + subtypeId;
		if (itemCache.containsKey(cacheKey2))
		{
			item = (T)itemCache.get(cacheKey2);
		}
		else if (itemCache.containsKey(cacheKey1))
		{
			item = (T)itemCache.get(cacheKey1);
		}
		else
		{
			List<T> result = idMethod.find(dc, query, identifier);
			if (result.size() == 0)
			{
				// Not found
				if (failIfNotFoundReference && !ignoreNotFound)
				{
					throw new ItemNotFoundException(itemType+"[" +idMethod + "=" + identifier + "]");
				}
			}
			else
			{
				// Pick out items where use-permission is granted
				List<T> accessibleItems = new ArrayList<T>();
				for (T t : result)
				{
					if (t.hasPermission(Permission.USE)) 
					{
						accessibleItems.add(t);
					}
				}
				if (accessibleItems.size() > 1)
				{
					// Several items found
					if (failIfMultipleFoundReferences)
					{
						throw new BaseException("Found " + accessibleItems.size() + " " + itemType + " [" +idMethod + "=" + identifier + "]");
					}
				}
				else if (accessibleItems.size() == 0)
				{
					// No use permission on item
					if (failIfNoUsePermissionToReference)
					{
						throw new PermissionDeniedException(Permission.USE, itemType+"[" +idMethod + "=" + identifier + "]");
					}					
				}
				else
				{
					item = accessibleItems.get(0);
				}
				itemCache.put(cacheKey2, item);
			}
		}
		return item;
	}
	
	private RequestInformation configureParser;
	private RequestInformation configureOptions;

	private Pattern getPattern(ParameterValues parameters, String name)
		throws BaseException
	{
		String regexp = (String)parameters.getValue(name);
		return regexp != null ? Pattern.compile(regexp) : null;
	}
	
	protected FlatFileParser getInitializedFlatFileParser(ParameterValues parameters)
		throws BaseException
	{
		FlatFileParser ffp = new FlatFileParser();
		ffp.setDataHeaderRegexp(getPattern(parameters, dataHeaderRegexpParameter.getName()));
		ffp.setDataSplitterRegexp(getPattern(parameters, dataSplitterRegexpParameter.getName()));
		Boolean trimQuotes = (Boolean)parameters.getValue(trimQuotesParameter.getName());
		if (trimQuotes != null) ffp.setTrimQuotes(trimQuotes);
		ffp.setIgnoreRegexp(getPattern(parameters, ignoreRegexpParameter.getName()));
		ffp.setDataFooterRegexp(getPattern(parameters, dataFooterRegexpParameter.getName()));
		ffp.setMinDataColumns(IntegerUtil.getInt((Integer)parameters.getValue(minDataColumnsParameter.getName())));
		ffp.setMaxDataColumns(IntegerUtil.getInt((Integer)parameters.getValue(maxDataColumnsParameter.getName())));
		return ffp;
	}
	
	protected RequestInformation getConfigureParserParameters(GuiContext context, boolean forJob)
	{
		if (configureParser == null)
		{
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();

			if (forJob)
			{
				// The file to import from
				parameters.add(fileParameter);
			}
			else
			{
				parameters.add(defaultsSection);
			}
			List<String> allowedModes = new ArrayList<String>();
			String defaultMode = "update";
			if (context == null || sc.hasPermission(Permission.CREATE, context.getItem()))
			{
				allowedModes.add("create+update");
				allowedModes.add("create");
				defaultMode = "create+update";
			}
			allowedModes.add("update");
			PluginParameter<String> modeParameter = new PluginParameter<String>(
					"mode", 
					"Mode", 
					"Select an import mode: \n" + 
					"create+update = create missing items, update existing items\n" +
					"create = create missing items only\n" +
					"update = update existing items only\n\n" +
					"NOTE! Not all modes may be available due to lack of permissions.",
					new StringParameterType(255, defaultMode, forJob, 1, 0, 0, allowedModes)
					);
			parameters.add(modeParameter);
			if (forJob)
			{
				addMoreJobParameters(parameters);
			}
			else
			{
				parameters.add(Parameters.dryRunParameter(null, null, null));
			}

			// Identification method
			if (forJob)
			{
				parameters.add(identificationSection);
			}
			Enumeration<String, String> idMethods = new Enumeration<String, String>();
			for (IdMethod idm : getIdMethods())
			{
				idMethods.add(idm.getMethod(), idm.getTitle());
			}
			PluginParameter<String> idMethodParameter = new PluginParameter<String>(
				"idMethod",
				"Identification method",
				"The method to use to find already existing items. You can use either the " +
				"internal BASE ID, the name or, for some items, some another appropriate identifer " +
				"property. You need to specify a column mapping for " +
				"the selected method.",
				new StringParameterType(null, idMethods.getKey(0), forJob, 1, 0, 0, idMethods)
			);
			parameters.add(idMethodParameter);
			
			if (getItemForSubtypes() != null)
			{
				DbControl dc = sc.newDbControl();
				try
				{
					// Get a list of possible subtypes
					List<ItemSubtype> subtypes = new ArrayList<ItemSubtype>(ItemSubtype.getQuery(getItemForSubtypes()).list(dc));
				
					if (subtypes != null && subtypes.size() > 0)
					{
						// Find a default subtype from the current context
						ItemContext cc = sc.getCurrentContext(getItemForSubtypes());
						int selectedSubtypeId = Values.getInt(cc.getPropertyValue("itemSubtype"));
						ItemSubtype defaultSubtype = selectedSubtypeId == 0 ? null : ItemSubtype.getById(dc, selectedSubtypeId);
						
						// Create parameter
						PluginParameter<ItemSubtype> subtypeParameter = new PluginParameter<ItemSubtype>(
							"itemSubtypes", 
							"Item subtypes", 
							"Only look for items of selected subtypes. If none is selected, all subtypes " +
							"will be searched. If a single subtype is selected, new items will automatically " +
							"be assigned that subtype.",
							new ItemParameterType<ItemSubtype>(ItemSubtype.class, defaultSubtype, false, 0, subtypes)
						);
						parameters.add(subtypeParameter);
					}
				}
				finally
				{
					if (dc != null) dc.close();
				}

			}
			
			if (forJob)
			{
				BooleanParameterType optionalBooleanType = new BooleanParameterType(null, false);
				ItemContext cc = sc.getCurrentContext(context.getItem());
				Set<Include> include = cc.getInclude();
				// Should we search in MINE, SHARED, IN_PROJECT, etc.
				parameters.add(new PluginParameter<Boolean>(
						"includeMine",
						"Owned by me",
						"Search items that are owned by yourself.",
						include.contains(Include.MINE) ? Boolean.TRUE : null,
						optionalBooleanType)
					);
				parameters.add(new PluginParameter<Boolean>(
						"includeShared",
						"Shared to me",
						"Search items that have been shared to you by other users.",
						include.contains(Include.SHARED) ? Boolean.TRUE : null,
						optionalBooleanType)
					);
					
				if (sc.getActiveProjectId() != 0)
				{
					parameters.add(new PluginParameter<Boolean>(
							"includeInProject",
							"In current project",
							"Search items that are shared to the current project.",
							include.contains(Include.IN_PROJECT) ? Boolean.TRUE : null,
							optionalBooleanType)
						);
				}
				if (sc.hasPermission(Permission.READ, context.getItem()))
				{
					parameters.add(new PluginParameter<Boolean>(
							"includeOthers",
							"Owned by others",
							"Search items that are owned by other users.",
							include.contains(Include.OTHERS) ? Boolean.TRUE : null,
							optionalBooleanType)
						);
				}
			}

			// Parser regular expressions
			parameters.add(parserSection);
			parameters.add(dataHeaderRegexpParameter);
			parameters.add(dataSplitterRegexpParameter);
			parameters.add(trimQuotesParameter);
			parameters.add(ignoreRegexpParameter);
			parameters.add(dataFooterRegexpParameter);
			parameters.add(minDataColumnsParameter);
			parameters.add(maxDataColumnsParameter);
			parameters.add(Parameters.charsetParameter(null, null, getCharset()));
			parameters.add(Parameters.decimalSeparatorParameter(null, null, getDecimalSeparator()));
			parameters.add(Parameters.dateFormatParameter(null, null, 
				sc.getUserClientSetting("formatter.date.formatstring")));
			addMoreParsingParameters(parameters);
			
			// Column mappings
			parameters.add(mappingSection);
			parameters.add(complexMappings);
			addMoreColumnMappingParameters(parameters);
			
			if (forJob)
			{
				configureParser = new RequestInformation
				(
					Request.COMMAND_CONFIGURE_JOB,
					"Select a file from which to import item data",
					"Here you select which file to import data from, and the " +
					"regular expressions and other settings used to parse it. Step 1 of 2.",
					parameters
				);
			}
			else
			{
				configureParser = new RequestInformation
				(
					Request.COMMAND_CONFIGURE_PLUGIN,
					"Specify parser regular expressions and settings",
					"This plug-in can be configured with the regular expressions and " +
					"settings used to parse files.",
					parameters
				);				
			}
		}
		return configureParser;
	}
	
	protected RequestInformation getConfigureOptionsParameters(GuiContext context)
	{
		if (configureOptions == null)
		{
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();

			String mode = (String)job.getValue("mode");
			boolean updateMode = mode == null || mode.contains("update");
			boolean createMode = mode == null || mode.contains("create");
			
			// Error handling options
			parameters.add(errorSection);
			parameters.add(Parameters.logFileParameter(null, null, null));
			parameters.add(Parameters.dryRunParameter(null, null, null));
			parameters.add(defaultErrorParameter);
			if (!createMode) parameters.add(itemNotFoundErrorParameter);
			if (!updateMode) parameters.add(itemExistsErrorParameter);
			parameters.add(multipleItemsFoundErrorParameter);
			if (updateMode) parameters.add(noWritePermissionToItemErrorParameter);
			parameters.add(referenceNotFoundErrorParameter);
			parameters.add(noUsePermissionToReferenceErrorParameter);
			parameters.add(multipleReferencesFoundErrorParameter);
			parameters.add(Parameters.invalidUseOfNullError(null, null, null));
			parameters.add(stringTooLongErrorParameter);
			parameters.add(numberFormatErrorParameter);
			parameters.add(numberOutOfRangeErrorParameter);
			addMoreErrorParameters(parameters);
			
			configureOptions = new RequestInformation
			(
				CONFIGURE_OPTIONS,
				"Other settings and error handling options",
				"Specify other settings and error handling options. Step 2 of 2.",
				parameters
			);
		}
		return configureOptions;	
	}


}
