/**
	$Id: PowerUserTest.java 5983 2012-02-23 14:45:07Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.test.roles;

import java.util.Arrays;
import java.util.EnumSet;
import java.util.List;

import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.ArrayBatch;
import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.ArraySlide;
import net.sf.basedb.core.BioPlateType;
import net.sf.basedb.core.BioWell;
import net.sf.basedb.core.ClonableProperty;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileParameterType;
import net.sf.basedb.core.Group;
import net.sf.basedb.core.Hardware;
import net.sf.basedb.core.IntegerParameterType;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.Plate;
import net.sf.basedb.core.PlateGeometry;
import net.sf.basedb.core.PlateType;
import net.sf.basedb.core.Platform;
import net.sf.basedb.core.PlatformVariant;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.Project;
import net.sf.basedb.core.Protocol;
import net.sf.basedb.core.Quantity;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.RawDataTypes;
import net.sf.basedb.core.ReporterCloneTemplate;
import net.sf.basedb.core.Software;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Unit;
import net.sf.basedb.test.FileUtil;
import net.sf.basedb.test.MouseData;
import net.sf.basedb.test.PluginUtil;
import net.sf.basedb.test.SequenceData;
import net.sf.basedb.test.TestUtil;

/**
	This class contains the tests for the power user created
	by the root user.
	<ol>
	<li>Create a project {@link #createProject(DbControl, String)} and activate it
	<li>Create several protocols: {@link #createProtocol(DbControl, String, String, AnnotationType[])}
	<li>Create hardware: {@link #createHardware(DbControl, String, String)}
	<li>Create software: {@link #createSoftware(DbControl, String, String)}
	<li>Create annotation types: {@link #createAnnotationType(DbControl, String, Type, Unit, Object[], Item[])}
	<li>Create plate type: {@link #createPlateType(DbControl, String, int, int)}
	<li>Create bioplate type: {@link #createBioPlateType(DbControl, String, Item, BioWell.LockMode)}
	<li>Create file formats for import:
	<li>Import plates:
	<li>Create array design: {@link #createArrayDesign(DbControl, Platform, String, String)}
	<li>Import features:
	<li>Create array batch: {@link #createArrayBatch(DbControl, String, ArrayDesign, Hardware, Protocol)}
	<li>Create array slides: {@link #createArraySlide(DbControl, String, ArrayBatch, int)}
	</ol>

	@author nicklas
	@version 2.0
	@base.modified $Date: 2012-02-23 15:45:07 +0100 (to, 23 feb 2012) $
*/
public class PowerUserTest
{

	public static boolean ok = true;
	
	/**
		Run the tests in this class.
	*/
	public static boolean run(boolean useBatchImporters)
	{
		DbControl dc = null;
		TestUtil.write("++Entering power user test\n");
		try
		{
			TestUtil.login("power", "power");
			
			// Project
			dc = TestUtil.getDbControl();
			Project p = createProject(dc, "Project A");
			dc.commit();
			TestUtil.getSessionControl().setActiveProject(p);
			
			dc = TestUtil.getDbControl();
			// Item subtype
			ItemSubtype qualityControl = createItemSubtype(dc, "Quality control", Item.EXTRACT, true);
			
			// Annotation types
			AnnotationType drugResistance = createAnnotationType(dc, "Drug resistance", 
					Type.STRING, null,
					new String[] { "high", "medium", "low" },  
					new Item[] { Item.BIOSOURCE }
				);
			
			Quantity time = Quantity.getById(dc, SystemItems.getId(Quantity.TIME));
			Unit hoursUnit = time.getUnit("hour");
			AnnotationType hours = createAnnotationType(dc, "Time", 
					Type.INT, hoursUnit, null,
					new Item[] { Item.SAMPLE } 
				);
			
			AnnotationType rin = createAnnotationType(dc, "RIN", 
					Type.FLOAT, null, null, 
					new Item[] { Item.EXTRACT }
				);
			
			AnnotationType dyeSwap = createAnnotationType(dc, "Dye swap", 
					Type.BOOLEAN, null, null, 
					new Item[] { Item.RAWBIOASSAY } 
				);
			
			Quantity potential = Quantity.getById(dc, SystemItems.getId(Quantity.POTENTIAL));
			Unit volt = potential.getUnit("V");
			AnnotationType pmtGain = createProtocolParameter(dc, "PMT gain", 
					Type.FLOAT, volt, null, new Item[] { Item.DERIVEDBIOASSAY }
			);
			
			// Protocols, hardware and software
			Protocol sampling = createProtocol(dc, "Sampling A", Protocol.SAMPLING);
			Protocol extraction = createProtocol(dc, "Extraction A", Protocol.EXTRACTION);
			Protocol labeling = createProtocol(dc, "Labeling A", Protocol.LABELING);
			Protocol libraryPrepartion = createProtocol(dc, "Library preparation A", Protocol.LIBRARY_PREPARATION);
			Protocol hybridization = createProtocol(dc, "Hybridization A", Protocol.HYBRIDIZATION);
			Protocol cBotSettings = createProtocol(dc, "cBot Settings A", Protocol.CLUSTER_GENERATION);
			Protocol scanning = createProtocol(dc, "Scanning A", Protocol.SCANNING, pmtGain);
			Protocol hiSeqSettings = createProtocol(dc, "HiSeq Settings A", Protocol.SEQUENCING);
			Protocol topHatSettings = createProtocol(dc, "TopHat Settings A", Protocol.ALIGNMENT);
			Protocol feature = createProtocol(dc, "Feature extraction A", Protocol.FEATURE_EXTRACTION);
			Protocol cufflinksSettings = createProtocol(dc, "Cufflinks Settings A", Protocol.FEATURE_EXTRACTION);
			Protocol printing = createProtocol(dc, "Printing A", Protocol.PRINTING);
			
			Hardware hybStation = createHardware(dc, "Hybridization station A", Hardware.HYBRIDIZATION_STATION);
			Hardware cBot = createHardware(dc, "cBot A", Hardware.CLUSTER_GENERATOR);
			Hardware robot = createHardware(dc, "Print robot A", Hardware.PRINT_ROBOT);
			Hardware scanner = createHardware(dc, "Scanner A", Hardware.SCANNER);
			Hardware hiSeq = createHardware(dc, "HiSeq 2000 A", Hardware.SEQUENCER);
			Software softwareA = createSoftware(dc, "Software A", Software.FEATURE_EXTRACTION);
			Software cufflinks = Util.findSoftware(dc, "Cufflinks");
			Software topHat = Util.findSoftware(dc, "TopHat");

			ReporterCloneTemplate cloneTemplate = createReporterCloneTemplate(dc, "Template A", "externalId", "symbol", "sequence");

			PlateType plateType = createPlateType(dc, "Plate type A", 16, 24);
			BioPlateType bioPlateType = createBioPlateType(dc, "Bioplate type A", null, BioWell.LockMode.UNLOCKED);
			
			PluginConfiguration platesImporter = PluginUtil.createPluginConfiguration(dc, "Plates for project A", 
					"net.sf.basedb.plugins.PlateFlatFileImporter", MouseData.getPlateParameters());

			PluginConfiguration genePixFeatureImporter = PluginUtil.createPluginConfiguration(dc, "Features from GenePix file",
					"net.sf.basedb.plugins.ReporterMapFlatFileImporter", MouseData.getGenePixReporterMapParameters());
			
			PluginDefinition printMapImporter = PluginDefinition.getByClassName(dc, "net.sf.basedb.plugins.PrintMapFlatFileImporter");

			PluginConfiguration gtfFeatureImporter = PluginUtil.createPluginConfiguration(dc, 
					"GTF features for project A", "net.sf.basedb.plugins.gtf.GtfReporterMapImporter", 
					SequenceData.getGtfFeatureParameters());
			
			PluginConfiguration rawDataImporter = PluginUtil.createPluginConfiguration(dc, "Raw data for project A", 
					"net.sf.basedb.plugins.RawDataFlatFileImporter", MouseData.getGenePixRawDataParameters(false));
			Util.setAnnotation(rawDataImporter, dyeSwap, false, null);
			
			PluginConfiguration rawDataImporterDyeSwap = PluginUtil.createPluginConfiguration(dc, "Raw data for project A (dye-swap)", 
					"net.sf.basedb.plugins.RawDataFlatFileImporter", MouseData.getGenePixRawDataParameters(true));
			Util.setAnnotation(rawDataImporterDyeSwap, dyeSwap, true, null);
		
			PluginDefinition arrayDesignImporter = PluginDefinition.getByClassName(dc, "net.sf.basedb.plugins.batchimport.ArrayDesignImporter");
			PluginDefinition arrayBatchImporter = PluginDefinition.getByClassName(dc, "net.sf.basedb.plugins.batchimport.ArrayBatchImporter");
			PluginDefinition arraySlideImporter = PluginDefinition.getByClassName(dc, "net.sf.basedb.plugins.batchimport.ArraySlideImporter");
			
			dc.commit();
			
			dc = TestUtil.getDbControl();

			File plates = FileUtil.uploadFile(dc, "/mouse", "plates_and_reporters.mouse.v4.37k.txt", File.PLATE);			
			File printMap = FileUtil.uploadFile(dc, "/mouse", "printmap.mouse.v4.37k.tam", File.PRINT_MAP);
			File cdfFile = FileUtil.uploadFile(dc, "/affymetrix/cdf", "MG_U74Av2.cdf.tar.bz2", File.REPORTER_MAP);
			File refSeqFile = FileUtil.uploadFile(dc, "/sequencing", "UCSC_Human_hg19_RefSeqGenes.gtf.tar.bz2", File.REPORTER_MAP);
			File batchImportFile_ArrayBatch = null;
			File batchImportFile_ArrayDesign = null;
			File batchImportFile_ArraySlide = null;
			if (useBatchImporters)
			{
				batchImportFile_ArrayBatch = FileUtil.uploadFile(dc, "/batchimport", "arraybatch_out.txt", null);
				batchImportFile_ArrayDesign = FileUtil.uploadFile(dc, "/batchimport", "arraydesign_out.txt", null);
				batchImportFile_ArraySlide = FileUtil.uploadFile(dc, "/batchimport", "arrayslide_out.txt", null);
			}
			dc.commit();
			
			importPlates(platesImporter, plates, "Plate A", plateType);
			
			ArrayDesign design = null;
			ArrayDesign affyDesign = null;
			ArrayDesign refSeqDesign = null;
			dc = TestUtil.getDbControl();
			Platform genericPlatform = Platform.getByExternalId(dc, Platform.GENERIC);
			Platform affymetrixPlatform = Platform.getByExternalId(dc, Platform.AFFYMETRIX);
			PlatformVariant seqPlatform = PlatformVariant.getByExternalId(dc, PlatformVariant.SEQUENCING_EXPRESSION);
			if (!useBatchImporters)
			{	
				design = createArrayDesign(dc, genericPlatform, "Array design A", "Plate A%");
				FileUtil.addDataFile(dc, design, DataFileType.GENERIC_PRINT_MAP, printMap, true);
				affyDesign = createArrayDesign(dc, affymetrixPlatform, "Affymetrix A", null);
				FileUtil.addDataFile(dc, affyDesign, DataFileType.AFFYMETRIX_CDF, cdfFile, true);
				refSeqDesign = createArrayDesign(dc, seqPlatform, "RefSeqDesign A", null);
				FileUtil.addDataFile(dc, refSeqDesign, DataFileType.REF_SEQ_GTF, refSeqFile, true);
			}
			else
			{
				importArrayDesigns(arrayDesignImporter, batchImportFile_ArrayDesign);
				design = Util.findArrayDesign(dc, "Array design A");
				design.setPlateList(Util.findPlates(dc, "Plate A%"));								
				affyDesign = Util.findArrayDesign(dc, "Affymetrix A");
				refSeqDesign = Util.findArrayDesign(dc, "RefSeqDesign A");
			}
			dc.commit();
			
			importFeatures(printMapImporter, printMap, design);
			importFeatures(gtfFeatureImporter, refSeqFile, refSeqDesign);
			
			dc = TestUtil.getDbControl();
			RawDataType genepix = RawDataTypes.getSafeRawDataType("genepix");
			p = Project.getById(dc, p.getId());
			p.addDefaultItem(genericPlatform);
			p.addDefaultItem(affymetrixPlatform);
			p.addDefaultItem(seqPlatform);
			p.setDefaultRawDataType(genepix);
			p.addDefaultItem(design);
			p.addDefaultItem(affyDesign);
			p.addDefaultItem(refSeqDesign);
			p.addDefaultItem(softwareA);
			p.addDefaultItem(cufflinks);
			p.addDefaultItem(topHat);
			p.addDefaultItem(hybStation);
			p.addDefaultItem(cBot);
			p.addDefaultItem(hiSeq);
			p.addDefaultItem(scanner);
			p.addDefaultItem(robot);
			p.addDefaultItem(scanning);
			p.addDefaultItem(sampling);
			p.addDefaultItem(extraction);
			p.addDefaultItem(labeling);
			p.addDefaultItem(libraryPrepartion);
			p.addDefaultItem(hybridization);
			p.addDefaultItem(cBotSettings);
			p.addDefaultItem(feature);
			p.addDefaultItem(hiSeqSettings);
			p.addDefaultItem(topHatSettings);
			p.addDefaultItem(cufflinksSettings);
			p.addDefaultItem(printing);			
			dc.commit();
			if (!useBatchImporters)
			{
				dc = TestUtil.getDbControl();
				ArrayBatch batch = createArrayBatch(dc, "Array batch A", design, robot, printing);
				ArraySlide slide1 = createArraySlide(dc, "Array slide A.1", batch, 1);
				ArraySlide slide2 = createArraySlide(dc, "Array slide A.2", batch, 2);
				ArraySlide slide3 = createArraySlide(dc, "Array slide A.3", batch, 3);
				ArraySlide slide4 = createArraySlide(dc, "Array slide A.4", batch, 4);
				
				// Affymetrix
				ArrayBatch affyBatch = createArrayBatch(dc, "Affymetrix batch A", affyDesign, null, null);
				ArraySlide affySlide1 = createArraySlide(dc, "Affymetrix slide A.1", affyBatch, 1);
				ArraySlide affySlide2 = createArraySlide(dc, "Affymetrix slide A.2", affyBatch, 2);
				ArraySlide affySlide3 = createArraySlide(dc, "Affymetrix slide A.3", affyBatch, 3);
				dc.commit();
			}
			else
			{
				importArrayBatches(arrayBatchImporter, batchImportFile_ArrayBatch);
				importArraySlides(arraySlideImporter, batchImportFile_ArraySlide);
			}
			TestUtil.logout();
		}
		catch (Throwable t)
		{
			ok = false;
			t.printStackTrace();
		}
		finally
		{
			if (dc != null) dc.close();
		}
		TestUtil.write("++Leaving power user test (" + (ok ? "OK" : "FAILED") + ")\n");
		return ok;
	}
	
	/**
		Create a project.
	*/
	public static Project createProject(DbControl dc, String name)
	{
		TestUtil.write("--Creating project: " + name + "\n");
		
		Project p = Project.getNew(dc);
		p.setName("Project A");
		
		Group groupA = Util.findGroup(dc, "Group A");
		
		p.setPermissions(groupA, EnumSet.of(Permission.USE));
		dc.saveItem(p);
		return p;
	}

	/**
		Create an item subtype.
	*/
	public static ItemSubtype createItemSubtype(DbControl dc, String name, Item itemType, 
			boolean pushAnnotations, ItemSubtype... related)
	{
		TestUtil.write("--Creating item subtype: " + name + "\n");
		ItemSubtype subtype = ItemSubtype.getNew(dc, itemType);
		subtype.setName(name);
		subtype.setPushAnnotations(pushAnnotations);
		if (related != null)
		{
			for (ItemSubtype r : related)
			{
				subtype.setRelatedSubtype(r);
			}
		}
		dc.saveItem(subtype);
		return subtype;
	}
	
	/**
		Create a protocol.
	*/
	public static Protocol createProtocol(DbControl dc, String name, String type, AnnotationType... parameters)
	{
		TestUtil.write("--Creating protocol: " + name + "\n");
		ItemSubtype pt = ItemSubtype.getById(dc, SystemItems.getId(type));
		Protocol p = Protocol.getNew(dc);
		p.setItemSubtype(pt);
		p.setName(name);
		if (parameters != null)
		{
			for (AnnotationType pp : parameters)
			{
				p.addParameter(pp);
			}
		}
		dc.saveItem(p);
		return p;
	}
	
	/**
		Create a reporter clone template.
	*/
	public static ReporterCloneTemplate createReporterCloneTemplate(DbControl dc, String name, String... properties)
	{
		TestUtil.write("--Creating reporter clone template: " + name + "\n");
		ReporterCloneTemplate t = ReporterCloneTemplate.getNew(dc);
		t.setName(name);
		if (properties != null)
		{
			for (String property : properties)
			{
				t.addClonableProperty(ClonableProperty.create(property));
			}
		}
		dc.saveItem(t);
		return t;
	}
	
	/**
		Create a hardware.
	*/
	public static Hardware createHardware(DbControl dc, String name, String type)
	{
		TestUtil.write("--Creating hardware: " + name + "\n");
		ItemSubtype ht = ItemSubtype.getById(dc, SystemItems.getId(type));
		Hardware hw = Hardware.getNew(dc);
		hw.setItemSubtype(ht);
		hw.setName(name);
		dc.saveItem(hw);
		return hw;
	}

	/**
		Create a software.
	*/
	public static Software createSoftware(DbControl dc, String name, String type)
	{
		TestUtil.write("--Creating software: " + name + "\n");
		ItemSubtype st = ItemSubtype.getById(dc, SystemItems.getId(type));
		Software sw = Software.getNew(dc);
		sw.setItemSubtype(st);
		sw.setName(name);
		dc.saveItem(sw);
		return sw;
	}
	
	/**
		Create an annotation type.
	*/
	public static AnnotationType createAnnotationType(DbControl dc, String name, 
			Type type, Unit unit, Object[] values, Item[] items)
	{
		TestUtil.write("--Creating annotation type: " + name + "\n");
		
		AnnotationType at = AnnotationType.getNew(dc, type);
		at.setName(name);
		at.setDefaultUnit(unit);
		if (items != null)
		{
			for (int i = 0; i < items.length; ++i)
			{
				at.enableForItem(items[i]);
			}
		}
		if (values != null)
		{
			at.setEnumeration(true);
			at.setDisplayAsList(false);
			at.setValues(Arrays.asList(values));
		}
		dc.saveItem(at);
		return at;
	}

	/**
		Create a protocol parameter.
	*/
	public static AnnotationType createProtocolParameter(DbControl dc, String name, 
			Type type, Unit unit, Object[] values, Item[] items)
	{
		TestUtil.write("--Creating protocol parameter: " + name + "\n");
		
		AnnotationType at = createAnnotationType(dc, name, type, unit, values, items);
		at.setProtocolParameter(true);
		return at;
	}

	
	/**
		Create a plate type.
	*/
	public static PlateType createPlateType(DbControl dc, String name, int rows, int columns)
	{
		TestUtil.write("--Creating plate type: " + name + "\n");
		
		PlateGeometry geometry = Util.findGeometry(dc, rows, columns);
		PlateType pt = PlateType.getNew(dc, geometry);
		pt.setName(name);
		dc.saveItem(pt);
		return pt;
	}
	
	/**
		Create a bioplate type.
	*/
	public static BioPlateType createBioPlateType(DbControl dc, String name, Item bioMaterialType, BioWell.LockMode lockMode)
	{
		TestUtil.write("--Creating bioplate type: " + name + "\n");
		
		BioPlateType pt = BioPlateType.getNew(dc);
		pt.setName(name);
		pt.setBioMaterialType(bioMaterialType);
		pt.setLockMode(lockMode);
		dc.saveItem(pt);
		return pt;
	}

	
	/**
		Create an array design.
	*/
	public static ArrayDesign createArrayDesign(DbControl dc, Platform platform, String name, String linkToPlatesPattern)
	{
		TestUtil.write("--Creating array design: " + name + "\n");
		ArrayDesign design = ArrayDesign.getNew(dc, platform);
		design.setName(name);
		
		if (linkToPlatesPattern != null)
		{
			List<Plate> plates = Util.findPlates(dc, linkToPlatesPattern);
			design.setPlateList(plates);
		}
		
		dc.saveItem(design);
		return design;
	}

	/**
		Create an array design.
	*/
	public static ArrayDesign createArrayDesign(DbControl dc, PlatformVariant variant, String name, String linkToPlatesPattern)
	{
		TestUtil.write("--Creating array design: " + name + "\n");
		ArrayDesign design = ArrayDesign.getNew(dc, variant);
		design.setName(name);
		
		if (linkToPlatesPattern != null)
		{
			List<Plate> plates = Util.findPlates(dc, linkToPlatesPattern);
			design.setPlateList(plates);
		}
		
		dc.saveItem(design);
		return design;
	}

	/**
		Create an array batch.
	*/
	public static ArrayBatch createArrayBatch(DbControl dc, String name, ArrayDesign design, Hardware robot, Protocol protocol)
	{
		TestUtil.write("--Creating array batch: " + name + "\n");
		ArrayBatch batch = ArrayBatch.getNew(dc, design);
		batch.setName(name);
		batch.setPrintRobot(robot);
		batch.setProtocol(protocol);
		dc.saveItem(batch);
		return batch;
	}

	/**
		Create an array slide.
	*/
	public static ArraySlide createArraySlide(DbControl dc, String name, ArrayBatch batch, int index)
	{
		TestUtil.write("--Creating array slide: " + name + "\n");
		ArraySlide slide = ArraySlide.getNew(dc, batch);
		slide.setName(name);
		slide.setBatchIndex(index);
		dc.saveItem(slide);
		return slide;
	}
	
	public static void importPlates(PluginConfiguration config, File file, String plateNamePrefix, PlateType plateType)
	{
		TestUtil.write("--Importing plates from file: " + file.getName() + "\n");
		DbControl dc = TestUtil.getDbControl();
		try
		{
			config = PluginConfiguration.getById(dc, config.getId());
			Job job = Job.getNew(dc, config.getPluginDefinition(), config, null);
			job.setName("Importing plates for project A");
			job.setParameterValue("file", new FileParameterType(), file);
			job.setParameterValue("plateType", new ItemParameterType<PlateType>(PlateType.class, null), plateType);
			job.setParameterValue("plateNamePrefix", new StringParameterType(), plateNamePrefix);
			job.setParameterValue("namePadding", new IntegerParameterType(), 4);

			dc.saveItem(job);
			dc.commit();
			PluginUtil.executeJob(job);
		}
		finally
		{
			if (dc != null) dc.close();
		}
		
	}
	
	/**
		Import features.
	*/
	public static void importFeatures(PluginDefinition plugin, File file, ArrayDesign design)
		throws Exception
	{
		TestUtil.write("--Importing features from file: " + file.getName() + "\n");
		DbControl dc = TestUtil.getDbControl();
		try
		{
			plugin = PluginDefinition.getById(dc, plugin.getId());
			Job job = Job.getNew(dc, plugin, null, null);
			job.setName("Importing features to array design: " + design.getName());
			job.setParameterValue("file", new FileParameterType(), file);
			job.setParameterValue("arrayDesign", 
				new ItemParameterType<ArrayDesign>(ArrayDesign.class, null), design);
			dc.saveItem(job);
			dc.commit();
			PluginUtil.executeJob(job);
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	/**
		Import features.
	*/
	public static void importFeatures(PluginConfiguration config, File file, ArrayDesign design)
		throws Exception
	{
		TestUtil.write("--Importing features from file: " + file.getName() + "\n");
		DbControl dc = TestUtil.getDbControl();
		try
		{
			config = PluginConfiguration.getById(dc, config.getId());
			Job job = Job.getNew(dc, config.getPluginDefinition(), config, null);
			job.setName("Importing features to array design: " + design.getName());
			job.setParameterValue("file", new FileParameterType(), file);
			job.setParameterValue("arrayDesign", 
				new ItemParameterType<ArrayDesign>(ArrayDesign.class, null), design);
			dc.saveItem(job);
			dc.commit();
			PluginUtil.executeJob(job);
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	
	public static void importArrayBatches(PluginDefinition batchImporter, File file)
	{
		TestUtil.write("--Importing array batches from file: " + file + "\n");
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PluginDefinition plugin = PluginDefinition.getById(dc, batchImporter.getId());
			Job job = Job.getNew(dc, plugin, null, null);
			job.setName("Batch import of Array batches to Project A");
			job.setParameterValue("file", new FileParameterType(), file);
			job.setParameterValue("dataSplitterRegexp", new StringParameterType(), "\\t");
			job.setParameterValue("dataHeaderRegexp", new StringParameterType(), "Name\\tArray design\\tDescription\\tProtocol\\tPrint robot");
			
			job.setParameterValue("nameColumnMapping", new StringParameterType(), "\\Name\\");
			job.setParameterValue("descriptionColumnMapping", new StringParameterType(), "\\Description\\");
			job.setParameterValue("hardwareColumnMapping", new StringParameterType(), "\\Print robot\\");
			job.setParameterValue("arrayDesignColumnMapping", new StringParameterType(), "\\Array design\\");
			job.setParameterValue("protocolColumnMapping", new StringParameterType(), "\\Protocol\\");
			
			dc.saveItem(job);
			dc.commit();
			PluginUtil.executeJob(job);
		}
		finally
		{
			if (dc != null) dc.close();
		}
		
	}

	public static void importArrayDesigns(PluginDefinition batchImporter, File file)
	{
		TestUtil.write("--Importing array designs from file: " + file.getName() + "\n");
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PluginDefinition plugin = PluginDefinition.getById(dc, batchImporter.getId());
			Job job = Job.getNew(dc, plugin, null, null);
			job.setName("Batch import of Array designs to Project A");
			job.setParameterValue("file", new FileParameterType(), file);
			job.setParameterValue("dataSplitterRegexp", new StringParameterType(), "\\t");
			job.setParameterValue("dataHeaderRegexp", new StringParameterType(), "Name\\tDescription\\tArrays/slide\\tPlatform\\tVariant\\tFile type\\tFile");
			
			// Item properties
			job.setParameterValue("nameColumnMapping", new StringParameterType(), "\\Name\\");
			job.setParameterValue("descriptionColumnMapping", new StringParameterType(), "\\Description\\");
			job.setParameterValue("numArraysColumnMapping", new StringParameterType(), "\\Arrays/slide\\");
			job.setParameterValue("platformColumnMapping", new StringParameterType(), "\\Platform\\");
			job.setParameterValue("variantColumnMapping", new StringParameterType(), "\\Variant\\");
			job.setParameterValue("fileTypeColumnMapping", new StringParameterType(), "\\File type\\");
			job.setParameterValue("fileColumnMapping", new StringParameterType(), "\\File\\");
			
			dc.saveItem(job);
			dc.commit();
			PluginUtil.executeJob(job);
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	public static void importArraySlides(PluginDefinition batchImporter, File file)
	{
		TestUtil.write("--Importing array slides from file: " + file.getName() + "\n");
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PluginDefinition plugin = PluginDefinition.getById(dc, batchImporter.getId());
			Job job = Job.getNew(dc, plugin, null, null);
			job.setName("Batch import of Array slides to Project A");
			job.setParameterValue("file", new FileParameterType(), file);
			job.setParameterValue("dataSplitterRegexp", new StringParameterType(), "\\t");
			job.setParameterValue("dataHeaderRegexp", new StringParameterType(), "Name\\tHybridization\\tDescription\\tBarcode\\tDestroyed\\tArray batch\\tIndex");
			
			// Item properties
			job.setParameterValue("nameColumnMapping", new StringParameterType(), "\\Name\\");
			job.setParameterValue("arrayBatchColumnMapping", new StringParameterType(), "\\Array batch\\");
			job.setParameterValue("hybridizationColumnMapping", new StringParameterType(), "\\Hybridization\\");
			job.setParameterValue("descriptionColumnMapping", new StringParameterType(), "\\Description\\");
			job.setParameterValue("barcodeColumnMapping", new StringParameterType(), "\\Barcode\\");
			job.setParameterValue("destroyedColumnMapping", new StringParameterType(), "\\Destroyed\\");
			job.setParameterValue("indexColumnMapping", new StringParameterType(), "\\Index\\");
			
			dc.saveItem(job);
			dc.commit();
			PluginUtil.executeJob(job);
		}
		finally 
		{
			if (dc != null) dc.close();
		}
	}
}
